#include  <stdlib.h>
#include  <stdio.h>
#include  <string.h>
#include  <ctype.h>
#include  "geosim.h"

/*------------------------------------------------------------------------------
  Named Constants
------------------------------------------------------------------------------*/

#define  CLEAR_DRAW_MENU_ITEM     "ClearDraw"
#define  CLEAR_TEXT_MENU_ITEM     "ClearText"
#define  FILL_COLOR_LIST          0
#define  INITIAL_ITEM             0
#define  TIME_TO_READ_MSG         2
#define  FILL_COLOR_BUTTON_LABEL  "Fill Color:"

/*--- Help message numbers ---*/
#define  AGAIN_MSG                2
#define  BOX_STYLE_CHECK_OFF_MSG  6
#define  BOX_STYLE_CHECK_ON_MSG   5
#define  BOX_STYLE_MENU_MSG       4
#define  FILL_COLOR_MENU_MSG      7
#define  INTRO_MSG                1
#define  MENU_MENU_MSG            3
#define  NEW_FILL_COLOR_MSG       8
#define  SAME_FILL_COLOR_MSG      9

/*------------------------------------------------------------------------------
  Drawing Object declarations
------------------------------------------------------------------------------*/

typedef enum {
  rectObj, textObj
  } ObjType;

typedef struct {
  int x, y;
  char *str;
  GSColor t_col;
  } TextObj;
	
typedef struct {
  int x, y, wd, ht;
  bool fill, border;
  GSColor f_col, b_col;
  } RectObj;
	
typedef struct DrawObj {
  ObjType type;
  union {
    TextObj text;
    RectObj rect;
  } object;
  struct DrawObj *next_object;
  } DrawObj;
	
/*------------------------------------------------------------------------------
  Function Prototypes
------------------------------------------------------------------------------*/

void     add_cursor_if_space  (char *out_string, int  field_wd);
DrawObj  *add_rect            (int x, int y, int wd, int ht, 
                               bool fill, bool border, 
                               GSColor f_col, GSColor b_col);
DrawObj  *add_text            (int x, int y, char *str, GSColor t_col);
bool     alert_done           (void);
bool     char_is_legal        (char ch);
bool     clear_draw           (void);
bool     clear_text           (void);
bool     do_again             (void);
bool     do_draw              (int mouse_x, int mouse_y, DragStatusType status);
bool     do_intro             (void);
bool     do_quit              (void);
bool     do_text              (int mouse_x, int mouse_y, char ch);
void     draw_all_objects     (void);
void     draw_object          (DrawObj *object);
void     init_pict            (GSPicture *pict_ptr, FileName pict_file);
void     init_colors          (void);
GSColor  logo_converter       (unsigned short code, int x, int y, int z);
bool     set_color            (char *item_name);
bool     show_funcs_menu      (void);
bool     show_box_styles      (void);
bool     show_colors          (void);
void     show_fill_color      (GSColor);
bool     toggle_border        (char *item_name);
bool     toggle_fill          (char *item_name);
void     write_hello          (void);
void     write_hello_again    (void);
bool     redraw_all           (void);
void     remove_all_objects   (void);

/*------------------------------------------------------------------------------
  Global Variables
------------------------------------------------------------------------------*/

bool       AlertShowing = FALSE;    /*--- Blocks typing when TRUE ---*/
bool       DrawBorder = FALSE;      /*--- Boxes get borders iff TRUE ---*/
bool       FillBorder = FALSE;      /*--- Boxes get filled iff TRUE ---*/
GSColor    FillColor;               /*--- Color to fill boxes with ---*/
FileName   GlobeFile = "logo.rl4";  /*--- File that contains globe logo ---*/
/*FileName   GlobeFile = "usmap.rl4";  /*--- File that contains globe logo ---*/
GSPicture  LogoPict;                /*--- Structure for globe logo ---*/
BigString  TextString;              /*--- User's input string ---*/
DrawObj    *first_object, *last_object;  /*--- Object list start and end ---*/
GSColor    aqua_color, mgray_color, white_color;  /*--- Logo colors ---*/

/*------------------------------------------------------------------------------
  Function Definitions
------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
  add_cursor_if_space 
    Add a text cursor ('_') to end of output string if there is enough
    space (in memory and on screen) for it.
------------------------------------------------------------------------------*/
void add_cursor_if_space (
  char *out_string, /*--- output string ---*/
  int  field_wd     /*--- width of output field ---*/
)
{
  /*--- if space in memory for text cursor, add it to temp buffer ---*/
  if (strlen(out_string) < BIGSTRINGLEN)
  {
    strcat(out_string, "_");
    /*--- if not enough space on screen for text cursor, remove it ---*/
    if (GStextwidth(out_string) > field_wd)
    {
      out_string[strlen(out_string)-1] = '\0';
    }
  }
}


/*------------------------------------------------------------------------------
  add_rect 
    Add a rectangle to the list of drawing objects.
------------------------------------------------------------------------------*/
DrawObj *add_rect(
  int x, int y, int wd, int ht, 
  bool fill, bool border, 
  GSColor f_col, GSColor b_col
)
{
  DrawObj *new_object;
  RectObj *rect;
  
  if ((fill == FALSE) && (border == FALSE))
    return NULL;
    
  new_object = (DrawObj *) malloc(sizeof(DrawObj));
  if (new_object == NULL) {
    GSelog(0, "Out of memory for new rectangle.\n");
    return NULL;
    }
    
  new_object->type = rectObj;
  
  rect = &(new_object->object.rect);
  rect->x = x;
  rect->y = y;
  rect->wd = wd;
  rect->ht = ht;
  rect->fill = fill;
  rect->border = border;
  rect->f_col = f_col;
  rect->b_col = b_col;
  
  /* put new object in object list */
  if (first_object == NULL)
    first_object = new_object;
  else
    last_object->next_object = new_object;
  last_object = new_object;
  new_object->next_object = NULL;

  GSenablemenuitem("funcs_menu", CLEAR_DRAW_MENU_ITEM);
  
  return new_object;
}


/*------------------------------------------------------------------------------
  add_rect 
    Add a text object to the list of drawing objects.
------------------------------------------------------------------------------*/
DrawObj *add_text(
  int x, int y, 
  char *str, 
  GSColor t_col
)
{
  DrawObj *new_object;
  TextObj *text;
  
  if (strlen(str) == 0)
    return NULL;
  
  new_object = (DrawObj *) malloc(sizeof(DrawObj));
  if (new_object == NULL) {
    GSelog(0, "Out of memory for new text.\n");
    return NULL;
    }
    
  text = &(new_object->object.text);
  
  text->str = (char *) malloc(strlen(str) + 1); //Typecast malloc to 
                                                //char 23 Jan jrf
  
  if (text->str == NULL) {
  	free(new_object);
    GSelog(0, "Out of memory for text string %s.\n", str);
    return NULL;
    }
  	
  new_object->type = textObj;
  
  text->x = x;
  text->y = y;
  text->t_col = t_col;
  strcpy(text->str, str);
  
  /* put new object in object list */
  if (first_object == NULL)
    first_object = new_object;
  else
    last_object->next_object = new_object;
  last_object = new_object;
  new_object->next_object = NULL;
  
  GSenablemenuitem("funcs_menu", CLEAR_DRAW_MENU_ITEM);

  return new_object;
}


/*------------------------------------------------------------------------------
  alert_done
    Called when "Done" button in PopUpMessage window is pushed.
    Removes the PopUpMessage window from the screen.
------------------------------------------------------------------------------*/
bool alert_done (
  void
)
{
  GSremovenamedwindow("alert");
  GSsetcurrentnamedwindow("root");
  show_fill_color(FillColor);
  GSsetcurrentnamedwindow("io");
  AlertShowing = FALSE;
  return TRUE;
}


/*------------------------------------------------------------------------------
  char_is_legal
    Returns TRUE if ch is a legal input character (A-Z, a-z, '.', '-', or ' ').
    Returns FALSE if ch is not a legal input character.
------------------------------------------------------------------------------*/
bool char_is_legal (
  char  ch  /*--- character to check for legality ---*/
)
{
  return ((isalnum(ch)) || (ch == '.') || (ch == '-') || (ch == ' '));
}


/*------------------------------------------------------------------------------
  clear_draw
    Called when "Clear Draw Area" item on "Menu" menu is selected.
    Clears the draw area.
------------------------------------------------------------------------------*/
bool clear_draw (
  void
)
{
  int      draw_area_x, draw_area_y, draw_area_wd, draw_area_ht;
  GSColor  draw_area_color;

  remove_all_objects();
  
  /*--- draw area is part of "io" window ---*/
  GSsetcurrentnamedwindow("io");

  /*--- 'clear' draw area by refilling with its background color ---*/
  GSgetfieldrect("dragbg", &draw_area_x, &draw_area_y, &draw_area_wd,
                 &draw_area_ht, &draw_area_color);
  GSsetcolor(draw_area_color);
  GSwfillrect(draw_area_x, draw_area_y, draw_area_wd, draw_area_ht);

  /*--- disable "Clear Draw Area" menu item, since area is now cleared ---*/
  GSdisablemenuitem("funcs_menu", CLEAR_DRAW_MENU_ITEM);
  return TRUE;
}

/*------------------------------------------------------------------------------
  clear_text
    Called when "Clear Text Field" item on "Menu" menu is selected.
    Clears the text input field.
------------------------------------------------------------------------------*/
bool clear_text (
  void
)
{
  int      background_x, background_y, background_wd, background_ht;
  GSColor  background_color;
  int      cursor_x, cursor_y;
  GSColor  cursor_color;

  strcpy(TextString, "");
  GSsetcurrentnamedwindow("io");

  /*--- 'clear' field by refilling it with its background color ---*/
  GSgetfieldrect("textbg", &background_x, &background_y, &background_wd,
                 &background_ht, &background_color);
  GSsetcolor(background_color);
  GSwfillrect(background_x, background_y, background_wd, background_ht);

  /*--- draw text cursor; use NULL for unneeded wd, ht in GSgetfieldrect ---*/
  GSgetfieldrect("textfg", &cursor_x, &cursor_y, NULL, NULL, &cursor_color);
  GSsetcolor(cursor_color);
  GSwwritemsg(cursor_x, cursor_y, "_");

  /*--- disable "Clear Text Area" button, since area is now cleared ---*/
  GSdisablemenuitem("funcs_menu", CLEAR_TEXT_MENU_ITEM);
  return TRUE;
}


/*------------------------------------------------------------------------------
  do_again         
    Called when "Again" button in "IOwin" window is pressed.
    Writes "Hello World" in field "hello2".
------------------------------------------------------------------------------*/
bool do_again (
  void
)
{
  write_hello_again();

  /*--- erase button, since pressing it again will have no effect ---*/
  GSinvisiblebutton("io", "again", DRAW);

  GSdisplayhelpmessage(AGAIN_MSG);
  return TRUE;
}

/*------------------------------------------------------------------------------
  do_draw
    Processes drag events in the draw area.
------------------------------------------------------------------------------*/
bool do_draw (
  int  mouse_x, /*--- x-coordinate of mouse cursor ---*/
  int  mouse_y, /*--- x-coordinate of mouse cursor ---*/
  DragStatusType  status   /*--- type of mouse event ---*/
)
{
  int         dfg_x, dfg_y;
  GSColor     dfg_color;
  int         dbg_x, dbg_y;
  int         cfg_x, cfg_y;
  GSColor     cfg_color;
  int         cbg_x, cbg_y, cbg_wd, cbg_ht;
  GSColor     cbg_color;
  static int  initial_x, initial_y;  
  static int  xor_x, xor_y, xor_wd, xor_ht;
  BigString   coordinate_string;
  DrawObj     *new_object;
  
  switch (status)
  {
    case MOUSE_CLICK: /*--- write contents of text area at mouse cursor ---*/
      GSgetfieldrect("dragfg", &dfg_x, &dfg_y, NULL, NULL, &dfg_color);
      
      new_object = add_text(mouse_x+dfg_x, mouse_y+dfg_y, TextString, dfg_color);
      if (new_object != NULL)
        draw_object(new_object);
      break;
      
    case DRAG_INIT: /*--- initilize XOR box coordinates and dimensions ---*/
      xor_x = initial_x = mouse_x;
      xor_y = initial_y = mouse_y;
      xor_wd = xor_ht = 0;
      break;
      
    case DRAG_PROCESS: /*--- move XOR box ---*/
      GSgetfieldrect("dragbg", &dbg_x, &dbg_y, NULL, NULL, NULL);
      
      /*--- erase old XOR box ---*/
      GSdrawxorbox(dbg_x+xor_x, dbg_y+xor_y, xor_wd, xor_ht);
      
      /*--- compute location and dimensions of new XOR box ---*/
      xor_x = MIN(mouse_x, initial_x);
      xor_y = MIN(mouse_y, initial_y);
      xor_wd = abs(mouse_x - initial_x);
      xor_ht = abs(mouse_y - initial_y);
      
      /*--- draw new XOR box ---*/
      GSdrawxorbox(dbg_x+xor_x, dbg_y+xor_y, xor_wd, xor_ht);
      break;
      
    case DRAG_FINAL: /*--- erase XOR box; draw a box if parameters are set ---*/
      GSgetfieldrect("dragbg", &dbg_x, &dbg_y, NULL, NULL, NULL);
      
      /*--- erase XOR box ---*/
      GSdrawxorbox(dbg_x+xor_x, dbg_y+xor_y, xor_wd, xor_ht);
      
      GSgetfieldrect("dragfg", &dfg_x, &dfg_y, NULL, NULL, &dfg_color);
      
      new_object = add_rect(dbg_x+xor_x, dbg_y+xor_y, xor_wd, xor_ht,
                         FillBorder, DrawBorder, FillColor, dfg_color);
      if (new_object != NULL)
        draw_object(new_object);
      break;
      
    case MOUSE_INSIDE: /*--- show current mouse cursor coordinates ---*/
      /*--- 'erase' old coordinates ---*/
      GSgetfieldrect("coordbg", &cbg_x, &cbg_y, &cbg_wd, &cbg_ht, &cbg_color);
      GSsetcolor(cbg_color);
      GSwfillrect(cbg_x, cbg_y, cbg_wd, cbg_ht);
      /*--- show new coordinates ---*/
      GSgetfieldrect("coordfg", &cfg_x, &cfg_y, NULL, NULL, &cfg_color);
      GSsetcolor(cfg_color);
      sprintf(coordinate_string, "(%d, %d)", mouse_x, mouse_y);
      GSwwritemsg(cfg_x, cfg_y, coordinate_string);
      break;
      
    case MOUSE_OUTSIDE: /*--- erase mouse cursor coordinates ---*/
      GSgetfieldrect("coordbg", &cbg_x, &cbg_y, &cbg_wd, &cbg_ht, &cbg_color);
      GSsetcolor(cbg_color);
      GSwfillrect(cbg_x, cbg_y, cbg_wd, cbg_ht);
      break;
      
  } /*--- switch (status) ---*/
  return(TRUE); 
}

/*------------------------------------------------------------------------------
  do_intro
    Writes "Hello World" in field "hello1".
    Demonstrates the use and deactivation of independent functions.
    This function is activated in the main function.
------------------------------------------------------------------------------*/
bool do_intro (
  void
)
{
  int      logo_x, logo_y, logo_wd, logo_ht;
  int      text_x, text_y, text_wd, text_ht;
  GSColor  text_color;

  /*--- deactivate this function (it should only ve called once) ---*/
  GSdeactivatefunction("intro");

  /*--- initialize fill color field to first color of fillcolor list ---*/
  GSsetcurrentnamedwindow("root");
  FillColor = GSgetlistcolor(FILL_COLOR_LIST, INITIAL_ITEM);
  show_fill_color(FillColor);
  GStogglecheck("col_menu", GSgetmenuitemname("col_menu",INITIAL_ITEM));
  GSdisplayversion("version.txt");

  write_hello();

  GSdisplayhelpmessage(INTRO_MSG);

  /*--- draw text cursor in text output field ---*/
  GSgetfieldrect("textfg", &text_x, &text_y, &text_wd, &text_ht, &text_color);
  GSsetcolor(text_color);
  GSwwritemsg(text_x, text_y, "_");

  /*--- draw globe logo ---*/
  GSgetfieldrect("logo", &logo_x, &logo_y, &logo_wd, &logo_ht, NULL);
  GSwdrawpict(&LogoPict, logo_x, logo_y, logo_wd, logo_ht, 0, 0, 1,
              logo_converter);
  return TRUE;
}

/*------------------------------------------------------------------------------
  do_quit
    Called when "Quit" item on "Menu" menu is selected.
    Quits giltest.
------------------------------------------------------------------------------*/
bool do_quit (
  void
)
{
  GSquit(NULL);
  return TRUE;
}

/*------------------------------------------------------------------------------
  do_text
    Called when a character is typed at the keyboard.
    Processes a character typed at the keyboard.
------------------------------------------------------------------------------*/
bool do_text (
  int   mouse_x,    /*--- not used here, but required by event handler ---*/
  int   mouse_y,    /*--- not used here, but required by event handler ---*/
  char  ch          /*--- character typed at the keyboard ---*/
)
{
  int        text_x, text_y, text_wd, text_ht;
  GSColor    text_color;
  int        background_x, background_y, background_wd, background_ht;
  GSColor    background_color;
  int        length;
  BigString  current_window;
  BigString  out_string;
  int        i;

  /*--- Don't process character if alert window is up. ---*/
  if (AlertShowing)
  {
    GSbell();
    return (FALSE);
  }

  /*--- Don't process character unless "io" is the current window. ---*/
  GSgetcurrentwindowname(current_window);
  length = strlen(current_window);
  for (i = 0; i < length; i++)
  {
    current_window[i] = tolower(current_window[i]);
  }
  if (strcmp(current_window, "io") != 0)
  {
    GSbell();
    return (FALSE);
  }

  GSsetcurrentnamedwindow("io");
  GSgetfieldrect("textfg", &text_x, &text_y, &text_wd, &text_ht, &text_color);
  if (char_is_legal(ch))
  {
    length = strlen(TextString);
    /*--- check length of string (in characters) in memory ---*/
    if (length < BIGSTRINGLEN)
    {
      /*--- copy string and add character to temporary buffer ---*/
      strcpy(out_string, TextString);
      out_string[length] = ch;
      out_string[length+1] = '\0';
      /*--- check length of string (in pixels) on screen ---*/
      if (GStextwidth(out_string) <= text_wd)
      {
        /*--- update permanent buffer if string still fits in field ---*/
        strcpy(TextString, out_string);
        if (strlen(TextString) == 1)
        {
          GSenablemenuitem("funcs_menu", CLEAR_TEXT_MENU_ITEM);
        }
      }
      else /*--- string will not fit in field with new character added ---*/
      {
        /*--- alert user that string would be too long ---*/
        GSalert(-1, TRUE, "typedtoofar", "");
        show_fill_color(GSconvertcolor("mgray"));
        /*--- disable typing until alert window is removed by user ---*/
        AlertShowing = TRUE;
        return FALSE;
      }
    }
  }
  else if ((ch == '\b') && (strlen(TextString) > 0))
  { /*--- rubout last character ---*/
    TextString[strlen(TextString)-1] = '\0';
    if (strlen(TextString) == 0) /*--- then there is no more text to clear ---*/
    {
      GSdisablemenuitem("funcs_menu", CLEAR_TEXT_MENU_ITEM);
    }
    strcpy(out_string, TextString);
  }
  else
  {
    return FALSE;
  }
  /*--- if we get here, the string has been updated, so rewrite it ---*/
  add_cursor_if_space(out_string, text_wd);
  GSgetfieldrect("textbg", &background_x, &background_y, &background_wd,
                 &background_ht, &background_color);
  GSsetcolor(background_color);
  GSwfillrect(background_x, background_y, background_wd, background_ht);
  GSsetcolor(text_color);
  GSsetfont(LARGE);
  GSwwritemsg(text_x, text_y, out_string);
  return TRUE;
}


/*------------------------------------------------------------------------------
  draw_all_objects
    Display all drawing objects.
------------------------------------------------------------------------------*/
void draw_all_objects(
  void
)
{
  DrawObj *object = first_object;
  
  while (object != NULL) {
    draw_object(object);
    object = object->next_object;
    }
}


/*------------------------------------------------------------------------------
  draw_object
    Display a drawing object.  Visual C++  doesn't like this void
------------------------------------------------------------------------------*/
void draw_object(
  DrawObj *object
)
{
  TextObj *text;
  RectObj *rect;
  
  GSsetcurrentnamedwindow("io");

  switch(object->type) {
  
    case rectObj:
      rect = &(object->object.rect);
      if (rect->fill) {
        GSsetcolor(rect->f_col);
        GSwfillrect(rect->x, rect->y, rect->wd, rect->ht);
        }
      if (rect->border) {
        GSsetcolor(rect->b_col);
        GSwdrawrect(rect->x, rect->y, rect->wd, rect->ht);
        }
      break;
      
    case textObj:
      text = &(object->object.text);
      GSsetcolor(text->t_col);
      GSwwritemsg(text->x, text->y, text->str);
      break;
      
    default:
      GSelog(-1, "Error:  unknown object type: %d.\n", object->type);
      break;
    }
}


/*------------------------------------------------------------------------------
  init_colors
    Convert color names to GSColors for logo.
------------------------------------------------------------------------------*/
void init_colors(
  void
)
{
  aqua_color = GSconvertcolor("aqua");
  mgray_color = GSconvertcolor("mgray");
  white_color = GSconvertcolor("white");
}

/*------------------------------------------------------------------------------
  init_pict
    Initialize the file for a picture and initialize the picture.
------------------------------------------------------------------------------*/
void init_pict (
  GSPicture *pict_ptr,  /*--- GIL picture structure ---*/
  FileName   pict_file  /*--- name of file which contains picture image ---*/
)
{
  FILE *fp;
  long  file_size;
  void *picture_memory;

  fp = GSopen(pict_file, "rb");
  if (fp == NULL)
  {
    GSelog(-1, "Unable to open logo file %s\n", pict_file);
  }
  file_size = GSfilesize(fp);
  picture_memory = (void *)malloc(file_size-8);
  if (picture_memory == NULL)
  {
    GSelog(-1, "Unable to allocate space for picture in file %s\n", pict_file);
  }
  if (!GSinitpic(fp, pict_ptr, picture_memory))
  {
    GSelog(-1, "Unable to initialize picture in file %s\n", pict_file);
  }
  fclose(fp);
}


/*------------------------------------------------------------------------------
  logo_converter
    Converts a pixel code from a runlength encoded image file to a color
    palette index.
------------------------------------------------------------------------------*/
GSColor logo_converter (
  unsigned short code,    /*--- pixel code from image run ---*/
  int x,
  int y,  /* x, y, z - unused params */
  int z
)
{
  switch (code)
  {
    case 255:
      return((unsigned char)-1);
    case 0:
      return(mgray_color);
    case 1:
      return(aqua_color);
    default:
      return(white_color);
  }
}


/*------------------------------------------------------------------------------
  redraw_all
    Called when the application window is exposed after being covered by
    another window. Redraws the application window.
------------------------------------------------------------------------------*/
bool redraw_all (
  void
)
{
  ButtonStatusType button_status;
  int              text_x, text_y, text_wd;
  GSColor          text_color;
  int              logo_x, logo_y, logo_wd, logo_ht;
  BigString        out_string;

  /*--- GSredrawnamedwindow is used to redraw a window which already
        appears on the screen ---*/
  GSredrawnamedwindow("root");
  GSdisplayversion("version.txt");
  show_fill_color(FillColor);
  GSredrawnamedwindow("io");
  
  /*--- redraw globe logo ---*/
  GSgetfieldrect("logo", &logo_x, &logo_y, &logo_wd, &logo_ht, NULL);
  GSwdrawpict(&LogoPict, logo_x, logo_y, logo_wd, logo_ht, 0, 0, 1,
              logo_converter);

  /*--- redraw text string in text input field ---*/
  GSgetfieldrect("textfg", &text_x, &text_y, &text_wd, NULL, &text_color);
  GSsetcolor(text_color);
  GSsetfont(LARGE);
  strcpy(out_string, TextString);
  add_cursor_if_space(out_string, text_wd);
  GSwwritemsg(text_x, text_y, out_string);

  /*--- first hello is written in intro, so it is always redrawn ---*/
  write_hello();

  /*--- second hello is written only after "again" button has been pressed ---*/
  GSgetbuttonspecs("io", "again", NULL, NULL, NULL, NULL, &button_status);
  if (button_status == INVISIBLE) /*--- then it has been pressed ---*/
  {
    write_hello_again();
  }

  /*--- redraw the current help message (GIL remembers it for you) ---*/
  GSdisplayhelpmessage(CURR_MSG);

  /*--- redraw alert window if it was showing ---*/
  if (AlertShowing)
  {
    GSalert(-1, TRUE, "typedtoofar", "");
  }
  
  draw_all_objects();
  
  GSsetcurrentnamedwindow("io");
  return TRUE;
}



/*------------------------------------------------------------------------------
  remove_all_objects
    Remove drawing objects from list.
------------------------------------------------------------------------------*/
void remove_all_objects(
  void
)
{
  DrawObj *object = first_object, *next_object;
  
  while (object != NULL) {
    
    next_object = object->next_object;
    
    if (object->type == textObj)
      free(object->object.text.str);
    free(object);
    
    object = next_object;
    }
  
  first_object = NULL;
  last_object = NULL;
}


/*------------------------------------------------------------------------------
  set_color
    Called when a fill color is selected from "Fill Color" menu.
    Sets the fill color to the color listed in position itemnum on the menu.
------------------------------------------------------------------------------*/
bool set_color (
  char *item_name /*--- name of selected menu item ---*/
)
{
  static int  prev_item_num = INITIAL_ITEM;

  int item_num;  /*--- number of selected menu item
                      (top item is 0, next lower item is 1, etc.) ---*/
  item_num = GSgetmenuitemnum("col_menu", item_name);  
/*------------------------------------------------------------------------
    if selected item is different from old item, GIL turns checkmark
    for selected item on, and GStogglecheck here turns old checkmark off;
    if selected item is the same as the old item, GIL turns old
    checkmark off, and GStogglecheck here turns it back on.
  ------------------------------------------------------------------------*/
  GStogglecheck("col_menu", GSgetmenuitemname("col_menu",prev_item_num));
  FillColor = GSgetlistcolor (FILL_COLOR_LIST, item_num);
  show_fill_color(FillColor);

  /*--- show a different message if color value is unchanged ---*/
  if (prev_item_num == item_num)
    GSdisplayhelpmessage(SAME_FILL_COLOR_MSG);
  else
    GSdisplayhelpmessage(NEW_FILL_COLOR_MSG);

  prev_item_num = item_num;
  return TRUE;
}

/*------------------------------------------------------------------------------
  show_box_styles
    Called when "Box Style" button is pushed.
    Shows the menu of box styles.
------------------------------------------------------------------------------*/
bool show_box_styles (
  void
)
{
  GSdisplayhelpmessage(BOX_STYLE_MENU_MSG);
  return TRUE;
}

/*------------------------------------------------------------------------------
  show_colors
    Called when "Fill Colors" button in "root" window is pressed.
    Shows the menu of fill colors to choose from.
------------------------------------------------------------------------------*/
bool show_colors (
  void
)
{
  GSdisplayhelpmessage(FILL_COLOR_MENU_MSG);
  return TRUE;
}
/*------------------------------------------------------------------------------
  show_fill_color
    Shows fill color in a box on the "Fill Colors" button.
------------------------------------------------------------------------------*/
void show_fill_color (
  GSColor fill_color
)
{
  int        button_x, button_y, button_wd, button_ht;
  int        rect_x, rect_y, rect_wd, rect_ht;
  BigString  current_window;
  ButtonStatusType btn_status;
  
  GSgetcurrentwindowname(current_window);
  GSsetcurrentnamedwindow("root");
  GSsetcolor(fill_color);
  
  /*--- determine area of button available for fill color rectangle ---*/
  GSgetbuttonspecs("root", "color", &button_x, &button_y, &button_wd,
                   &button_ht, &btn_status);
  GSsetfont(LARGE);
  rect_x = button_x + GStextwidth(FILL_COLOR_BUTTON_LABEL) + 10;
  rect_y = button_y + 7;
  rect_wd = button_wd - GStextwidth(FILL_COLOR_BUTTON_LABEL) - 15;
  rect_ht = button_ht - 13;

  /*--- draw a solid rectangle of current fill color on "color" button ---*/
  GSwfillrect(rect_x, rect_y, rect_wd, rect_ht);

  /*--- if color _is_ "mgray", button is disabled, so make border "mgray" ---*/
  if (btn_status == DISABLED)
    GSsetcolor(GSconvertcolor("white"));

  /*--- draw a border around the solid rectangle ---*/
  GSwdrawrect(rect_x - 1, rect_y - 1, rect_wd + 2, rect_ht + 2);
  GSsetcurrentnamedwindow(current_window);
}


/*------------------------------------------------------------------------------
  show_funcs_menu
    Called when "Menu" button is pushed.
    Shows the menu of user functions to choose from.
------------------------------------------------------------------------------*/
bool show_funcs_menu (
  void
)
{
  GSdisplayhelpmessage(MENU_MENU_MSG);
  return TRUE;
}


/*------------------------------------------------------------------------------
  toggle_border
    Called when "Draw Black Border" item on "Box Styles" menu is selected.
    Toggles the DrawBorder parameter.
------------------------------------------------------------------------------*/
bool toggle_border (
  char *item_name  /*--- name of selected menu item ---*/
)
{
  DrawBorder = !DrawBorder;
  if (DrawBorder)
  {
    GSdisplayhelpmessage(BOX_STYLE_CHECK_ON_MSG);
  }
  else
  {
    GSdisplayhelpmessage(BOX_STYLE_CHECK_OFF_MSG);
  }
  return TRUE;
}

/*------------------------------------------------------------------------------
  toggle_fill
    Called when "Fill With Color" item on "Box Style" menu is selected.
    Toggles the FillBorder parameter.
------------------------------------------------------------------------------*/
bool toggle_fill (char *item_name)
{
  FillBorder = !FillBorder;
  if (FillBorder)
  {
    GSdisplayhelpmessage(BOX_STYLE_CHECK_ON_MSG);
  }
  else
  {
    GSdisplayhelpmessage(BOX_STYLE_CHECK_OFF_MSG);
  }
  return TRUE;
}

/*------------------------------------------------------------------------------
  write_hello
    write "Hello World." in field "hello1" in "io" window.
------------------------------------------------------------------------------*/
void write_hello (
  void
)
{
  int      hello1_x, hello1_y;
  GSColor  hello1_color;

  /*--- write "Hello World" ---*/
  GSgetfieldrect("hello1", &hello1_x, &hello1_y, NULL, NULL, &hello1_color);
  GSsetcolor(hello1_color);
  GSsetfont(LARGE);
  GSsetcurrentnamedwindow("io");
  GSwwritemsg(hello1_x, hello1_y, "Hello World.");
}
/*------------------------------------------------------------------------------
  write_hello_again
    write "Hello World, again." in field "hello2" in "io" window.
------------------------------------------------------------------------------*/
void write_hello_again (
  void
)
{
  int      hello2_x, hello2_y;
  GSColor  hello2_color;

  /*--- get coordinates and color for message text ---*/
  GSgetfieldrect("hello2", &hello2_x, &hello2_y, NULL, NULL, &hello2_color);

  /*--- set text attributes and write message ---*/
  GSsetcolor(hello2_color);
  GSsetfont(LARGE);
  GSwwritemsg(hello2_x, hello2_y, "Hello World, again.");
}

/*------------------------------------------------------------------------------
  FuncNames array
    function name -> function pointer mapping
------------------------------------------------------------------------------*/
IntrFunction FuncNames[] = {
/* string in .inf file    function pointer
   -------------------    ------------------------ */
  { "agn_func",           do_again              },
  { "box_menu",           show_box_styles       },
  { "clear_draw_func",    clear_draw            },
  { "clear_text_func",    clear_text            },
  { "col_menu",           show_colors           },
  { "done_func",          alert_done            },
  { "draw_func",          (bool(*)())do_draw },
  { "funcs_menu",         show_funcs_menu       },
  { "intro",              do_intro              },
  { "kbd_hndlr",          (bool(*)())do_text },
  { "quit_prog_func",     do_quit               },
  { "redraw_func",        redraw_all            },
  { "set_color_func",     (bool(*)())set_color             },
  { "toggle_border_func", (bool(*)())toggle_border         },
  { "toggle_fill_func",   (bool(*)())toggle_fill           },
};

int NumFuncs = GSnumifuncs(FuncNames);

/*------------------------------------------------------------------------------
  Main routine: Do initializations and start interface handler.
------------------------------------------------------------------------------*/
int GS_MAIN (
	GS_MAIN_ARGS_F /*--- formal arguments to main() ---*/
)
{
  first_object = last_object = NULL;
  
  /*--- initialize output.log ---*/
/*  if (!GSinitlog ("wt"))*/
  if (!GSinitlog ("w"))
  {
    printf("Unable to initialize output.log\n");
    GSdelay(TIME_TO_READ_MSG);
    exit(-1);
  }

  init_pict(&LogoPict, GlobeFile);

  /*--- initialize giltest interface ---*/
  GSinterfaceinit("GeoSim Interface Library (GIL) Test", /*--- progrm name ---*/
                  "giltest.inf", /*--- interface file name ---*/
                  BSTORE,        /*--- use backing store ---*/
                  "w",           /*--- error.log mode ---*/
                  G640x480x16,   /*--- 16-color graphics mode ---*/
                  GS_MAIN_ARGS_A); /*--- arguments to GS_MAIN ---*/

  init_colors();

  /*--- activate keyboard input function (do_text) ---*/
  
  GScharfunc("kbd_hndlr");

  /*--- start event loop ---*/
  GSinterface();
  return(0);
}