#ifndef INTRFACE_INC
#define INTRFACE_INC

#ifdef __cplusplus
extern "C" {
#endif
/**********************************************************************
Include file shared by modules accessing the GeoSim interface
interpreter.
**********************************************************************/

#include "gssys.h"
#include "gsincl.h"
#include <stdio.h>

#define GSenabledragarea(wn,dn)    GSenablebutton((wn),(dn),NO_DRAW)
#define GSdisabledragarea(wn,dn)   GSdisablebutton((wn),(dn),NO_DRAW)
#define GSinvisibledragarea(wn,dn,draw)   GSinvisiblebutton((wn),(dn),(draw))
#define GSvisibledragarea(wn,dn,draw)   GSvisiblebutton((wn),(dn),(draw))
#define GSnumifuncs(farray)        sizeof((farray))/sizeof(IntrFunction)

#define boolean bool

typedef	enum {JUST_LEFT, JUST_RIGHT, JUST_CENTER} GSjustType;

enum
{
  NO_DRAW = 0,
  DRAW    = 1
};

enum
{
  PREV_LIST_ITEM = -2,
  CURR_LIST_ITEM = -1
};

enum
{
  NO_BSTORE = 0,
  BSTORE    = 1
};

enum
{
  NO_POPUP = 0,
  POPUP    = 1
};

enum
{
  THIN  = 1,
  THICK = 3
};

enum
{
  CURR_MSG = -1,
  PREV_MSG = -2
};

typedef enum
{
  DRAG_INIT     = 1,
  DRAG_PROCESS  = 2,
  DRAG_FINAL    = 3,
  MOUSE_INSIDE  = 4,
  MOUSE_CLICK   = 5,
  MOUSE_OUTSIDE = 6,
  KEY_PRESS     = 7
} DragStatusType;

typedef enum
{
  ENABLED       =  1,  /*--- Users can see and access element ---*/
  DISABLED      =  0,  /*--- Users can see, but cannot access element ---*/
  INVISIBLE     = -1,  /*--- Users cannot see element ---*/
  TMP_DISABLED  = -2   /*--- Disabled while popup window is up ---*/
} ButtonStatusType, MenuItemStatusType, DragAreaStatusType;

typedef enum
{
  SMALL = 10,
  LARGE = 15
} FontSizeType;

/*--- Graphics file formats supported for screen dumps ---*/
typedef enum
{
  NO_FILE = 0,  /*--- No File Format.  I.e., don't dump screen to a file ---*/
  BMP_16,       /*--- 16 color Windows 3.1 bitmap format ---*/
  PS_GRAY,      /*--- PostScript Level 2.0 grayscale format ---*/
  PS_COLOR,     /*--- PostScript Level 2.0 color format ---*/
  GIF_16        /*--- 16 color GIF87a format ---*/
} GraphicsFileType;

/*--- Key encodings ---*/
enum
{
  KeyRubout =  8,
  KeyTab    =  9,
  KeyEnter  = 13,
  KeyEscape = 27,
  KeyUp     = 0x1E,
  KeyLeft   = 0x1C,
  KeyDown   = 0x1F,
  KeyRight  = 0x1D,
  KeyF1 = 0x13
};

/*--- GSdohelpscreen page commands ---*/
typedef enum
{
  FirstPage,    /*--- when help window is first opened ---*/
  NextPage,     /*--- go forward one page of help file ---*/
  PrevPage,     /*--- go backward one page of help file ---*/
  CurrPage      /*--- redraw current page of help file ---*/
} HelpScreenPage;

typedef unsigned char GSColor;

typedef enum disptype
{
  Highlight,
  Unhighlight,
  New
} DispType;

typedef struct
{
  char *name;
  boolean (*funcptr)();
} IntrFunction;

/*--- Runlength encoded picture structure.  This is accessable
      to the user so that they can see the size of the image. ---*/
typedef struct
{
  unsigned long tag;      /* Identifier tag for image type */
  int runsize;            /* Length of runlength code in bytes (depends on tag)*/
  unsigned short rows;    /* Number of rows and cols for this image */
  unsigned short cols;    /* Number of rows and cols for this image */
  unsigned short *rowvec; /* Offsets for rows within codes */
  unsigned char *codes;   /* The runlength codes */
} GSPicture;

/*--- Types of control groups (i.e. Windows, or Menus) ---*/
typedef enum controlgrouptype
{
  NONE,
  WINDOW,
  MENU
} ControlGroupType;

/*--- Information for a control (i.e. button or menu item) ---*/
typedef struct
{
  ControlGroupType type;   /*--- NONE, WINDOW, MENU ---*/
  BigString gname;         /*--- window or menu group name ---*/
  BigString time;
  union
  {
    BigString name; /*--- name, if in window ---*/
    int number;     /*--- physical position number, if in menu ---*/
  } control;
} ControlType;

/*--- Graphics modes. GWIDTHxHEIGHTxCOLORS ---*/
typedef enum graphicsmodetype
{
  G640x480x256 = 256, 
  G640x480x16 = 16
} GraphicsModeType;


typedef int GSlistHandle;
typedef int GSsclistHandle;

char *str_dup(char *);

/*-------------------------Prototypes for intrpub.c---------------------------*/

void    GSactivatefunction	(char *);
boolean GSbigendian		(void);
void    GSclearlastcontrol	(void);
void    GScreatenamedpopup	(char *);
GSColor GSconvertcolor		(char *);
void    GSdeactivatefunction	(char *);
void    GSdisablebutton		(char *, char *, int);
void    GSdisablemenuitem	(char *, char *);
void    GSdrawnamedwindow	(char *, int, int);
void    GSdrawxorbox		(int, int, int, int);
void    GSdrawinvertbox		(int, int, int, int);
void	GSdrawinvertline	(int, int, int, int);
void	GSdrawxorline		(int, int, int, int);
void    GSenablebutton		(char *, char *, int);
void    GSenablemenuitem	(char *, char *);
long    GSfilesize		(FILE *);
void    GSgetbuttonspecs	(char *, char *, int *, int *, int *, int *,  
				 ButtonStatusType *);
void	GSgetdragareaspecs	(char *, char *, int *, int *, int *, int *,
				 DragAreaStatusType *, GSColor *);
void	GSsetdragareaspecs	(char *, char *, int *, int *, int *, int *,
				 GSColor *);
void	GSsetbuttonspecs	(char *, char *, int *, int *, int *, int *,
                                 GSColor *);
void	GSsetbuttontext		(char *, char *, char *);
void	GSgetwindowspecs	(char *, int *, int *, int *, int *, int *);
char    *GSgetcurrentwindowname	(char *);
void    GSgetcursorxy		(int *, int *);
void	GSgetlastcontrol	(ControlType *);
GSColor GSgetlistcolor          (int, int);
void    GSgetfieldrect          (char *, int *, int *, int *, int *, GSColor *);
void	GSsetfieldrect		(char *, int *, int *, int *, int *, GSColor *);
char    *GSgetmsg                (char *);
void    GSinterface              (void);
void    GSinterfaceinit          (BigString, BigString, int, char *,
                                  GraphicsModeType, GS_MAIN_ARGS_F);
void    GSinvisiblebutton        (char *, char *, int);
void    GSinvisiblemenuitem	(char *, char *);
void    GSloadwindows            (FILE *);  
long    GSmemavail               (void);
void    GSmouseinit              (void);
FILE    *GSopen                  (char *, char *);
void    GSquit                   (char *, ...);
void    GSredrawactivewindows    (void);
void    GSredrawnamedwindow      (char *);
void    GSremovenamedwindow      (char *);
void    GSremovetopwindow        (void);
void    GSsavewindows            (FILE *);
void    GSsetactivenamedwindow   (char *);
void    GSsetcurrentnamedwindow  (char *);
void    GSsetdeactivenamedwindow (char *);
void    GSsetlinesize            (int);
void    GSsetmenuorigin          (char *, int *, int *);
void    GSsetwindowspecs	(char *, int *, int *, int *, int *);
void    GStogglecheck            (char *, char *);
int		GSgetmenuitemnum		(char *, char *);
char	*GSgetmenuitemname		(char *, int);
void    GSvisiblebutton          (char *, char *, int);
void    GSvisiblemenuitem	(char *, char *);
void    GSwclearrect             (int, int, int, int);
void    GSwdrawellipse           (int, int, int, int);
void    GSwdrawline              (int, int, int, int);
void    GSwdrawpoint             (int, int);
void    GSwdrawrect              (int, int, int, int);
void    GSwfastrow               (int, int, GSColor[], int);
void    GSwfillellipse           (int, int, int, int);
void    GSwfillpoly              (int, int *);
void    GSwfillrect              (int, int, int, int);
GSColor GSwgetpoint              (int, int);
void    GSwsetmouseposition      (int, int);
void    GSwwritecommanum         (int, int, double);
void    GSwwriteint              (int, int, int, long);
void    GSwwritemsg              (int, int, char *);
void	GSwwriteclippedmsg	 (int, int, int, int, int, int, char *);
void    GSwwritenamedmsg         (int, int, char *);
void    GSwwritereal             (int, int, int, int, double);
void	GSframefield		 (char *, char *, GSColor, GSColor);
void	GSframedragarea		 (char *, char *, GSColor, GSColor);



/*-------------------- Video (system dependent) routines ---------------------*/

GSColor GSgetcolor               (void);
void    GSsetcolor               (GSColor);
void    GSsetfont                (FontSizeType);
int     GStextheight             (char *);
int     GStextwidth              (char *);


/*-------------------------- Prototypes: picutils.c --------------------------*/

void    GSgetpixel              (GSPicture *, int, int, unsigned short *, int);
void	GSgetfilepixel		(FILE *, int, int, unsigned short *, int);
boolean GSinitpic                (FILE *, GSPicture *, void *);
void    GSwdrawfilepict          (FILE *, int, int, int, int, int, int,
                                 GSColor(*)(unsigned short));/*int,int,int));*/
void    GSwdrawpict		(GSPicture *, int, int, int, int, int, int, int,
                                 GSColor(*)(unsigned short,int,int,int));
void	GSgetfilepictspecs	(FILE *, unsigned short *, unsigned short *);


/*---------------------------- Prototypes: list.c ----------------------------*/

void		GSdrawlistitem	(GSlistHandle, DispType, int);
void		GShandlelist	(GSlistHandle, int, int, DragStatusType);
void		GSlistpage	(GSlistHandle, int);
GSlistHandle	GSlistsetup	(char *,
                                 char *,
                                 int,
                                 char *,
                                 int,
                                 int,
                                 GSjustType,
                                 int (*)(GSlistHandle, int),
                                 char *(*)(int),
                                 void (*)(GSlistHandle, int),
                                 void (*)(int),
                                 void (*)(void),
                                 void (*)(int));
GSlistHandle	GSdismantlelist	(GSlistHandle);
void GSlistfromitem(GSlistHandle list, int firstItem);


/*--------------------------- Prototypes: gshelp.c ---------------------------*/

void    GSalert                  (int helpmsg, boolean, char *, ...);
void    GSdisplayhelpmessage     (int);
void    GSdisplayversion         (BigString);
boolean GSdohelpscreens          (BigString, HelpScreenPage, boolean, int);
int     GSgetcurrmsg             (void);


/*--------------- prototypes in ??misc.c - platform dependent. ---------------*/

void    GSbell                   (void);
void    GSclick                  (void);
boolean GShasFPU                 (void);
void	GSdelay			 (unsigned long);

extern short Etester;
#ifndef GSbigendian
#define GSbigendian()  ( *((char *) &Etester) )
#endif
/*-------------------------- Prototypes: gsread.c ----------------------------*/

boolean GSgenreader      (FILE *, boolean(*)(BigString, ReaderStatusType, void *), char *, void *);
boolean GSrdboolean              (void);
boolean GSrdformat               (int *, FormatList);
long    GSrdinteger              (void);
boolean GSrdphrase               (char *, int);
double  GSrdreal                 (void);
boolean GSrdstringeol            (char *, int);
boolean GSrdword                 (char *, int);


/*---------------------------- Prototypes: log.c -----------------------------*/

boolean GSinitlog                (char *);
void    GSolog                   (char *, ...);
void    GSelog                   (int, char *, ...);


/*-------------------------- Prototypes: eventhnd.c --------------------------*/

void    GScharfunc               (char *);
void    GSclearcharfunc          (void);


/*-------------------------- Prototypes: scrndump.c --------------------------*/

boolean GSdumpscreen             (char *, GraphicsFileType);


/*---------------------------- Prototypes: login.c ---------------------------*/

void    GSdologinscreen          (char *, int, boolean(*)(void), GSColor,
                                  char *);


/*-------------------------- Prototypes: eventhnd.c ---------------------------*/

void    GSdoeventsandreturn      (void);

/*-------------------------- Prototypes: thumb.c ---------------------------*/
void			GSdismantlesclist(GSsclistHandle);
GSsclistHandle 	GSscrollbarsetup(
						char *wname,
						char *daname,
						GSColor buttonCol, 
						GSColor upLeftCol, 
						GSColor botRightCol,
						int numListItems, int numVisibleItems,
						int cols,
						int curPos,
						void (*sclistDraggedFunc)(int),
						char *upbtn, char *dwnbtn
						);
void			GSdrawsclist(GSsclistHandle);
void 			GSsetsclistnumitems(GSsclistHandle sclist,
									int numItems, boolean redraw);
void 			GSsetsclistpos(GSsclistHandle sclist, 
									int newPos, boolean redraw);
int				GSgetsclistpos(GSsclistHandle);
int				GSgetmaxpos(GSsclistHandle);
void 			GShandlesclist(GSsclistHandle, int, int, DragStatusType);
void			GShandleupbutton(GSsclistHandle);
void			GShandledownbutton(GSsclistHandle);
void			GSdisablesclistbuttons(GSsclistHandle);
void			GSenablesclistbuttons(GSsclistHandle);
void GSsclistsetup(
	char	*wname,			/*--- name of window holding list and sclist ---*/
	char	*daname,		/*--- name of drag-area holding list ---*/
	int	cols,			/*--- Number of columns on page ---*/
	char	*labelfield,		/*--- Field descriptor for a label ---*/
	int	colgap,			/*--- Gap (in pixels) between columns ---*/
	int	rowgap,			/*--- Gap (in pixels) between rows ---*/
	GSjustType just,		/*-- justification of labels --*/	
	char	*(*label_func)(int),	/*--- returns requested label ---*/
	void	(*settext_func)(GSlistHandle, int), /* sets text attributes for items */
	void	(*in_func)(int),	/*--- response when cursor in list ---*/
	void	(*out_func)(void),	/*--- response when cursor leaves list ---*/
	void	(*click_func)(int),	/*--- response when user clicks ---*/
	char *thDaname,
	GSColor buttonCol, GSColor topLeftCol, GSColor botRightCol,
	int numListItems,
	int curPos,
	void (*sclistDraggedFunc)(int),
	char *upbtn, char *dwnbtn,
	GSlistHandle *newList,
	GSsclistHandle *newThumb
	);

/*-----  Global Variables (defined in $(GEOSIM)/utils/intrface.c) ------*/

extern boolean GSDebug;
extern boolean GSpermit_log;
extern int GraphicsMode;            /* used on all platforms now */ 

#ifdef __cplusplus
} /* end extern "C" */
#endif

#endif /* #ifndef INTRFACE_INC */
