#include "geosim.h"

#ifndef INTRTYPE_INC
#define INTRTYPE_INC
#define MAXBSTORES 5
#define MAXALIASES   50
#define COLORNAMELEN  9
#define DARKBUTTONBORDER    1   /* palette index for 3D button borders;
                                   used in intrface.c  */
#define LIGHTBUTTONBORDER   3   /* palette index for 3D button borders;
                                   used in intrface.c  */
#define MAXWINDOW           100 /* Maximum number of windows. */
#define MAXPOPUP            25  /* Maximum number of popup menus, and max
                                   number of menu items on a menu */
#define TSIZE               15  /* Text height in pixels */
#define POPUPOFFSET         5   /* Offset from edge of menu box for popup
                                   item label */
#define CHECKWID            10  /* Width of a check mark for checklist */

#define NOCOLOR             16

/*
No longer using this.  CLK_TCK is a macro which
should be in the platform time.h header - use CLK_TCK
instead.  JBR

#define GS_CLOCKS_PER_SEC      1000000
*/

#ifndef SEEK_SET 
#define SEEK_SET            0
#endif

#ifndef SEEK_CUR 
#define SEEK_CUR            1
#endif

#ifndef SEEK_END 
#define SEEK_END            2
#endif
#define FGCOLOR             99
#define BGCOLOR             98
#define HIGHLIGHT           97

#define PALBARWIDTH256  470
#define PALBARHEIGHT256 160
#define PALBARWIDTH16   250
#define PALBARHEIGHT16  60

#define MAXOVERRLAPBUTTONS  10  /* Maximum number of overlapping drag rect
                                   buttons */
#define PAL_SIZE 256
#define NUM_STD_COLORS 2 /* Black, white */
#define NUM_DFLT_COLORS 6 /* dgray, mgray, lgray, bluegray, lblue, peach */
#define UP_LEFT             0
/*#define CENTER              1*/


/* These definitions should become variables if we support different resolutions, etc.*/
#define SCREENWIDTH        640
#define SCREENHEIGHT       480
#define BITSPERPIXEL       4
#define BM                 19778    /* Is this defined elsewhere in a Windows header? -Bo*/

/* Refugees from intrface.c:
	these were moved so that the parser could see them.  JBR
*/
#define INFNAMELEN    31 /* max characters for window/button name */
#define ARROWMARGIN  4

/* drawable graphics objects supported */
#define GR_LINE 1
#define GR_WLINE 2
#define GR_RECT 3
#define GR_WRECT 4
/*#define GR_BITMAP 5  Is this really needed?  It's never implemented.  JBR */
#define GR_LABEL 6
#define GR_LABEL_SMALL 7


/* Graphics type definitions. */
/* WARNING: Any type that will be used with "getname" MUST have the
   pointer to the string to be searched for as the first item in the
   structure definition. */

/* Event types */
enum { EvExpose, EvMouseDown, EvMouseUp, EvKeyDown, EvNone };

typedef struct
{
  char     type, key;
  short    mousex, mousey;
} GSEvent;

/* Rectangle styles */
enum { HOLLOW, FILL };
typedef enum
{
  UP_ARROW_SYM = '^',
  DOWN_ARROW_SYM = 'v',
  LEFT_ARROW_SYM = '<',
  RIGHT_ARROW_SYM = '>',
  NO_ARROW_SYM = '\0'
} ArrowSymType;

/* Descriptor for a field */
typedef struct FieldType
{
  char *name;             /* Field Name (as known to caller) */
  int x, y, wd, ht;       /* Bounding box */
  GSColor color;            /* Color index */
} Field;

typedef struct
{
  char *name;
  char *text;
} Message;

typedef struct
{
  char *name;
  int dtime;   /*  delta time to call  */
  int itime;   /*  last time (initial) called  */
  boolean mouse_off;            /* TRUE iff the mouse cursor should be hidden */
  boolean active;
  boolean (*func)(int, int, int, ...);
} Function;

/* Description of the current configuration (state) of the interface */
typedef struct
{
  int		level;		/* Stacking height of current window */
  int		speedup;	/* Speedup rate factor for pressed buttons */
  int		cw;		/* Current active window */
  int		cb;		/* Currently "hit" button */
  GSColor	colorlist[5][5]; /* A list of user defined colors.  This needs to
                                   be fixed for arbitrary length list. CAS */
  GSColor	highlight;	/* Highlight color */
  int		numwindows;
  int		numpopups;
  int		activepopupwindow;
  int		buttonborder;
  int		buttonTLcolor;
  int		buttonBRcolor;
  int		diagonals;
  int		windowborder;
  int		numfields;	/* Number of fields */
  Field		*fields;	/* Pointer to array of fields */
  int		numpics;	/* Number of pictures */
  int		nummsgs;	/* Number of messages */
  Message	*msgs;		/* Pointer to array of messages */
  int		numfuncs;	/* Number of activatable functions */
  Function	*funcs;		/* Points to array of user-activated functions */
  boolean	winbstore;      /* TRUE if application wants bstore for windows. */
  struct
  {
    boolean	inuse;		/* Occupied? */
    char	*bstore1;	/* Backing store space */
    char	*bstore2;	/* Backing store space */
    char	*bstore3;	/* Backing store space */
    int		bstoreht;	/* Height of image in backing store */
    unsigned	bsize;		/* size of backing store */
  } backingstore[MAXBSTORES+1];
  ControlType	lastcontrol;	/* last control pressed (button or menu item) */
  boolean	(*keyfunc)(int, int, char); /* Key handler function to call
                                               when a key is pressed */
  void		(*redrawfunc)(void);	/* Redraw all function to call on
                                           Expose Events */
} Iconfig;

typedef enum
{
  LEFT = 'L',
  CENTER = 'C',
  RIGHT = 'R'
} BtnLabPosType;

/* Descriptor for a button */
typedef struct
{
  char            *name;               /* Name for button */
  char            *text;               /* Text displayed in button */
  ArrowSymType     arrow;
  int              x, y, wd, ht;       /* Bounding box */
  int              popupmenunum;       /* What popupmenu this part of */
  FontSizeType     tsize;              /* Text size */
  ButtonStatusType active;          
  GSColor          efgcolor;           /* Enabled text color */
  GSColor          dfgcolor;           /* Disabled text color */
  GSColor          bgcolor;            /* Background color */
  boolean          (*func)(int, int, DragStatusType, ...); /* What to do when button is
                                                   pressed */
  boolean          border;             /* Border present? */
  boolean          repeatable;         /* TRUE iff the callback funct should be
                                          executed as long as button is held
                                          down */
  boolean          depressed;          /* TRUE iff the button is currently
                                          drawn depressed */
  boolean          drag;               /* TRUE iff this is a dragable */
  BtnLabPosType    labpos;            /* LEFT, CENTER, RIGHT */
} GSButton;


/* Descriptor for a graphic object */
typedef struct GraphicType
{
  int x, y, wd, ht;
  char color;
  char shadowcolor;
  char type;
  char attrib;
  char *text;
} Graphic;


/* Descriptor for a window */
typedef struct
{
  char     *name;         /* Name (as known to caller) */
  int       active;       /* Stacking height of this window */
  int       bstore_to_use; /* Which backing store can this
				window use, if any? */
  boolean   bstore;       /* Is backing store currently saved for this
                             window? */
  boolean   popup;        /* Is this a popup window? */
  int       x, y, wd, ht; /* Bounding box */
  GSColor   fgcolor;      /* Forground color */
  GSColor   bgcolor;      /* Background color */
  GSColor   bordercolor;  /* Border color */
  int       bordersize;   /* Width of border */
  int       numbuttons;   /* Number of buttons on window*/
  GSButton *wb;           /* Pointer to array of buttons on this window */
  int       numgraphics;  /* Number of graphics on window */
  Graphic  *graphics;     /* Pointer to array of graphic items on window */
} GSWindow;

/* Descriptor for a menu item */
typedef struct
{
  char              *name;         /* Name (as known to caller) */
  char              *text;
  MenuItemStatusType active;   /* -1 = invisible, 0 = disabled, 1 = enabled  */
  boolean            checked;  /* TRUE iff this item should have a check mark */
  boolean            (*func)(char *);
} Menuitem;


/* Descriptor for a popup menu */
typedef struct
{
  char *name;
  boolean active;     /* TRUE iff active */
  boolean checklist;  /* TRUE iff this is a checklist menu */
  GSColor enablefg, disablefg, bg, bcolor;
  int x,y;
  int wd, ht;
  int numitems;
  Menuitem *item;
  boolean (*func)(void);
} Popupmenu;

/* Descriptor for a scroll list */
typedef struct
{
  int sclistposition;
  int sclistrange;
  int sclistht;
  int numpages;
  int maxpage;
  int *currpage;
} GSScrollList;

/*--------------------------------------------------------------------------*/

#ifndef BIG_ENDIAN_DEF
extern short Etester;
#define BIG_ENDIAN_DEF
#endif

/*--------------------------------------------------------------------------*/

/*  prototype for support functions in intrface.lib  */
void    writemsg(int, int, char *);
void    drawbutton(int, int, boolean);
void    erasebutton(int, int);
void    drawwindow(int);
void    drawwindowborders (int, int);
void    removewindow(int);
GSColor convertcolor(char *[], char *[], char *);
void    createpopupmenu(int);
void    init_default_palette(char *[], char []);
void    readfilepass1(BigString);
void    readinterfacefile(BigString);
void    assignpopups(void);
boolean getchars(char *, int, int);
void    highlight(int, int);
void    processbuttoninput(int, int, boolean);
void    processapplicationkey(int, int, char);
void    processkeyboardinput(void);
void    init_all_windows(void);
void    GSPinit_system(char *, GS_MAIN_ARGS_F);
void    GSPflush(void);
int     getshiftstatus (void);
void    processmouseinside (void);
void    init_palette (void);
void    writepalettefile (char *);
void    getbuttoninfo (char *, char *, int *, int *, int *, int *, int *);
boolean init_error_log (char *);
int     pickitem (int, int);
int     pickitemslot (int, int);
int     getname(void *, int, int, char *, char *);
void    toggle_check(int, int, GSColor);
void    doallbuttons (int, int, int, boolean);
void    GSPstart_timer(void);
long    GSPread_timer(void);
void    GSPdraw_depressed_button(int, int);
void    GSPdraw_released_button( int, int);
void    createpopupmenu(int);
void    removepopup(void);
void    setlastcontrol(ControlGroupType, char *, char *, int);
void    process_arrow(int);
void    display_popup_label(int, int, int);
void    GSPsetcurrentwindow(int winnum);
void    disable_buttons_for_popup(int);
void    enable_buttons_after_popup(int);
int     GSPgetlinesize(void);
void	drawshadedframe(int, int, int, int, GSColor, GSColor);


/* ???video.c  Platform dependent video routines*/

/* these are macros in DOS - JBR */

#ifndef GSPinvertmode
void GSPinvertmode(int);
#endif
#ifndef GSPxormode
void GSPxormode(int);
#endif

void GSPfastrow(int, int, GSColor *, int);
void setpalettebar (void);
void writergbpalette(char *);
unsigned far GSPimagesize( int, int, int, int);
FontSizeType GSPgetfont(void);       /* returns the current font: SMALL or LARGE */
GSColor *GSPget_row(int row_n, GSColor *row_b);
void GSPset_focus(void);
void GSPinit_palette( char pal[3*256],    /* pal of r g b colors */
		      int n);              /* number of colors */
void GSPset_clip_rect(int x, int y, int wd, int ht);
void GSPframe_rect(int x, int y, int wd, int ht);
void GSPfill_rect(int x, int y, int wd, int ht);
void GSPline(int x1, int y1, int x2, int y2);
void GSPdraw_string(int x, int y, char *msg);
void GSPfillpoly(
  int numpoints,   /* number of vertices in polygon */
  int *polypoints); /* array of 2*numpoints ints that are the x-y coords of the vertices */
boolean GSPinit_bstore(int, int, int);
boolean GSPto_bstore( int, int, int, int, int);
boolean GSPfrom_bstore( int, int, int);
void GSPdraw_pixel(int x, int y);
GSColor GSPget_pixel(int x, int y);
void GSPframe_ellipse(int, int, int, int);
void GSPfill_ellipse(int, int, int, int);
void closegraph();

/* These two GSPsuspend/restoregraphics should removed ASAP (01/19/94) -Bo */
void GSPsuspendgraphics(void);
void GSPrestoregraphics(void);


/* ???misc.c   Platform specific routines */


/* mouse prototypes.  - platform dependent */
int      mouse_present(void);
void     reset_mouse(void);
void     set_mouse_position(int, int);
void     get_mouse_position(short *, short *);
int      get_mouse_press(int, int *, int *, int *);
int      get_mouse_release(int, int *, int *, int *);
void     set_mouse_borders(int, int, int, int );

/*-------------------------Prototypes for intrface.c---------------------------*/

boolean button_text_specials(char, void *);
void resolve_specials(char *, boolean (*)(char,void*), void*);
boolean (*getfunctionptr(char *fname))();


/* dosevent.c */
void     GSPinit_events(void);
void     GSPclose_events(void);
void     GSPclear_events(void);
int      GSPnext_event(GSEvent far *);

/* eventhnd.c */
void GSPdo_events_forever(void);
boolean GSPinside_button( int, int, int, int);
void GSPkeyevent(int, int, char);
void GSPdoactivefuncts(int, int);
void GSPprocessmouseinside(int, int);
boolean GSPgetbutton (int, int, int *, int *);
int GSPget_popup_item(int, int);
boolean GSPhandle_popups(int, int);

/* scrndump.c */
boolean GSPscreen_to_bmp( FILE * );
boolean GSPscreen_to_ps( FILE *, boolean );
boolean GSPscreen_to_gif(void);

/* selword.c */
int GSiselword (char *, char **);
int GSselword (char *, char **);

#endif /* INTRTYPE_INC */
