/* mouselib.c
**
** Library of mouse routines.
** by Colin Klipsch (based on code by Mark Lattanzi)
** July, 1992
*/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

/*---------------------------------------------------------------------*/

int    mouse_present(void)
    {
    /* Detect whether a mouse is present; return TRUE if so, FALSE
    ** otherwise.
    */

    void    (interrupt far *int_handler)();
    long             vec;
    unsigned char    ch;

    int_handler = getvect(0x33);
    ch = *((unsigned char far *) int_handler);
    vec = (long) int_handler;
    return (vec != 0  &&  ch != 0xCF);
    }


void    get_mouse_position(short far *x, short far *y)
    {
    /* Read the mouse's position into *x and *y. */

    asm mov    ax, 3
    asm int    33h
    asm les    bx, x
    asm mov    es:[bx], cx
    asm les    bx, y
    asm mov    es:[bx], dx
    }

void   set_mouse_position(int x, int y)
    {
    /* Put the mouse at the given position. */

    asm mov    cx, x
    asm mov    dx, y
    asm mov    ax, 4
    asm int    33h
    }


int    get_mouse_press(int btn, int far *x, int far *y, int far *stat)
    {
    /* Return the number of mouse button presses that have occured since
    ** the last time this function was called.  The button of interest
    ** is encoded in _button_: 0 for the left button, 1 for the right, 2 for
    ** the middle.  Also puts the position of the mouse at the last press
    ** into *x and *y, and puts the current state of all the buttons into
    ** *stat (bit 0 represents the left button, bit 1 the right, bit 2
    ** the middle).
    */

    asm mov    bx, btn
    asm mov    ax, 5
    asm int    33h
    asm push   bx
    asm les    bx, x
    asm mov    es:[bx], cx
    asm les    bx, y
    asm mov    es:[bx], dx
    asm les    bx, stat
    asm mov    es:[bx], ax
    asm pop    ax

    return _AX;
    }


int    get_mouse_release(int btn, int far *x, int far *y, int far *stat)
    {
    /* Very similar to get_mouse_press, except here we deal with button
    ** releases.  See the comments for get_mouse_press.
    */

    asm mov    bx, btn
    asm mov    ax, 6
    asm int    33h
    asm push   bx
    asm les    bx, x
    asm mov    es:[bx], cx
    asm les    bx, y
    asm mov    es:[bx], dx
    asm les    bx, stat
    asm mov    es:[bx], ax
    asm pop    ax

    return _AX;
    }

void   set_mouse_borders(int left, int top, int right, int bottom)
    {
    /* Confine the mouse to the given rectangle. */

    asm mov    cx, left
    asm mov    dx, right
    asm mov    ax, 7
    asm int    33h
    asm mov    cx, top
    asm mov    dx, bottom
    asm mov    ax, 8
    asm int    33h
    }

