#define  BSTORE_CHECK '\0'

/*================================ image.c ================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloc.h>
#include <dos.h>
#include <graphics.h>
#include <ctype.h>
#include "geosim.h"
#include "gspriv.h"

#define VIDEO 0x10

static int Offx=0, Offy=0;     /* Upper left corner of current clip rectangle */

static union REGS regs;
static struct SREGS sregs;

extern int Cursx;
extern int Cursy;
extern char Interface_Palette[];
extern char *bm1;  /*  backing store for popup palette  - intrpub.c */
extern char *bm2;  /*  backing store for popup palette  */
extern Iconfig InfState;        /* Current state of the interface */

extern char *bm1;  /*  backing store for popup palette editor */
extern char *bm2;  /*  backing store for popup palette editor */
extern GSWindow *Win;
static int CurrentFont;


void GSflushevents(void) {}
void GSPflush(void) {}

/*--------------------------------------------------------------------------*/
/*****************************************************************************
Return non-negative code for the next keypress, or -1 if there is no
keypress waiting.
*****************************************************************************/
static int DOSget_key(void)
{
  int    key;

  if (!kbhit())  return -1;

  if (!(key = getch()))  key = 256+getch();

  switch (key)
  {
    case 328: key = KeyUp;  break;
    case 336: key = KeyDown;  break;
    case 331: key = KeyLeft;  break;
    case 333: key = KeyRight;  break;
    case 339: key = KeyRubout;  break;
  }

  return (key>=0 && key<=127)? key : -1;
}

/***********************************************************************
Get the string typed by user.  Return TRUE iff user ended with ENTER.
GLOBALS: None.
***********************************************************************/
static boolean DOSgetchars(
  char *str,    /* space for the string. */
  int x, int y) /* location for string */
{
  int num = 0;
  int savecolor;

  savecolor = GSgetcolor();
  while ((*str = DOSget_key()) != KeyEnter)
    if (isalnum(*str) || *str == '.' || *str == '-')
    {
      str++ ;
	  *str = '\0';
      GSPdraw_string(x, y, str-1);
      x += textwidth(str-1) ;
      num++;
    }
    else if (*str == KeyRubout && num) /* BACKSPACE */
    {
      *str = '\0';
      GSsetcolor(0);
      x -= textwidth(str-1);
      GSPdraw_string(x,y,str-1);
      str--;
      *str = '\0';
      GSsetcolor(savecolor);
      num--;
    }
    else if (*str == KeyEscape)
      return(FALSE);

  *str = '\0' ;
  return(TRUE);
}


/***********************************************************************
 Read and return a string typed by user.
 GLOBALS: None.
***********************************************************************/

/* Apparantly, this is never used. JBR */
#if 0

static boolean DOSgetstring(
  int x, int y,  /* Position of string */
  char *msg,     /* Text to print for user */
  char *inp)  /* The string */
{
  boolean ret;
  int savecolor;

  savecolor = GSgetcolor();
  GSPdraw_string(x, y, msg);
  ret = DOSgetchars(inp, x+textwidth(msg)+8, y);
  GSsetcolor(0);
  GSPdraw_string(x, y, msg);
  GSPdraw_string(x+textwidth(msg)+8, y, inp);
  GSsetcolor(savecolor);
  return(ret);
}
#endif


/***********************************************************************
  Return a character array of red, green, blue values for each color in
  the pallette.  Red_i = values[3*i], Green_i = values[3*i+1],
  Blue_i = values[3*i+2]
************************************************************************/
static void readrgbpalette (char *gcm)
{
 int i;

 if (GraphicsMode == G640x480x16)
 {
   regs.x.ax = 0x1017;
   regs.x.bx = 0;
   regs.x.cx = 256;
   sregs.es = FP_SEG((void far *)gcm);
   regs.x.dx = FP_OFF((void far *)gcm);
   int86x (VIDEO, &regs, &regs, &sregs);
 }
 else
 {
   for (i=0; i<256; i++)
   {
	 gcm[i*3] = (char)Interface_Palette[i*3];
	 gcm[i*3+1] = (char)Interface_Palette[i*3+1];
	 gcm[i*3+2] = (char)Interface_Palette[i*3+2];
   }
 }
}

GSColor *GSPget_row(int rn, GSColor values[])
{
  int i;

  for (i = 0; i < SCREENWIDTH; i++)
    values[i] = getpixel(i, rn);
  return values;
}

/***********************************************************************
Quickly draw a row of pixels to the VGA screen.  The colors of
the "n" pixels are supplied in the array "row", with one color
per character.  (Only the lower four bits are used.)  The row
will be drawn starting at pixel (x,y) in the current window,
proceeding to the right.  Negative values in the array will
be skipped over.
***********************************************************************/
/*void  GSPfastrow(int x, int y, ColorIndexType *row, int n)*/
void  GSPfastrow(int x, int y, GSColor *row, int n)
{
  char far    *psc;    /* pointer to screen memory location */
  int i;

  if (GraphicsMode == G640x480x256)
  {
  /* following is strictly BGI calls to draw the row in SVGA */
	for (i=0; i<n; i++)
	  if (row[i] != 255)
		putpixel(x+i-Offx, y-Offy, row[i]);
	return;
  }
  /* following is fast assembly, but only works in VGA, not SVGA */
  psc = MK_FP(0xA000, 80*y + x/8);
  outport(0x3CE, 0x0205);    /* select VGA write mode 2 */
  outport(0x3C4, 0xFF02);    /* write to all four bit planes */

  asm push   ds
  asm lds    si, row
  asm les    di, psc

  asm mov    cx, x
  asm and    cx, 7
  asm mov    ah, 80h
  asm shr    ah, cl
  asm mov    al, 8

  asm mov    dx, 3CEh
  asm mov    cx, n

  next:
  asm mov    bl, ds:[si]
  asm cmp    bl, 0
  asm jl     skip

  asm out    dx, ax
  asm test   bl, es:[di]
  asm mov    es:[di], bl

  skip:
  asm ror    ah, 1
  asm jnc    okay
  asm inc    di

  okay:
  asm inc    si
  asm loop   next

  asm pop    ds

  outport(0x3CE, 0x0005);    /* Go back to write-mode 0. */
}

/**********************************************************************
 Initialize graphics.
 GLOBALS: GraphicsMode
**********************************************************************/
void GSPinit_system(char *appl_name, int dumint, char **dumptr)
{
  int g1, g2;
  int gerr;
  int userdrivernumber1, userdrivernumber2;
  int i = 0;
  struct palettetype pal;
  unsigned palette_editor_winsize;   /*--- for calculating palette editor's
					   backing store req'mt ---*/

  appl_name[0] = appl_name[0]; /*--- to appease compiler ---*/
  if (GraphicsMode == G640x480x256)
  /*--- If application requests 256 color mode, check to make sure an
	S3 chip is not installed.  If it is, don't use 256 color mode
	(no mouse support for S3 in 256 color mode). ---*/
  {
	userdrivernumber1 = installuserdriver("Svgas3", NULL);
	if ((gerr = registerfarbgidriver(SvgaS3_fdriver)) < 0)
    {
	  GSelog (-1, "Error registering SVGA BGI driver |%s|.\n",
		  grapherrormsg(gerr));
    }

    g1 = userdrivernumber1;
    g2 = DETECT;
    initgraph(&g1, &g2, ".");
    gerr = graphresult();
    closegraph();
    if ((g1 != userdrivernumber1) || (gerr != 0))
    { /* S3 not detected.  So try SVGA256M driver */
	  userdrivernumber2 = installuserdriver("Svga256M", NULL);
	  if ((gerr = registerfarbgidriver(SVGA256M_driver_far)) < grOk)
	    {
	    GSelog (-1, "Error registering SVGA BGI driver |%s|.\n", 
		    grapherrormsg(gerr));
	    }

      g1 = userdrivernumber2;
	  g2 = SVGA640x480x256;
	  initgraph(&g1, &g2, ".");
	  gerr = graphresult();
	  if (gerr == grOk)
      {  /* Initialize SVGA mode stuff */
	    palette_editor_winsize = GSPimagesize(0, 0, PALBARWIDTH256+4,
						  PALBARHEIGHT256+4);
      }
      else /* SVGA not available */
      {
	closegraph();
	GraphicsMode = G640x480x16;   /* SVGA not available!  So don't use 256 color mode */
	    GSelog (0, "SVGA video card not detected.  Switching to 16 color mode.\n");
      }
    }
    else
    {
      GraphicsMode = G640x480x16;   /* S3 detected!  So don't use 256 color mode */
	  GSelog (0, "S3 video card detected.  Switching to 16 color mode.\n");
    }
  }

  if (GraphicsMode == G640x480x16)
  {
	if ((gerr = registerfarbgidriver(EGAVGA_driver_far)) < grOk)
	  {
	  GSelog (-1, "Error registering BGI driver |%s|.\n",  grapherrormsg(gerr));
	  }
	g1 = VGA ;
	g2 = VGAHI ;
	initgraph(&g1, &g2, ".");
	gerr = graphresult();
	if (gerr != grOk)
	  GSelog (-1, "Graphics error: %s.\n", grapherrormsg(gerr));

	for (i=0; i<16; i++)         /* Use palette numbers to 0-15 */
	  pal.colors[i] = i;
	setallpalette (&pal);

	palette_editor_winsize = GSPimagesize(0, 0, PALBARWIDTH16+4, PALBARHEIGHT16+4);
  }

  if (registerfarbgifont (small_font_far) < 0)
	GSelog (-1, "Error setting up little font.\n");
  gerr = graphresult();
  if (gerr < 0)
	GSelog (-1, "Graphics error: %s.\n", grapherrormsg(gerr));

  GSPxormode(OFF);
  settextstyle (DEFAULT_FONT, HORIZ_DIR, 1);
  setfillstyle(SOLID_FILL, 0);


#ifdef DEBUG
  /* allocate backing store req'd for palette editor */
  if ((bm1 = (char *)malloc (palette_editor_winsize)) == NULL)
	GSelog (-1, "Out of memory for palette editor store 1.\n");
  if ((bm2 = (char *)malloc (palette_editor_winsize)) == NULL)
	GSelog (-1, "Out of memory for palette editor store 2.\n");
#endif
}


/**********************************************************************
Set Palette colors interactively. (EGA/VGA only)
GLOBALS: Interface_Palette[]
**********************************************************************/
void setpalettebar (void)
{
#ifdef DEBUG
  int i, j;
  register int x;
  register int y;
  int w, h;
  int dh, numrows, numcols;
  int right, left, top, bottom;
  char rgb[800];
  int status;
  int oldcx, oldcy;
  int px, py;
  BigString tmpstr;
  BigString inp;
  int val;
  struct palettetype pal;
  char ch;
  boolean press = FALSE;
  int oldfont, oldcolor;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();

  for(i=0; i<768; i++)
	 rgb[i]=Interface_Palette[i];

  if (GraphicsMode == G640x480x16)
  {
	readrgbpalette (rgb);
	getpalette (&pal);
	dh = 15;
	numrows = 1;
	numcols = 16;
	w = PALBARWIDTH16;
	h = PALBARHEIGHT16;
  }
  else
  {
	dh = 14;
	numrows = 8;
	numcols = 32;
	w = PALBARWIDTH256;
	h = PALBARHEIGHT256;
  }
  x = 5; y=5;

  GSmouseoff ();
  getimage (x-1, y-1, x+w+1, y+h/2, bm1);
  getimage (x-1, y+h/2, x+w+1, y+h+1, bm2);

  setcolor (15);
  rectangle (x-1, y-1, x+w+1, y+h+1);
  setfillstyle (SOLID_FILL, 0);
  bar (x, y, x + w, y + h);
  x += 5;
  y += 30;
  for(i=0; i<numrows; i++)
  {
	top = y+dh*i;
	bottom = y+dh*(i+1);
	for(j=0; j<numcols; j++)
	{
	  left = x+dh*j;
	  right = x+dh*(j+1);
	  rectangle (left-1, top-1, right+1, bottom+1);
	  setfillstyle (SOLID_FILL, i*numcols+j);
	  bar(left, top, right, bottom);
	}
  }
  GSmouseon ();
  while (TRUE)
	{ /* Mini event loop */
	oldcx = Cursx; oldcy = Cursy;
	get_mouse_position((short far *)&Cursx, (short far *)&Cursy);
	if (kbhit())
	  if ((ch = getch()) == 'q')
		break;
	  else
	  if (ch == 13)
		press = TRUE;
	if (press || get_mouse_press(0, &px, &py, &status)) /* mouse button pressed */
	  {
	  press = FALSE;
	  if ((Cursx > x) && (Cursx < x + dh*numcols) &&
		  (Cursy > y) && (Cursy < y + dh*numrows) )
		{
		GSmouseoff ();
	GSsetcolor(15);
		GSsetfont(LARGE);
		i = (((Cursy-y)/(dh))*numcols + ((Cursx-x)/(dh)))*3;
	GSsetcolor(0);
		bar (x+1, y+dh*numrows+2, x+150, y+dh*numrows+14);
		GSsetcolor (15);
		DOSgetstring (x, y+dh*numrows+12, "Enter red value:", inp);
		val = atoi (inp);
		if (val < 0) val = 0; if (val > 63) val = 63;
		rgb[i] = (char)val;
		i++;
		GSsetcolor (0);
		bar (x+1, y+dh*numrows+2, x+150, y+dh*numrows+14);
		GSsetcolor (15);
		DOSgetstring (x, y+dh*numrows+12, "Enter green value:", inp);
		val = atoi (inp);
		if (val < 0) val = 0; if (val > 63) val = 63;
		rgb[i] = (char)val;
		i++;
		GSsetcolor (0);
		bar (x+1, y+dh*numrows+2, x+150, y+dh*numrows+14);
		GSsetcolor (15);
		DOSgetstring (x, y+dh*numrows+12, "Enter blue value:", inp);
		val = atoi (inp);
		if (val < 0) val = 0; if (val > 63) val = 63;
		rgb[i] = (char)val;
		i-=2;
		setrgbpalette(i/3, rgb[i], rgb[i+1], rgb[i+2]);
		GSsetcolor (0);
		bar (x+1, y+dh*numrows+2, x+150, y+dh*numrows+14);
		GSmouseon ();
		oldcx = 0; oldcy = 0;
		}
	  }
	if (oldcx != Cursx || oldcy != Cursy)
	  {
	  if ((Cursx > x) && (Cursx < x + dh*numcols) &&
		  (Cursy > y) && (Cursy < y + dh*numrows) )
		{
		GSmouseoff ();
	GSsetcolor(15);
		GSsetfont(LARGE);
		GSsetcolor (0);
		bar (x+1, y-21, x+200, y-3);
		i = (((Cursy-y)/(dh))*numcols + ((Cursx-x)/(dh)))*3;
		sprintf (tmpstr, "pal#=%3d r=%2d g=%2d b=%2d",
		  i/3, rgb[i], rgb[i+1], rgb[i+2]);
		GSsetcolor (15);
		GSPdraw_string(x+1, y-10, tmpstr);
		GSmouseon ();
		}
	  }
	}

  for(i=0; i<768; i++)
	 Interface_Palette[i] = rgb[i];
  writepalettefile (rgb);

  GSmouseoff ();
  x-=5;
  y-=30;
  putimage (x-1, y-1, bm1, COPY_PUT);
  putimage (x-1, y+h/2, bm2, COPY_PUT);
  GSmouseon ();

  GSPclear_events();
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
#endif
}


/**********************************************************************
Returns the number of bytes required to store a bit image
replaces Borland imagesize() to handle SVGA images.
GLOBALS: GraphicsMode
**********************************************************************/

/* changed to unsigned long - JBR */

unsigned long GSPimagesize( int left,
			   int top,
			   int right,
					   int bottom)
{
/*
   Use the following formulas:
	* imsize = ((xwid >> 1)+1)*ywid+4;      (16-color)
	* imsize = xwid*ywid+4;                 (256-color)
	* imsize = 2*xwid*ywid+4;               (32768/65536-color)
*/

	if (GraphicsMode == G640x480x16)
		return( (((right-left+4) >>1 )+1)*(bottom-top+1)+4 );
	else
		return( (right-left+4)*(bottom-top+1)+4 );
}

/***********************************************************************
 Return the current font setting.
 GLOBALS: none.
***********************************************************************/
FontSizeType GSPgetfont(void)
{
  return CurrentFont;
}
/*------------------------------------------------------------------------------
  Returns the height of text in screen pixels
------------------------------------------------------------------------------*/
int GStextheight (
  char *text
)
{
  return (textheight(text));
}

/*------------------------------------------------------------------------------
  Returns the current global line width
------------------------------------------------------------------------------*/
int GSPgetlinesize (void)
{
  struct linesettingstype lineinfo;

  getlinesettings(&lineinfo);
  return (lineinfo.thickness);
}

/*------------------------------------------------------------------------------
  Returns the width of text in screen pixels
------------------------------------------------------------------------------*/
int GStextwidth (
  char *text
)
{
  return (textwidth(text));
}

/***********************************************************************
 Set the font.
 GLOBALS: none.
***********************************************************************/
void GSsetfont (FontSizeType num)     /* either SMALL or LARGE */
{
  settextjustify(LEFT_TEXT, BOTTOM_TEXT);
  if (num <= SMALL)
    settextstyle(SMALL_FONT, HORIZ_DIR, 0);
  else
    settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
  CurrentFont = num;
}



/***********************************************************************
 Set draw color for lines, text, everything.
 GLOBALS: none.
***********************************************************************/
void GSsetcolor(GSColor color)
{
  if (GraphicsMode == G640x480x16)
	color = color % 16;
  setcolor(color);
  setfillstyle(SOLID_FILL, color);
}

/***********************************************************************
 Get draw color for lines, text, everything.  Private.
 GLOBALS: none.
***********************************************************************/
GSColor GSgetcolor(void)
{
  return (GSColor)getcolor();
}

/***********************************************************************
 Set the thickness for all future lines.
 GLOBALS: None.
***********************************************************************/
void GSsetlinesize(int size)
{
  setlinestyle (SOLID_LINE, 0, size);
}


/***********************************************************************
  Initialize system color palette
***********************************************************************/
void GSPinit_palette( char pal[3*256],    /* pal of r g b colors */
		      int n)              /* number of colors */
{
  int i;
  struct palettetype bgipal;

  if (GraphicsMode == G640x480x16)
  {
    getpalette(&bgipal);         /* grab current palette definition */
    for (i=0; i<16; i++)         /* Use palette numbers to 0-15 */
	  bgipal.colors[i] = i;
    setallpalette (&bgipal);     /* set new palette definition */
  }

  for (i=0; i < 3*n; i+=3)      /* set rgb values for each palette index */
      setrgbpalette(i/3, pal[i], pal[i+1], pal[i+2]);
}


/***********************************************************************
 THIS should removed ASAP (01/19/94) -Bo
 Suspends graphics mode (enters text mode in DOS).
 GLOBALS: None.
***********************************************************************/
void GSPsuspendgraphics(void)
{
  /* restore system to text mode */
  restorecrtmode();
  GSPclose_events();
}


/***********************************************************************
 THIS should removed (01/19/94) -Bo
 Restores graphics mode and palette.
 GLOBALS: None.
***********************************************************************/
void GSPrestoregraphics(void)
{
  /* return to graphics mode */
  setgraphmode(getgraphmode());

  GSPinit_palette(Interface_Palette, (GraphicsMode == G640x480x256)?256:16);
  GSPinit_events();
}


/***********************************************************************
  Sets the clipping rectangle to x, y, wd, & ht.
***********************************************************************/
void GSPset_clip_rect(int x, int y, int wd, int ht)
{
  int result;

  if (x < 0)
     x = 0;
  else if (x > SCREENWIDTH)
     x = SCREENWIDTH-1;
  if (y < 0)
     y = 0;
  else if (y > SCREENHEIGHT)
     y = SCREENHEIGHT;

  if (x+wd > SCREENWIDTH)
     wd = SCREENWIDTH-x;
  if (y+ht > SCREENHEIGHT)
     ht = SCREENHEIGHT-y;

  Offx = x;
  Offy = y;

  if (!((wd>0)&&(ht>0)))
     setviewport(0,0,0,0, TRUE);
  else
      setviewport(x, y, x+wd-1, y+ht-1, TRUE);

  if ((result=graphresult()) != grOk)
	GSelog(-1, "setviewport failed.\n"
		"x: %d;\ty: %d\twd: %d\tht: %d",
		x, y, wd, ht);
}

/***********************************************************************
***********************************************************************/
void GSPframe_rect(int x, int y, int wd, int ht)
{
  rectangle(x-Offx, y-Offy, x-Offx+wd-1, y-Offy+ht-1);
}

/***********************************************************************
***********************************************************************/
void GSPfill_rect(int x, int y, int wd, int ht)
{
  bar(x-Offx, y-Offy, x-Offx+wd-1, y-Offy+ht-1);
}

/***********************************************************************
***********************************************************************/
void GSPframe_ellipse(int x, int y, int wd, int ht)
{
    ellipse(x+wd/2-Offx, y+ht/2-Offy, 0, 360, wd/2, ht/2);
}

/***********************************************************************
***********************************************************************/
void GSPfill_ellipse(int x, int y, int wd, int ht)
{
    fillellipse(x+wd/2-Offx, y+ht/2-Offy, wd/2, ht/2);
}

/***********************************************************************
***********************************************************************/
void GSPline(int x1, int y1, int x2, int y2)
{
  line(x1-Offx, y1-Offy, x2-Offx, y2-Offy);
}

/***********************************************************************
Write a message at position (x, y).  Static because only the interface
handler should write to absolute coordinates.
GLOBALS: None.
***********************************************************************/
void GSPdraw_string(int x, int y, char *msg)
{
  HugeString tempstr;
  int i;

  for(i=0; msg[i] != '\0'; i++)
    tempstr[i]=isprint(msg[i])?msg[i]:' ';

  tempstr[i]='\0';

  if (CurrentFont == SMALL) {
    outtextxy(x-Offx+1, y-2-Offy, tempstr);
  }
  else {
    outtextxy(x-Offx+1, y-Offy, tempstr);
  }
}


/***********************************************************************
 Draw a polygon using current line style and color.  Then fills the
 poly using the current color.
***********************************************************************/
void GSPfillpoly(
  int numpoints,   /* number of vertices in polygon */
  int *polypoints) /* array of 2*numpoints ints that are the x-y coords of the vertices */
{
  int i;

  for (i=0; i<numpoints; i++)
  {
    polypoints[2*i]   -= Offx;  /* subtract x offset of current clip rect */
    polypoints[2*i+1] -= Offy;  /* subtract x offset of current clip rect */
  }

  fillpoly(numpoints, polypoints);
}

/**********************************************************************
Set up memory for backing store
**********************************************************************/
boolean GSPinit_bstore(int bstorenum, int wd, int ht)
{

  /* InfState.bsize will be at least 1/3 the size of the largest popup.
	 If bstore == TRUE, then InfState.bsize will be 1/3 the size of the
	 largest of all the popup menus and all the windows. */

  InfState.backingstore[bstorenum].bsize = GSPimagesize(0, 0, wd, (int)(ht/3.0+1) );

  if ((InfState.backingstore[bstorenum].bstore1 = malloc(InfState.backingstore[bstorenum].bsize)) == NULL)
    return FALSE;
  if ((InfState.backingstore[bstorenum].bstore2 = malloc(InfState.backingstore[bstorenum].bsize)) == NULL)
    return FALSE;
  if ((InfState.backingstore[bstorenum].bstore3 = malloc(InfState.backingstore[bstorenum].bsize)) == NULL)
    return FALSE;
  InfState.backingstore[bstorenum].bstore1[0] = BSTORE_CHECK;
  InfState.backingstore[bstorenum].bstore2[0] = BSTORE_CHECK;
  InfState.backingstore[bstorenum].bstore3[0] = BSTORE_CHECK;
  return TRUE;
}


/**********************************************************************
copy an image from screen to backing store
**********************************************************************/
boolean GSPto_bstore(int bstorenum, int left, int top, int wd, int ht)
{
  unsigned long size;
  int OldCurrWinNum;
  struct viewporttype op;

  if ((wd > SCREENWIDTH) || (ht > SCREENHEIGHT))
  {
    GSelog(0, "Error.  Width or Height is greater than maximum allowed.\n"
	      "  Max Width is %d.  Requested Width is %d.  Max Height is %d.  Requested Height is %d",
	      SCREENWIDTH, SCREENHEIGHT, wd, ht);
    return FALSE;
  }
  OldCurrWinNum = InfState.cw;
  GSPsetcurrentwindow(0);
  /* added this - JBR */
  getviewsettings(&op);
  GSPset_clip_rect(0, 0, 640, 480);
  if (InfState.backingstore[bstorenum].bstore1[0] != BSTORE_CHECK) {
    GSelog(0, "Backing store %d is unready.\n", bstorenum);
    return FALSE;
    }
  size = GSPimagesize(left, top, left + wd-1, top + ht/3 );
  if ((size == 0) || ( size > InfState.backingstore[bstorenum].bsize)) {
    GSelog(0, "Backing store error: size=%d; store %d size=%d\n",
                  size, bstorenum, InfState.backingstore[bstorenum].bsize);
    return FALSE;
    }
  getimage(left-Offx, top-Offy, left-Offx + wd-1, 
		top-Offy + ht/3, InfState.backingstore[bstorenum].bstore1);
  getimage(left-Offx, top-Offy + ht/3, left-Offx +  wd-1, 
		top-Offy + 2*ht/3, InfState.backingstore[bstorenum].bstore2);
  getimage(left-Offx, top-Offy + 2*ht/3, left-Offx + wd-1, 
		top-Offy + ht, InfState.backingstore[bstorenum].bstore3);
  InfState.backingstore[bstorenum].bstoreht = ht;
/*
  GSPset_clip_rect(op.left, op.right, op.right - op.left, op.bottom-op.top);*/
  GSPset_clip_rect(op.left, op.top, op.right - op.left, op.bottom-op.top);
  GSPsetcurrentwindow(OldCurrWinNum);
  return TRUE;
}

/**********************************************************************
copy an image from backing store to screen
**********************************************************************/
boolean GSPfrom_bstore(int bstorenum, int left, int top)
{
  int OldCurrWinNum;
  struct viewporttype op;

  OldCurrWinNum = InfState.cw;
  GSPsetcurrentwindow(0);
  getviewsettings(&op);
  GSPset_clip_rect(0, 0, 640, 480);
  if (InfState.backingstore[bstorenum].bstore1[0] == BSTORE_CHECK)
    return FALSE;
  putimage(left-Offx, top-Offy, InfState.backingstore[bstorenum].bstore1, COPY_PUT);
  putimage(left-Offx, top-Offy + InfState.backingstore[bstorenum].bstoreht/3, InfState.backingstore[bstorenum].bstore2, COPY_PUT);
  putimage(left-Offx, top-Offy + 2*InfState.backingstore[bstorenum].bstoreht/3, InfState.backingstore[bstorenum].bstore3, COPY_PUT);
  InfState.backingstore[bstorenum].bstore1[0] = BSTORE_CHECK;
  InfState.backingstore[bstorenum].bstore2[0] = BSTORE_CHECK;
  InfState.backingstore[bstorenum].bstore3[0] = BSTORE_CHECK;
/*  GSPset_clip_rect(op.left, op.right, op.right - op.left, op.bottom-op.top);*/
  GSPset_clip_rect(op.left, op.top, op.right - op.left, op.bottom-op.top);
  GSPsetcurrentwindow(OldCurrWinNum);
  return TRUE;
}

/**********************************************************************
**********************************************************************/
void GSPdraw_pixel(int x, int y)
{
  putpixel(x-Offx, y-Offy, getcolor());
}

/**********************************************************************
**********************************************************************/
GSColor GSPget_pixel(int x, int y)
{
  return((GSColor)getpixel(x-Offx, y-Offy));
}

/**********************************************************************
  No-op. This function is required in X-WIndows version to set input
  focus to the application window. It does nothing in the DOS version.
  Unfortunately, it must be called from GSinterfaceinit in order to
  work in the X version, so the DOS version must call it as well.
**********************************************************************/
void GSPset_focus(void)
{
}
