#include <stdio.h>
#include "geosim.h"
#include "gspriv.h"

typedef enum {NORMAL_STATE, MOUSEDOWN_STATE} EventLoopState;

#define CTRL_Q        17 /* Quit key */
#define CTRL_R        18 /* palette key */
#define CTRL_T        20 /* Fast cursor toggle */
#define CTRL_P        16 /* Fast cursor toggle */


/*----------------------------------GLOBALS-----------------------------------*/

static EventLoopState CurrState = NORMAL_STATE; /* Current interface state */
static int            ButtonSaveCurrWinNum;
static boolean        Depressed = FALSE;        /* TRUE iff current button is
                                                   drawn depressed */
static int            Oldx;                     /* mouse coordinates of ... */
static int            Oldy;                     /* ...previous event */
static int            Drag_initx;               /* mouse coordinates as ... */
static int            Drag_inity;               /* ... drag begins */
static GSButton      *CurrButt;                 /* the button most recently
                                                   pressed */
static int            CurrWinnum = 0;           /* Window number of CurrButt */
static int            CurrButtnum = 0;          /* Button number of CurrButt */
static boolean        InDragProcess = FALSE;    /* TRUE iff the mouse is being
                                                   dragged */
static boolean        FirstPressed = FALSE;     /* TRUE when repeatable button
                                                   is pressed, but only first
                                                   time through event loop. */
static boolean        FirstExposeSeen = FALSE;


/*-------------------- External Globals - in intrface.c ----------------------*/

extern GSWindow      *Win;                      /* All of the windows */
extern Popupmenu     *Popups;                   /* All of the popup menus */
extern Iconfig        InfState;                 /* Current state of interface */
extern int            Fast;                     /* For a faster k-board cursor */
extern int            Cursx, Cursy;             /* Pixel coordinates of cursor */
extern int            PopupSaveMsg;

/*------------------------------------------------------------------------------
  Calls the application's redraw-all function.
  Then redraws whatever popup menu is active.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSPdo_evexpose (
  void
)
{
  int      i;
  int      num;
  GSColor  oldcolor;
  int      oldlinesize;
  int      save_cw;
  int      slot = 0;

  /*--- save current interface state ---*/
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();
  save_cw = InfState.cw;

  /*--- call application's redraw function;
        then redraw whatever popup menu was active ---*/
  GSmouseoff();
  InfState.redrawfunc();
  GSmouseon();

  if (InfState.activepopupwindow != -1)
  {
    GSsetlinesize(THIN);
    GSsetcurrentnamedwindow("root");
    num = InfState.activepopupwindow;
    GSsetcolor(Popups[num].bg);
    GSPfill_rect(Popups[num].x, Popups[num].y, Popups[num].wd, Popups[num].ht);
    GSsetcolor(Popups[num].bcolor);
    GSPframe_rect(Popups[num].x, Popups[num].y, Popups[num].wd, Popups[num].ht);
    for (i=0; i < Popups[num].numitems; i++)
    {
      if (Popups[num].item[i].active != -1)
      {
        display_popup_label(num, i, slot);
        slot++;
      }
    }
  }
  /*--- restore previous interface state ---*/
  GSsetcolor(oldcolor);
  GSPsetcurrentwindow(save_cw);
  GSsetlinesize(oldlinesize);
}

/*------------------------------------------------------------------------------
  Handles EvMouseDown events.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSPdo_evmousedown (
  void
)
{
  static  int nextwinnum=0, nextbuttnum=0;
  static  boolean nextbutton = FALSE;
  static  boolean gotbutton = FALSE;

  if (CurrState == NORMAL_STATE)
  {
    if (InfState.activepopupwindow >= 0)
    /*--- If there is a popup window on screen, do that ---*/
    {
      CurrState = MOUSEDOWN_STATE;
    }
    else
    {
      do
      { /*--- get the last active button ---*/
        nextbutton = GSPgetbutton(Cursx, Cursy, &nextwinnum, &nextbuttnum);
        if (nextbutton)
        {
          if (Win[nextwinnum].wb[nextbuttnum].active==1)
          { /*--- 1 == active and visible ---*/
            CurrButt = &(Win[nextwinnum].wb[nextbuttnum]);
            CurrWinnum = nextwinnum;
            CurrButtnum = nextbuttnum;
            gotbutton = TRUE;
          }
          nextbuttnum++;
        }
      } while(nextbutton); /*--- end do while (gotbutton) ---*/
      nextbutton = FALSE; /*--- reset next button stuff ---*/
      nextwinnum = 0;
      nextbuttnum = 0;
      if (gotbutton) 
      { /*--- if cursor is in a button ---*/
        CurrState = MOUSEDOWN_STATE;
        ButtonSaveCurrWinNum = InfState.cw;
        GSPsetcurrentwindow(CurrWinnum);
        InfState.cb = CurrButtnum;
        if (CurrButt->drag)
        { /*--- and if CurrButton IS draggable ---*/
          Drag_initx = Cursx; /*--- set x and y at start of potential DRAG ---*/
          Drag_inity = Cursy;
        } /* end else CurrButton IS draggable */
        else
        { /*--- CurrButton is not draggable ---*/
          Depressed = TRUE;
          GSPdraw_depressed_button(CurrWinnum, CurrButtnum);
          if (CurrButt->repeatable)
            FirstPressed = TRUE;
        }
        gotbutton = FALSE;
      } /*--- end if (gotbutton) ---*/
      else
      { /*--- else cursor pressed outside a button or menu ---*/
        GSbell();
      }
    }
  } /*--- end if (CurrState == NORMAL_STATE) ---*/
}

/*------------------------------------------------------------------------------
  Handles EvMouseUp events.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSPdo_evmouseup (
  void
)
{
  if (CurrState == MOUSEDOWN_STATE )
  { /*--- mouse button is pressed ---*/
    CurrState = NORMAL_STATE; /*--- reset the state ---*/
    if (InfState.activepopupwindow >= 0)
    { /*--- If there is a popup window on screen, do that ---*/
      GSPhandle_popups(Cursy, GSPget_popup_item(Cursx, Cursy));
    }
    else if (CurrButt->active == 1)
    {
      if (CurrButt->drag)
      { /*--- the current button is draggable ---*/
        GSmouseoff();
        setlastcontrol(WINDOW, Win[CurrWinnum].name,
                       Win[CurrWinnum].wb[CurrButtnum].name, -1);
        if (InDragProcess)
        { /*--- if we are in a DRAG process, ---*/
          InDragProcess = FALSE;
          /*--- send DRAG_FINAL to callback func ---*/
          (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                              Cursy-(Win[CurrWinnum].y+CurrButt->y),
                              DRAG_FINAL,
                              CurrButt->name); 
        }
        else if (GSPinside_button(Cursx, Cursy, CurrWinnum, CurrButtnum))
        { /*--- else send MOUSE_CLICK to callback func ---*/
          (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                              Cursy-(Win[CurrWinnum].y+CurrButt->y),
                              MOUSE_CLICK,
                              CurrButt->name); 
        }
        GSmouseon();
      } /*--- end if (CurrButt->drag) ---*/
      else
      {
        if (Depressed)
        {
          /*--- undepress the button ---*/
          GSPdraw_released_button(CurrWinnum, CurrButtnum);  
          Depressed = FALSE;
        }
        if (GSPinside_button(Cursx, Cursy, CurrWinnum, CurrButtnum))
        {
          /*--- execute the button func iff mouse is inside the button ---*/
          setlastcontrol(WINDOW, Win[CurrWinnum].name,
                         Win[CurrWinnum].wb[CurrButtnum].name, -1);
          /*--- restore current window to state before button press ---*/
          GSPsetcurrentwindow(ButtonSaveCurrWinNum);
          GSmouseoff();
          PopupSaveMsg = GSgetcurrmsg();
/*----------------------------------------------------------
          (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                              Cursy-(Win[CurrWinnum].y+CurrButt->y), 0);
----------------------------------------------------------*/
          (*((boolean(*)(char *))CurrButt->func))
                                          (Win[CurrWinnum].wb[CurrButtnum].name);

          if (CurrButt->popupmenunum != -1)
          {
            createpopupmenu(CurrButt->popupmenunum);
          }
          GSmouseon();
        }
      }
    }
    /*CurrState = NORMAL_STATE;*/ /*--- reset the state ---*/
    if (CurrState == MOUSEDOWN_STATE )
      FirstPressed = FALSE;
    CurrWinnum = 0;
    CurrButtnum = 0;
  }
}


/*------------------------------------------------------------------------------
  Handles whatever should be down if no event exists
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSPdo_evnone (
  void
)
{
  static int holding = 0;
  static int threshold;
  static long tused;
  if ((CurrState == MOUSEDOWN_STATE) && (InfState.activepopupwindow < 0))
  { /*--- mouse button is pressed ---*/
    /*--- If there is NO popup window on screen, do that ---*/
    if ((GSPinside_button(Cursx, Cursy, CurrWinnum, CurrButtnum)) &&
        (CurrButt->active == 1))
    { /*--- mouse is still in original button, and button is still active ---*/
      GSPsetcurrentwindow(CurrWinnum);
      if ((!CurrButt->drag) && (!Depressed))  
      { /*--- if it's not already drawn depressed, ---*/
        Depressed = TRUE; /*--- draw the button depressed ---*/
        GSPdraw_depressed_button(CurrWinnum, CurrButtnum);
      }
      if (CurrButt->repeatable) /* OK, I'm positive that ONLY a real button (not a
      								dragarea) can be repeatable.  If so, then the
      								button's callback expects the name of the button as
      								an argument.
      								I'm fixing the call to the button's func below based
      								on this assumption. JBR
      								*/
      { /*--- the current button is repeatable ---*/
        if (FirstPressed)
        { /*--- if this is the first pass after it was pressed ---*/
          holding = 0; /*--- initialize the threshold, etc. ---*/
          threshold = 1000;
          GSPstart_timer();
          FirstPressed = FALSE; /*--- reset FirstPressed ---*/
        }
        else
        { /*--- get time since last GSstart_timer() ---*/
          tused = GSPread_timer() * InfState.speedup;
          if (tused > threshold)
          { /*--- if enough time has passed, ---*/
            setlastcontrol(WINDOW, Win[CurrWinnum].name,
            Win[CurrWinnum].wb[CurrButtnum].name, -1);
            GSmouseoff(); /*--- execute the button's function ---*/
#if 0
		/* This is the old way - passing in mouse coordinates is appropriate for
			a dragarea, but not for a real button. - JBR */
            (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                                Cursy-(Win[CurrWinnum].y+CurrButt->y), 0);
#else
			(*((boolean(*)(char *))CurrButt->func))(Win[CurrWinnum].wb[CurrButtnum].name);
#endif
            GSmouseon();
            /*--- change the threshold, depending on how long this button has
                  been held down ---*/
            if (holding < 4)
              threshold = 500;
            else if (holding < 8)
              threshold = 200;
            else
              threshold = 0;
            GSPstart_timer();
              holding++;
          } /*--- if not enough time has passed, do nothing ---*/
        }
      } /*--- end if (CurrButt->repeatable) ---*/
      else if (CurrButt->drag) /*--- the current button is draggable ---*/
      {
        if (InDragProcess) /*--- if we are in a DRAG ---*/
        {
          if ((Cursx != Oldx) || (Cursy != Oldy))
          { /*--- and if mouse has also moved since last event ---*/
            GSmouseoff();
            (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                                Cursy-(Win[CurrWinnum].y+CurrButt->y),
                                DRAG_PROCESS);
            GSmouseon();
          }
        }
        else if ((Cursx != Drag_initx) || (Cursy != Drag_inity))
        { /*--- else if the mouse has moved in dragrect ---*/
          InDragProcess = TRUE; /*--- begin DRAG process ---*/
          setlastcontrol(WINDOW, Win[CurrWinnum].name,
                         Win[CurrWinnum].wb[CurrButtnum].name, -1);
          GSmouseoff();
          (*(CurrButt->func))(Cursx-(Win[CurrWinnum].x+CurrButt->x),
                              Cursy-(Win[CurrWinnum].y+CurrButt->y), DRAG_INIT);
          GSmouseon();
        }
      } /*--- end if (CurrButt->drag) ---*/
    }
    else
    { /*--- mouse outside CurrButton ---*/
      if ((!CurrButt->drag) && (Depressed))
      { /*--- if not a drag area & depressed = true, undepress the button ---*/
        GSPdraw_released_button(CurrWinnum, CurrButtnum); 
        Depressed = FALSE; /*--- reset depressed = false ---*/
      }
    } /*--- end else mouse outside CurrButton ---*/
  } /*--- end if (CurrState == MOUSEDOWN_STATE ) ---*/
}


/*------------------------------------------------------------------------------
  Processes one event and returns.
  Globals: Oldx, Oldy, Cursx, Cursy
------------------------------------------------------------------------------*/
static void GSPdo_one_event (
  GSEvent *event
)
{

  Oldx = Cursx;
  Oldy = Cursy;
  if (event->type != EvExpose) /*dhines*/
  { /*dhines*/
    Cursx = event->mousex;
    Cursy = event->mousey;
  } /*dhines*/

  switch (event->type)
  {
    case EvExpose:
      GSPdo_evexpose();
      break;
    case EvMouseDown:
      GSPdo_evmousedown();
      break;
    case EvMouseUp:
      GSPdo_evmouseup();
      break;
    case EvKeyDown:
      GSPkeyevent(Cursx, Cursy, event->key);
      break;
    case EvNone:
      GSPdo_evnone();
      break;
  } /* end switch (event.type) */

  if (InfState.activepopupwindow >= 0)
  { /*--- if there are any popup menus, do their highlights ---*/
    highlight(Cursx, Cursy);
  }
  if ((Oldx != Cursx) || (Oldy != Cursy))
  { /*--- if mouse has moved since last time through loop ---*/
    GSPprocessmouseinside(Cursx, Cursy);
  }
  GSPdoactivefuncts(Cursx, Cursy);
} /* end void GSPdo_one_event(GSEvent) */


/*------------------------------------------------------------------------------
  Processes all events on the queue until the queue is empty. Then returns to
  caller. (as opposed to GSPdo_events_forever which never returns).
------------------------------------------------------------------------------*/
void GSdoeventsandreturn (
  void
)
{
  GSEvent event;

  while (GSPnext_event(&event))
  {
	if (event.type == EvNone) break; /* VC++: No events to handle -- CAS */
    GSPdo_one_event(&event);
    GSPflush(); /* TEST */
  }
}

/*------------------------------------------------------------------------------
  Main Event loop. Continuously checks for and processes all events.
------------------------------------------------------------------------------*/
void GSPdo_events_forever (
  void
)
{
  GSEvent event;

  while (!FirstExposeSeen)
  { /*--- ignore all events until the first expose event occurs ---*/
    (void)GSPnext_event(&event);
    if (event.type == EvExpose) FirstExposeSeen = TRUE;
  }

  do
  { /*--- do events till the cows come home ---*/
    GSPdo_one_event(&event);
    (void)GSPnext_event(&event);
  } while (TRUE);

}

/*------------------------------------------------------------------------------
  Returns TRUE iff the cursx and cursy coords are inside the button or dragrect
  specified by winnum, bnum.
  Globals: Win
------------------------------------------------------------------------------*/
boolean GSPinside_button (
  int cursx,    /*--- cursor x position ---*/
  int cursy,    /*--- cursor y position ---*/
  int winnum,   /*--- window number ---*/
  int bnum      /*--- button number ---*/
)
{
  int x, y;

  x = Win[winnum].x + Win[winnum].wb[bnum].x;
  y = Win[winnum].y + Win[winnum].wb[bnum].y;
  if ((cursx > x) && (cursx < x + Win[winnum].wb[bnum].wd) &&
      (cursy > y) && (cursy < y + Win[winnum].wb[bnum].ht))
  {
    return(TRUE);
  }
  else
  {
    return(FALSE);
  }
}


/*------------------------------------------------------------------------------
  Looks for and executes active functions in the list.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSPdoactivefuncts (
  int cursx,
  int cursy
)
{
  int i;

  for (i = 0; i < InfState.numfuncs; i++)
  { /*--- check for active function to call ---*/
    if (InfState.funcs[i].active)
    {
      if (InfState.funcs[i].mouse_off)
      { /*--- hide mouse if funct desires it ---*/
        GSmouseoff();
      }
      (*(InfState.funcs[i].func))(cursx, cursy, 0);
      GSmouseon();
    } /*--- end if (InfState.funcs[i].active) ---*/
  }
}


/*------------------------------------------------------------------------------
  Process keyboard input.
  Handles some keys (arrow cursor keys, CTRL-Q, etc) and passes others
  on to the application's key handler function, if it exists.
  GLOBALS: Win, InfState
------------------------------------------------------------------------------*/
void GSPkeyevent (
  int  cursx,
  int  cursy,
  char ch
)
{
  switch (ch)
  {
/* Removed - JBR
    case KeyUp: case KeyDown: case KeyLeft: case KeyRight:
      process_arrow(ch) ;
      break;
*/

#ifdef DEBUG
    case CTRL_R:
      setpalettebar ();
      break;
#endif
    case CTRL_T:
      Fast = !Fast;
      break;
    case CTRL_Q:
      GSelog (-1, "Program terminated by the user pressing 'Control-Q'.");
      break;
/*    case KeyF1:
      printf("Ouch!\n");
      break;*/
    default:
      if (InfState.keyfunc != NULL)
      {
        GSmouseoff();
        (*(InfState.keyfunc))( (cursx-Win[InfState.cw].x),
                               (cursy-Win[InfState.cw].y), ch);
        GSmouseon();
      }
      break;
  }
}


/*------------------------------------------------------------------------------
  Activate a registered function to handle character events.
  Only one keyboard handler may be active at one time.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GScharfunc (
  char *name
)
{
  int i;

  i = getname(InfState.funcs, sizeof (Function), InfState.numfuncs, name,
              "function");
  InfState.keyfunc = (boolean(*)(int, int, char))InfState.funcs[i].func;
}


/*------------------------------------------------------------------------------
  Deactivate the keyboard handler function.
  GLOBALS: InfState
------------------------------------------------------------------------------*/
void GSclearcharfunc (
  void
)
{
  InfState.keyfunc = NULL;
}

/*------------------------------------------------------------------------------
 NOTE: This should only be called if the mouse has moved.  This function
 will not check for mouse movement between invocations.

 Process mouse up -- no button press, but maybe in a dragrect
                  -- or perhaps just left a dragrect
 GLOBALS:
------------------------------------------------------------------------------*/
void GSPprocessmouseinside (
  int cursx,
  int cursy
)
{
  int i=0, j=0, k=0;
  static int lastw[MAXOVERRLAPBUTTONS] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
  static int lastb[MAXOVERRLAPBUTTONS] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
  GSButton *butt;
  int savecurrentwin;

  savecurrentwin = InfState.cw;

  /*--- Check for going outside of all buttons we were in before ---*/
  for (k=0; k<MAXOVERRLAPBUTTONS; k++)
  {
    if (lastw[k] != -1)
    {
      if (!GSPinside_button(cursx, cursy, lastw[k], lastb[k]) &&
          (Win[lastw[k]].active == InfState.level))
      {
        GSPsetcurrentwindow(lastw[k]);
        GSmouseoff();
        (*(Win[lastw[k]].wb[lastb[k]].func))
            (cursx-(Win[lastw[k]].x+Win[lastw[k]].wb[lastb[k]].x),
             cursy-(Win[lastw[k]].y+Win[lastw[k]].wb[lastb[k]].y),
             MOUSE_OUTSIDE);
        GSmouseon();
        /*--- clear that button from the list ---*/
        lastw[k] = -1;
        lastb[k] = -1;
      }
      else if (!(Win[lastw[k]].active==InfState.level))
      { /*--- clear that button from the list ---*/
        lastw[k] = -1;
        lastb[k] = -1;
      }
    }
  }

  if (GSPgetbutton (cursx, cursy, &i, &j))
  {
    do
    {
      butt = &(Win[i].wb[j]);

      if (butt->drag)
      { /*--- save the current drag rectangle ---*/
        for (k = 0; k < MAXOVERRLAPBUTTONS; k++)
        {
          if (lastw[k] != -1)
          { /*--- see if this drag rect button is already in the list ---*/
            if ((lastw[k] == i) && (lastb[k] == j))
            {
              break;
            }
          }
          else
          {
            lastw[k] = i;
            lastb[k] = j;
            break;
          }
        }
        if (k == MAXOVERRLAPBUTTONS)
        {
          GSelog(0, "processmouseinside: Too many overlapping drag areas.");
          GSelog(-1, "Max = |%d|\n", MAXOVERRLAPBUTTONS);
        }

        /*--- process inside a drag rectangle, but button is up ---*/
        GSPsetcurrentwindow(i);
        GSmouseoff();
        (*(butt->func))(cursx-(Win[i].x+butt->x),
                        cursy-(Win[i].y+butt->y), MOUSE_INSIDE);
        GSmouseon();
      }
      j++;
    } while (GSPgetbutton(cursx, cursy, &i, &j));
  }
  GSPsetcurrentwindow(savecurrentwin);
  return;
}


/*------------------------------------------------------------------------------
Return TRUE iff the cursor is currently in an active button.
  Begin search for window and button at the window i and button j.
  i.e. if i == 0 and j == 0, search from the beginning (Win[0].wb[0])
  Return the first button that the cursor is in.
GLOBALS: InfState, Win, Popups, Cursx, Cursy.
------------------------------------------------------------------------------*/
boolean GSPgetbutton (
  int x,
  int y,
  int *i,      /* begin search at window i and return window number in i */
  int *j)      /* begin search at button j and return button number in j */

{
  for (; *i < InfState.numwindows; (*i)++)
  {
    if (Win[*i].active == InfState.level)
    {
      for (; *j < Win[*i].numbuttons; (*j)++)
      {
        if (GSPinside_button(x, y, *i, *j )) /* Found button */
        {
          return(TRUE);
        }
      }
      *j = 0;
    }
  }
  return(FALSE);
}


/*------------------------------------------------------------------------------
  Returns the item number of the menu item the cursor is in,
  or -1 if cursor outside of menu.

  GLOBALS: InfState, Popups
------------------------------------------------------------------------------*/
int GSPget_popup_item (
  int currx,
  int curry
)
{
  int i;

  i = InfState.activepopupwindow;
  if ((Popups[i].x <= currx) && (currx <= Popups[i].x + Popups[i].wd) &&
      (curry >= Popups[i].y) && (curry <= (Popups[i].y + Popups[i].ht)) )
  { /*--- cursor is inside of menu ---*/
    return ( pickitem(curry, i));
  }
  else
  { /*--- cursor is outside of menu ---*/
    return (-1);
  }
}

/*------------------------------------------------------------------------------
  Executes the callback function of a menu item, if the cursor is inside
  a menu.
  GLOBALS: InfState, Popups
------------------------------------------------------------------------------*/
boolean GSPhandle_popups (
  int curry,
  int itemnum
)
{
  int i;

  i = InfState.activepopupwindow;
  if ((curry <= Popups[i].y+1)  ||
      (itemnum >= Popups[i].numitems) || (itemnum == -1))
  { /*--- if click outside menu */
    setlastcontrol(NONE, "", "", -1);
    GSmouseoff();
    removepopup();
    if (Popups[i].func != NULL)
    {
      (*Popups[i].func)();
    }
    GSmouseon();
    return FALSE;
  }
  if (Popups[i].item[itemnum].active == 0)
  { /*--- if menu item is disabled ---*/
    GSbell();
    setlastcontrol(NONE, "", "", -1);
    return FALSE;
  };
  if (Popups[i].item[itemnum].func != NULL)
  {
    GSmouseoff();
    if (Popups[i].checklist)
    { /*--- if this is a checklist menu ---*/
      toggle_check(i, itemnum, InfState.highlight); /*--- toggle checkmark ---*/
    }
    else
    {
      setlastcontrol(NONE, "", "", -1); /*--- is this necessary 01/27/94 -Bo ---*/
      removepopup();
    }
    setlastcontrol(MENU, Popups[i].name, "", itemnum);
    (*Popups[i].item[itemnum].func)(Popups[i].item[itemnum].name);
    GSmouseon();
  }
  return TRUE;
}



