/**********************************************************************
GeoSim standard interface file interpreter and interface handler.
**********************************************************************/

#define PARSER

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "geosim.h"
#include "gspriv.h"

#ifdef PARSER
#include "ifparse.h"

void ifparse();
void ifrestart(FILE *);
void iferror(char *msg);
 /* I don't know why these protos aren't in ifparse.h. JBR */
 
#endif

#define ESC_SYM       '\\'
#define CHECKMARK     "*" /* Character to use marking items in  Cheklists */

#define CURSOR_RESO    5 /* Number of pixels in one cursor move */

#define CTRL_P        16
#define CTRL_Q        17 /* Quit key */
#define CTRL_R        18 /* palette key */
#define CTRL_T        20 /* Fast cursor toggle */

#define CURSIZE       16 /* Size of cursor bounding square */


#define NOTEXT "@@"  /* symbol for no label text in button */


/***********************************************************************
GLOBAL VARIABLES
***********************************************************************/

Iconfig InfState;      /* Current state of the interface */
Popupmenu *Popups;   /* All of the popup menus */
GSWindow *Win; /* All of the windows */
int GraphicsMode;
int PopupSaveMsg=-1;
static int PopUpSaveCurrWinNum;

/* Internal prototypes */
void drawarrow (int wnum, int bnum, int tx, int twd, int awd, int aht);
/**/

#ifdef PARSER
  int msgs, aliases, functions, windows, popups, fields;
  int windata[MAXWINDOW][2]; /* windata[i][0] is number of buttons,
                                windata[i][1] is number of graphics */
  int popdata[MAXPOPUP]; /* Number of plabels in each popup */
  boolean firstPass;

  boolean (*tempptr)(int,int,int,...);

#endif

#ifdef PARSER
  int        currmsg = -1;             /* Index for current message */
  int        currpop = -1;             /* Index for current popup menu */
  int        currplab;                 /* Index for current label in current
                                          popup menu */
  int        currwin = -1;             /* Index for current window */
  int        currbutt;                 /* Index for current button in current
                                          window */
  int        currgraphic;              /* Index for current graphic in current
                                          window */
  int        currfield = -1;           /* Index for current field */
  int        currfunc = -1;            /* Index for current function */
  int        currcolor;
  int        clist = 0;  
  int        curr_alias = 0;           /* Why not -1? CAS *//* Index for current alias*/
#endif

#ifdef PARSER
 HugeString lineBuf;
 int lineNumber;
#endif

#ifdef PARSER
  extern FILE *ifin;
#endif

char *Colorlist[PAL_SIZE+1]
#if 0
= {
  "BLACK", "BLUE", "GREEN", "CYAN", "RED",
  "MAGENTA", "BROWN", "LGRAY", "DGRAY", "LBLUE", "LGREEN",
  "LCYAN", "LRED", "LMAGENTA", "YELLOW", "WHITE", ""
}
#endif
;

char *Std_Colorlist[NUM_STD_COLORS+1] = {"black", "white", NULL};
char *Dflt_Colorlist[NUM_DFLT_COLORS+1] = 
	{"dgray", "mgray", "lgray", "bluegray", "lblue", "peach", NULL};

char dflt_rgb[(NUM_STD_COLORS+NUM_DFLT_COLORS)*3] =
			{ 0, 0, 0,
			  63, 63, 63, /* white */
			  50, 45, 45, /* dgray */
			  30, 30, 55, /* mgray */
			  53, 53, 53, /* lgray */
			  0, 50, 50,  /* bluegray */
			  10, 10, 55, /* lblue */
			  63, 47, 50  /* peach */
			};
/*  default color list  */
char *Alias[MAXALIASES+1];           /* color aliases for file */
char *Acolor[MAXALIASES+1]; /* colors for aliases */

char Interface_Palette[3*PAL_SIZE];

/* FuncNames must be declared in the module that is
using this module.  The calling module must have an actual function
for each entry on FuncNames, unless the value is NULL (in which case,
no call will be made). */
extern IntrFunction FuncNames[]; /* Association of function names to ptrs */
extern int NumFuncs;  /* The number of elements in FuncNames */
/* Position of the cursor (in pixels) */
int Cursx;
int Cursy;

/* Maximum screen size (in pixels) */
int Maxx=640;
int Maxy=480;

/*  For a faster k-board cursor  */
int Fast = 0;

#ifdef PARSER
void iferror(char *msg)
{
	GSelog(-1, "Error on line: %d.\nLine: |%s|\nMessage: %s.\n", 
										lineNumber, lineBuf, msg);
}
#endif

char *str_dup(char *oldstr)
{
	char *newstr = (char *) (malloc(strlen(oldstr)+1));
	if (newstr != NULL)
		strcpy(newstr, oldstr);
	return newstr;
}

/***********************************************************************
Return the index in a generic structure array whose name matches
  reqname.
GLOBALS: none.
***********************************************************************/
int getname (
  void *stuff,   /* Array to look in */
  int structlen, /* Length of an array item */
  int structnum, /* Number of items in an array */
  char *reqname, /* The string we are searching for */
  char *type)    /* A label for the type of object for error msg */
{
  char **ptr;
  int cnt;
  
  /* This construction for incrementing ptr is significant.
     If you just add the structlen, it will go up by structlen pointer
     steps (typically 2*structlen actual bytes), which is not what
     is wanted here. CAS */

  /* Uh, sure...      JBR */
  
  for(ptr=(char**)stuff, cnt=0; cnt<structnum; cnt++, ptr = (char **)(((char *) ptr)+structlen))
      if (GSstricmp(*ptr, reqname) == 0)
           return(cnt);

  if (type != NULL) /* if type is NULL, then its OK, otherwise error exit */
    GSelog (-1, "getname: Couldn't find |%s| named |%s|\n", type, reqname);
  return(-1);
}


/*------------------------------------------------------------------------------
return (logical position of) item selected in a popup menu --
i.e, the position on the list of visible menu items.
------------------------------------------------------------------------------*/
int pickitemslot (int y, int num)
{
  return ((y - Popups[num].y - TSIZE/4) / TSIZE);
}


/*------------------------------------------------------------------------------
  return the (physical position of) an item selected in a popup menu --
  i.e., the position of the item from among all items both visible or
  invisible
------------------------------------------------------------------------------*/
int pickitem (int y, int num)
{
  int slot;
  int i;

  slot = (y - Popups[num].y - TSIZE/4) / TSIZE + 1;
  for (i=0; i<Popups[num].numitems; i++)
  {
    if (Popups[num].item[i].active != -1)
      slot--;
    if (slot == 0) return (i);
  }
  return (-1);
}

/*------------------------------------------------------------------------------
 Returns origin and width and height of a button
 GLOBALS: InfState, Win
------------------------------------------------------------------------------*/
void getbuttoninfo (
  char *wname,
  char *bname,
  int *x,
  int *y,
  int *w,
  int *h,
  int *status)
{
  int i,j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  *x = Win[i].x + Win[i].wb[j].x;
  *y = Win[i].y + Win[i].wb[j].y;
  *w = Win[i].wb[j].wd;
  *h = Win[i].wb[j].ht;
  *status = Win[i].wb[j].active;
}


/**********************************************************************
Convert a string to a function pointer.
GLOBALS: FuncNames (declared by caller).
**********************************************************************/
boolean (*getfunctionptr(char *fname))()
{
  int funcnum;
/*  long nothing;*/
  char *ptr = fname;
  boolean (*retval)();
  if(fname[0] == '@')
    ptr++;
  funcnum = getname(&FuncNames, sizeof(IntrFunction), NumFuncs, ptr, "function");
  retval = FuncNames[funcnum].funcptr;
/*  nothing = (long)retval;
  nothing++;
  if (retval)
  	nothing--;*/
  return(retval);
}

static void make_triang(int t[], int x1, int y1, int x2, int y2, int x3, int y3)
{
	t[0] = x1; t[1] = y1; t[2] = x2; t[3] = y2; t[4] = x3, t[5] = y3;
}

/*------------------------------------------------------------------------------
------------------------------------------------------------------------------*/
void drawarrow(
  int wnum,
  int bnum,
  int tx,
  int twd,
  int awd,
  int aht
)
{
	int arrowcoords[6];
	int arrow_dim = MIN(awd, aht);
	int aleft, aright, atop, abottom; /* bounding box for arrow */
	int ty = Win[wnum].y + Win[wnum].wb[bnum].y, 
						tht = Win[wnum].wb[bnum].ht;
	
	aleft = tx + twd + (awd - arrow_dim)/2 + ARROWMARGIN;
	atop = ty + (aht - arrow_dim)/2 + ARROWMARGIN;
	aright = aleft + arrow_dim;
	abottom = atop + arrow_dim;
	
	switch(Win[wnum].wb[bnum].arrow)	{

		case UP_ARROW_SYM:
			make_triang(arrowcoords, (aleft+aright)/2, atop, aleft, abottom, aright, abottom);
			break;
			
		case DOWN_ARROW_SYM:
			make_triang(arrowcoords, (aleft+aright)/2, abottom, aleft, atop, aright, atop);
			break;
		
		case LEFT_ARROW_SYM:
			make_triang(arrowcoords, aleft, atop, aleft, abottom, aright, (atop+abottom)/2);
			break;
			
		case RIGHT_ARROW_SYM:
			make_triang(arrowcoords, aleft, (atop+abottom)/2, aright, atop, aright, abottom);
			break;
		}
#if 0	
	switch (Win[wnum].wb[bnum].arrow)
	{
	case UP_ARROW_SYM:  
	  arrowcoords[0] = tx + twd + awd/2 + ARROWMARGIN;      /* top point x */
	  arrowcoords[1] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   ARROWMARGIN;                         /* top point y */
	  arrowcoords[2] = tx + twd + ARROWMARGIN;              /* left point x */
	  arrowcoords[3] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* left point y */
	  arrowcoords[4] = tx + twd + awd + ARROWMARGIN;        /* right point x */
	  arrowcoords[5] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* right point y */
	  break;
	case DOWN_ARROW_SYM:  
	  arrowcoords[0] = tx + twd + awd/2 + ARROWMARGIN;      /* bottom point x */
	  arrowcoords[1] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* bottom point y */
	  arrowcoords[2] = tx + twd + ARROWMARGIN;              /* left point x */
	  arrowcoords[3] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   ARROWMARGIN;                         /* left point y */
	  arrowcoords[4] = tx + twd + awd + ARROWMARGIN;        /* right point x */
	  arrowcoords[5] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   ARROWMARGIN;                         /* right point y */
	  break;
	case LEFT_ARROW_SYM:  
	  arrowcoords[0] = tx + twd + awd + ARROWMARGIN;      /* top point x */
	  arrowcoords[1] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   ARROWMARGIN;                         /* top point y */
	  arrowcoords[2] = tx + twd + ARROWMARGIN;              /* left point x */
	  arrowcoords[3] = Win[wnum].y + Win[wnum].wb[bnum].y - awd/2 +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* left point y */
	  arrowcoords[4] = tx + twd + awd + ARROWMARGIN;        /* bottom point x */
	  arrowcoords[5] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* bottom point y */
	  break;
	case RIGHT_ARROW_SYM:  
	  arrowcoords[0] = tx + twd + ARROWMARGIN;      /* top point x */
	  arrowcoords[1] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   ARROWMARGIN;                         /* top point y */
	  arrowcoords[2] = tx + twd + awd+ ARROWMARGIN;         /* right point x */
	  arrowcoords[3] = Win[wnum].y + Win[wnum].wb[bnum].y - awd/2 +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* right point y */
	  arrowcoords[4] = tx + twd + ARROWMARGIN;        /* bottom point x */
	  arrowcoords[5] = Win[wnum].y + Win[wnum].wb[bnum].y +
	                   Win[wnum].wb[bnum].ht - ARROWMARGIN; /* bottom point y */
	  break;
	}
#endif
	GSPfillpoly(3, arrowcoords);
}

/*------------------------------------------------------------------------------
  Draw a button in a window.
  GLOBALS: Win, InfState.
------------------------------------------------------------------------------*/
void drawbutton (
  int		wnum, /*--- index for window ---*/
  int		bnum, /*--- index for button ---*/
  boolean	textonly
)
{
  int x, y, w, h;
  int xt, yt, twd, tht;
  BigString text1;
  char *text2;
  int i;
  FontSizeType oldfont;
  int oldcolor, oldlinesize;
  int awd = 0, aht = 0, amarg = 0;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();
  GSmouseoff();

  x = Win[wnum].x + Win[wnum].wb[bnum].x;
  y = Win[wnum].y + Win[wnum].wb[bnum].y;
  w = Win[wnum].wb[bnum].wd;
  h = Win[wnum].wb[bnum].ht;
  if (!textonly)
  {
    GSsetcolor(Win[wnum].wb[bnum].bgcolor);
    GSsetlinesize(THIN);
    GSPfill_rect(x, y, w, h);
    if (Win[wnum].wb[bnum].border)
    {
      GSsetcolor(InfState.buttonborder);
      GSPframe_rect(x, y, w, h);
      GSsetcolor (InfState.buttonTLcolor);
      GSPline (x+1, y+1, x+w-2, y+1);
      GSPline (x+1, y+1, x+1, y+h-2);
      GSsetcolor (InfState.buttonBRcolor);
      GSPline (x+2, y+h-2, x+w-2, y+h-2);
      GSPline (x+w-2, y+2, x+w-2, y+h-2);
      GSsetcolor(LIGHTBUTTONBORDER);
      GSPdraw_pixel (x+w-2, y+1);
      GSsetcolor(DARKBUTTONBORDER);
      GSPdraw_pixel (x+1, y+h-1);
    }
  }
  if ((!Win[wnum].wb[bnum].drag) &&
      (strcmp(Win[wnum].wb[bnum].text, NOTEXT) != 0))
  { /* Drag buttons have no label */
    if (Win[wnum].wb[bnum].active == 1)
    {
      GSsetcolor(Win[wnum].wb[bnum].efgcolor);
      GSsetfont(Win[wnum].wb[bnum].tsize);
    }
    else
    {
      GSsetcolor(Win[wnum].wb[bnum].dfgcolor);
      GSsetfont(Win[wnum].wb[bnum].tsize);
    }
    /*  break text into two lines if necessary  */
    text2 = strchr (Win[wnum].wb[bnum].text, '@');
    if (text2 == NULL)
    {
      /*  only 1 line of text in the button */
      twd = GStextwidth(Win[wnum].wb[bnum].text);
      tht = GStextheight(Win[wnum].wb[bnum].text);
      if (Win[wnum].wb[bnum].arrow != NO_ARROW_SYM)
      {
        amarg = ARROWMARGIN;
        awd = aht = Win[wnum].wb[bnum].ht - 2 * amarg;
        awd = Win[wnum].wb[bnum].wd - 2 * amarg;
      }
      switch (Win[wnum].wb[bnum].labpos)
      {
        case LEFT:
          xt = x + 5;
          break;      
        case RIGHT:
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) - 6;
          break;      
        case CENTER:
/*          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) / 2 - 1;*/
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) / 2;
          break;      
      }
#if 0
	/* why didn't the Mac like this code? */
      yt = y + ((Win[wnum].wb[bnum].ht - tht) / 2) + tht +
               ((Win[wnum].wb[bnum].ht - tht) % 2);
#else
		yt = y;
		yt += ((Win[wnum].wb[bnum].ht - tht) / 2);
		yt += tht;
		yt += ((Win[wnum].wb[bnum].ht - tht) % 2);
#endif
      /*yt = y + ((Win[wnum].wb[bnum].ht - tht) / 2) + tht + 1;*/
      GSPdraw_string(xt, yt, Win[wnum].wb[bnum].text);
      if (Win[wnum].wb[bnum].arrow != NO_ARROW_SYM)
      {
        if (twd > 0) {
          drawarrow(wnum, bnum, xt, twd, awd, aht);
	}
        else {
          drawarrow(wnum, bnum, x, twd, awd, aht);
	}
      }
    }
    else /*-- text2 != NULL --*/
    {
      text2++;
      i = 0;
      while ((Win[wnum].wb[bnum].text[i] != '@') &&
             (Win[wnum].wb[bnum].text[i] != '\0'))
      {
        text1[i] = Win[wnum].wb[bnum].text[i];
        i++;
      }
      text1[i] = '\0';

      twd = GStextwidth(text1);
      tht = GStextheight(text1);
      if (Win[wnum].wb[bnum].arrow != NO_ARROW_SYM)
      {
        amarg = ARROWMARGIN;
        awd = aht = Win[wnum].wb[bnum].ht - 2 * amarg;
        awd = Win[wnum].wb[bnum].wd - 2 * amarg;
      }
      switch (Win[wnum].wb[bnum].labpos)
      {
        case LEFT:
          xt = x + 5;
          break;
	case RIGHT:
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) - 6;
          break;      
        case CENTER:
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) / 2;
          break;
      }
/*      xt = x + (Win[wnum].wb[bnum].wd - twd) / 2 - 1;*/
      yt = (Win[wnum].wb[bnum].ht - tht - tht - 2);
      yt = (int)(yt / 2);
      yt = yt + tht;
      yt += y;
      GSPdraw_string(xt, yt, text1);
      twd = GStextwidth(text2);
      switch (Win[wnum].wb[bnum].labpos)
      {
        case LEFT:
          xt = x + 5;
          break;
	case RIGHT:
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) - 6;
          break;      
        case CENTER:
          xt = x + (Win[wnum].wb[bnum].wd - twd - awd - amarg) / 2;
          break;
      }
/*      xt = x + (Win[wnum].wb[bnum].wd - twd) / 2 - 1;*/
      yt += tht;
      GSPdraw_string(xt, yt, text2);
    }
    if (Win[wnum].wb[bnum].active == 0 &&
        InfState.diagonals &&
        Win[wnum].wb[bnum].border)
    {
      GSsetcolor(InfState.buttonborder);
      GSPline (x, y+h-1, x+w-1, y);
    }
  }
  GSmouseon();
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
}

/**********************************************************************
 Draw a depressed button in a window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSPdraw_depressed_button( int wnum,
                               int bnum)
{
  int x, y, w, h;
  FontSizeType oldfont;
  int oldcolor, oldlinesize;
  int savecurrentwin;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();

  savecurrentwin = InfState.cw;
  GSPsetcurrentwindow (wnum);
  GSsetlinesize(THIN);

  x = Win[wnum].x + Win[wnum].wb[bnum].x;
  y = Win[wnum].y + Win[wnum].wb[bnum].y;
  w = Win[wnum].wb[bnum].wd;
  h = Win[wnum].wb[bnum].ht;
  GSmouseoff();
  if (Win[wnum].wb[bnum].border)
  {
    GSsetcolor(InfState.buttonborder);
    GSPframe_rect(x, y, w, h);
    GSsetcolor (InfState.buttonBRcolor);
    GSPline (x+1, y+1, x+w-2, y+1);
    GSPline (x+1, y+1, x+1, y+h-2);
    GSsetcolor (InfState.buttonTLcolor);
    GSPline (x+2, y+h-2, x+w-2, y+h-2);
    GSPline (x+w-2, y+2, x+w-2, y+h-2);
    GSsetcolor(DARKBUTTONBORDER);
    GSPdraw_pixel (x+w-2, y+1);
    GSsetcolor(LIGHTBUTTONBORDER);
    GSPdraw_pixel (x+1, y+h-1);
  }
  GSmouseon();
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
  GSPsetcurrentwindow (savecurrentwin);
}


/**********************************************************************
 Draw a released button in a window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSPdraw_released_button( int wnum,
                               int bnum)
{
  int x, y, w, h;
  FontSizeType oldfont;
  int oldcolor, oldlinesize;
  int savecurrentwin;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();
  savecurrentwin = InfState.cw;

  GSPsetcurrentwindow (wnum);
  GSsetlinesize(THIN);

  x = Win[wnum].x + Win[wnum].wb[bnum].x;
  y = Win[wnum].y + Win[wnum].wb[bnum].y;
  w = Win[wnum].wb[bnum].wd;
  h = Win[wnum].wb[bnum].ht;
  GSmouseoff();
  if (Win[wnum].wb[bnum].border)
  {
    GSsetcolor(InfState.buttonborder);
    GSPframe_rect(x, y, w, h);
    GSsetcolor (InfState.buttonTLcolor);
    GSPline (x+1, y+1, x+w-2, y+1);
    GSPline (x+1, y+1, x+1, y+h-2);
    GSsetcolor (InfState.buttonBRcolor);
    GSPline (x+2, y+h-2, x+w-2, y+h-2);
    GSPline (x+w-2, y+2, x+w-2, y+h-2);
    GSsetcolor(LIGHTBUTTONBORDER);
    GSPdraw_pixel (x+w-2, y+1);
    GSsetcolor(DARKBUTTONBORDER);
    GSPdraw_pixel (x+1, y+h-1);
  }
  GSmouseon();
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
  GSPsetcurrentwindow (savecurrentwin);
}

/**********************************************************************
Erase a button.
GLOBALS: Win.
**********************************************************************/
void erasebutton(
  int wnum,  /* index for window */
  int bnum)  /* index for button */
{
  int x, y;

  x = Win[wnum].x + Win[wnum].wb[bnum].x;
  y = Win[wnum].y + Win[wnum].wb[bnum].y;
  GSsetcolor(Win[wnum].bgcolor);
  GSPfill_rect(x, y, Win[wnum].wb[bnum].wd, Win[wnum].wb[bnum].ht);

}


/**********************************************************************
Draw a graphics object
GLOBALS: Win, InfState.
**********************************************************************/
static void drawgraphic(Graphic *gr)
{
  int x, y;
  FontSizeType oldfont;
  int oldcolor, oldlinesize;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();

  GSsetlinesize(THIN);
  x = Win[InfState.cw].x + gr->x;
  y = Win[InfState.cw].y + gr->y;
  switch (gr->type) {
  case GR_LINE:
    GSsetcolor(gr->color);
    GSPline(x, y, Win[InfState.cw].x + gr->wd, Win[InfState.cw].y + gr->ht);
    break;

  case GR_WLINE:
    GSsetcolor(gr->color);
    GSPline(x, y, Win[InfState.cw].x + gr->wd, Win[InfState.cw].y + gr->ht);
    if (gr->x == gr->wd)
    {
      GSPline(x+1, y, x+1, Win[InfState.cw].y + gr->ht);
    }
    else
    {
      GSPline(x, y+1, Win[InfState.cw].x + gr->wd, y+1);
    }
    break;

  case GR_RECT:
    GSsetcolor(gr->color);
    if (gr->attrib == FILL)
      GSPfill_rect(x, y, gr->wd, gr->ht);
    else
      GSPframe_rect(x, y, gr->wd, gr->ht);
    break;

  case GR_WRECT:
    GSsetcolor(gr->color);
    if (gr->attrib == FILL)
      GSPfill_rect(x, y, gr->wd, gr->ht);
    else
    {
      GSPframe_rect(x, y, gr->wd, gr->ht);
      GSPframe_rect(x+1, y+1, gr->wd-2, gr->ht-2);
    }
    break;

  case GR_LABEL:
    if (gr->shadowcolor != -1)
    {  /* Write the shadow of the text down and to right by one pixel */
      GSsetcolor(gr->shadowcolor);
      GSsetfont (LARGE);
      GSPdraw_string(x+1, y+1, gr->text);
    }
    GSsetcolor(gr->color);
    GSsetfont (LARGE);
    GSPdraw_string(x, y, gr->text);
    break;

  case GR_LABEL_SMALL:
    if (gr->shadowcolor != -1)
    {  /* Write the shadow of the text down and to right by one pixel */
      GSsetcolor(gr->shadowcolor);
      GSsetfont (SMALL);
      GSPdraw_string(x+1, y+1, gr->text);
    }
    GSsetcolor(gr->color);
    GSsetfont(SMALL);
    GSPdraw_string(x, y, gr->text);
    break;
  } /* switch */
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
}


/**********************************************************************
Draw window "num" borders.
GLOBALS: Win.
**********************************************************************/
void drawwindowborders (int num, int color)
{
  int i;
  int oldcolor, oldlinesize;

  oldlinesize = GSPgetlinesize();
  oldcolor = GSgetcolor();

  GSsetlinesize(THIN);
  GSPsetcurrentwindow (num);
  GSsetcolor(color);

  GSPset_clip_rect(Win[num].x, Win[num].y, Win[num].wd, Win[num].ht);
  
  for (i = 0; i < Win[num].bordersize; i++)
  {
    GSPframe_rect (Win[num].x + i,
                   Win[num].y + i,
                   Win[num].wd - 2 * i,
                   Win[num].ht - 2 * i);     
  }
  GSPset_clip_rect(Win[num].x + Win[num].bordersize,
                   Win[num].y + Win[num].bordersize,
                   Win[num].wd - 2 * Win[num].bordersize,
                   Win[num].ht - 2 * Win[num].bordersize);

  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
}

/**********************************************************************
Draw window "num".
GLOBALS: Win, InfState.
**********************************************************************/
void drawwindow(int num)
{
  int i;
  int oldcolor;

  oldcolor = GSgetcolor();

  GSPsetcurrentwindow(num);
  GSsetcolor(Win[num].bgcolor);
  GSPfill_rect(Win[num].x, Win[num].y, Win[num].wd, Win[num].ht);
  drawwindowborders (num, Win[num].bordercolor);
  for (i=0; i < Win[num].numbuttons; i++)
    if (Win[num].wb[i].active != -1)
      drawbutton (num, i, FALSE);
  for (i=0; i<Win[num].numgraphics; i++) /* Put up the graphic objects */
    drawgraphic(&(Win[num].graphics[i]));

  GSsetcolor(oldcolor);
}


/**********************************************************************
Remove window with index winnum.
GLOBALS: Win, InfState.
**********************************************************************/
void removewindow(int winnum)
{
  int i;

  GSPsetcurrentwindow(0);
  if (Win[winnum].active)
  {
    if (Win[winnum].bstore == BSTORE) {
	  InfState.backingstore[Win[winnum].bstore_to_use].inuse = FALSE;
      if (!GSPfrom_bstore(Win[winnum].bstore_to_use, Win[winnum].x,
          Win[winnum].y))
        GSelog (-1, "Error restoring backing store for window %s\n",
                Win[winnum].name);
	  }

    if(Win[winnum].popup)
    {
      InfState.level--;
      for (i = 0; i < InfState.numwindows; i++)
      {
        if (Win[i].active == InfState.level)
        {
          enable_buttons_after_popup(i);
        }
      }
    }
    Win[winnum].active = 0;
  }
}

/**********************************************************************
Display one label from a popup menu item.
**********************************************************************/
void display_popup_label(int popnum, int itemnum, int slotnum)
{
  int offset = POPUPOFFSET;
  FontSizeType oldfont; 
  int oldcolor;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();

  if (Popups[popnum].checklist)
    offset += CHECKWID;
  if (Popups[popnum].item[itemnum].active == 1)
    GSsetcolor(Popups[popnum].enablefg);
  else
    GSsetcolor(Popups[popnum].disablefg);
  GSsetfont(LARGE);

  GSPdraw_string(Popups[popnum].x + offset + 1,
                 Popups[popnum].y + slotnum*TSIZE + TSIZE,
                 Popups[popnum].item[itemnum].text);
  if (Popups[popnum].item[itemnum].checked)
  {
    if (Popups[popnum].item[itemnum].active == 1)
      GSsetcolor(Popups[popnum].enablefg);
    else
      GSsetcolor(Popups[popnum].disablefg);
    GSsetfont(LARGE);
    /* Kludge for now with "+" -- CAS */
    GSPdraw_string(Popups[popnum].x + POPUPOFFSET + 1,
                   Popups[popnum].y + slotnum*TSIZE + TSIZE, CHECKMARK);
  }
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
}

/**********************************************************************
Bring up a popup menu.
GLOBALS: Popups, InfState.
**********************************************************************/
void createpopupmenu(int num)
{
  int i;
  int numactive = 0;
  int slot = 0;  /* Physical position of current item on display list */
  FontSizeType oldfont;
  int oldcolor, oldlinesize;

  oldfont = GSPgetfont();
  oldcolor = GSgetcolor();
  oldlinesize = GSPgetlinesize();

  GSsetlinesize(THIN);
  PopUpSaveCurrWinNum = InfState.cw;
  GSsetcurrentnamedwindow("root");

  InfState.activepopupwindow = num;
  InfState.level++;
  Popups[num].active = TRUE;
  for (i=0; i<Popups[num].numitems; i++)
    if (Popups[num].item[i].active != -1) numactive++;
  Popups[num].ht = TSIZE/2 + numactive * TSIZE;

  if (InfState.backingstore[POPUPBSTORE].inuse)
	GSelog(-1, "Error - bstore %d is already occupied.\n", POPUPBSTORE);
  InfState.backingstore[POPUPBSTORE].inuse = TRUE;
  if (!GSPto_bstore(POPUPBSTORE, Popups[num].x, Popups[num].y, Popups[num].wd,
      Popups[num].ht))
    GSelog (-1, "Error putting Popupwindow into backing store: %d %d %d %d\n",
            Popups[num].x, Popups[num].y, Popups[num].wd, Popups[num].ht);

  GSsetcolor(Popups[num].bg);
  GSPfill_rect(Popups[num].x, Popups[num].y, Popups[num].wd, Popups[num].ht);
  GSsetcolor(Popups[num].bcolor);
  GSPframe_rect(Popups[num].x, Popups[num].y, Popups[num].wd, Popups[num].ht);
  for (i=0; i < Popups[num].numitems; i++)
  {
    if (Popups[num].item[i].active != -1)
    {
      display_popup_label(num, i, slot);
      slot++;
    }
  }
  GSsetfont(oldfont);
  GSsetcolor(oldcolor);
  GSsetlinesize(oldlinesize);
}

/**********************************************************************
Remove a popup menu (and restore the screen under it).
GLOBALS: Popups, InfState.
**********************************************************************/
void removepopup(void)
{
  InfState.level--;
  GSPsetcurrentwindow(0);
  Popups[InfState.activepopupwindow].active = FALSE;
  if (!GSPfrom_bstore(POPUPBSTORE, Popups[InfState.activepopupwindow].x,
                      Popups[InfState.activepopupwindow].y))
    GSelog (-1, "Error restoring backing store for popupmenu\n");
  InfState.backingstore[POPUPBSTORE].inuse = FALSE;
  InfState.activepopupwindow = -1;
  GSdisplayhelpmessage(PopupSaveMsg);
  GSPsetcurrentwindow(PopUpSaveCurrWinNum);
}


/**********************************************************************
Convert a color name to a color index. (EGA/VGA only)
GLOBALS: None.
**********************************************************************/
GSColor convertcolor (
 char *alias[],
 char *acolor[],
 char *color
)
{
  int currcolor;

  currcolor = GSiselword(color, Colorlist);
  if (currcolor == -1)
    {
    currcolor = GSiselword(color, alias);
    if (currcolor != -1)
      currcolor = GSiselword(acolor[currcolor], Colorlist);
    if (currcolor == -1)
      {
      /*  user specified invalid color  */
      GSelog (-1, "convertcolor: color alias not found.\n\tAlias = |%s|\n",
              color);
      }
    }
  return((GSColor)currcolor);
}


/**********************************************************************
Get the next line.  Skip empty lines and comment lines.
GLOBALS: None.
**********************************************************************/
static boolean getline (
  HugeString str,   /* string */
  FILE      *fptr   /* file to read */
)
{
  char *ptr;

  do
  {
    if(fgets(str, HUGESTRINGLEN+1, fptr) == NULL)
      return(FALSE);
    if(str[strlen(str)-1] != '\n')
    {
      GSelog (-1, "Line too long in interface file\n\t|%s|\n", str);
      /* this generates an unnecessary error if the 
      interface file doesn't end with '\n'. JBR
      */
    }
    /* skip spaces in preparation for next line */
    for(ptr=str; (*ptr == ' ') || (*ptr == '\t'); ptr++);
  } while ((*ptr == '\n') || (*ptr == '#'));
  return(TRUE);
}

void resolve_specials (
  char *string,
  boolean (*func)(char, void*),
  void *data
)
{
  char *p1;
  char *p2;
  boolean delete_char;

  for (p1 = string; *p1; p1++)
  {
    if (*p1 == ESC_SYM)
    {
      for (p2 = p1; *p2; p2++)
      	*p2 = *(p2+1); /*--- remove symbol ---*/
      delete_char = (*func)(*p1, data);      /*--- process next char ---*/
      if (delete_char)
      	for (p2 = p1; *p2; p2++)
      		*p2 = *(p2+1); /*--- remove symbol ---*/
    }
  }
}

boolean button_text_specials (
  char special,
  void *butt
)
{
  switch (special)
  {
    case UP_ARROW_SYM:
      ((GSButton *)butt)->arrow = UP_ARROW_SYM;
      break;
    case DOWN_ARROW_SYM:
      ((GSButton *)butt)->arrow = DOWN_ARROW_SYM;
      break;
    case LEFT_ARROW_SYM:
      ((GSButton *)butt)->arrow = LEFT_ARROW_SYM;
      break;
    case RIGHT_ARROW_SYM:
      ((GSButton *)butt)->arrow = RIGHT_ARROW_SYM;
      break;
    default:
    	return FALSE;
  }
  return TRUE;
}

/*-----------------------------------------------------------------------------
  dobutton.
  Obsolete with parser
-----------------------------------------------------------------------------*/
#ifndef PARSER
static void dobutton(
  BigString linestr,
  int  currwin,
  int currbutt,
  FontSizeType size,
  boolean drag,
  char *alias[],
  char *acolor[]
)
{
  GSButton *butt;
  BigString tmpstr, tmpstr2, tmpstr3;
  BigString fg, color, bg;
  int k;
  int numfields;
  int tmpint;
  GSString labpos;

  if (currwin == -1)
    GSelog (-1, "No active window for buttons.\n");
  butt = &(Win[currwin].wb[currbutt]);
  butt->repeatable = FALSE;
  butt->arrow = NO_ARROW_SYM; /*--- to start off ---*/
  if (drag)
  {
    if (sscanf(linestr, "%*s ( %s %d %d %d %d %s %s %d )",
               tmpstr, &(butt->x), &(butt->y), &(butt->wd), &(butt->ht), bg,
               tmpstr3, &tmpint) != 8)
    {
      GSelog (-1, "Invalid drag area spec.\n\t|%s|\n", linestr);
    }
    butt->active = (ButtonStatusType)tmpint;
    butt->border = FALSE;
  }
  else /*--- NOT drag ---*/
  {
    numfields = sscanf(linestr,
                       "%*s ( %s |%[^|]| %d %d %d %d %s %s %s %s %d %d %d %s )",
                       tmpstr, tmpstr2, &(butt->x), &(butt->y), &(butt->wd),
                       &(butt->ht), fg, color, bg, tmpstr3, &(butt->border),
                       &tmpint, &(butt->repeatable), &labpos);
    butt->active = (ButtonStatusType)tmpint;
    if (numfields != 14)
    {
      GSelog (-1, "Invalid button spec.\n\t|%s|\n", linestr);
    }
    butt->labpos = (BtnLabPosType) labpos[0];
  }
  if (strlen(tmpstr) > INFNAMELEN)
    GSelog (0, "Name too long: |%s|\n", tmpstr);
  if ((butt->name = (char *) malloc(strlen(tmpstr)+1)) == NULL)
    GSelog (-1, "Out of memory for button name - |%s|.\n", tmpstr);
  else
    strcpy(butt->name, tmpstr);
  if (!drag)
  {
    if (strlen(tmpstr2) > INFNAMELEN)
      GSelog (0, "Label too long: |%s|\n", tmpstr2);
    resolve_specials(tmpstr2, button_text_specials, (void *)butt);
    if ((butt->text = (char *) malloc(strlen(tmpstr2)+1)) == NULL)
      GSelog (-1, "Out of memory for button label - |%s|.\n", tmpstr2);
    else
      strcpy(butt->text, tmpstr2);
    butt->tsize = size;
  }
  else
    butt->text = NULL;
  butt->func = (boolean(*)(int, int, int, ...))getfunctionptr(tmpstr3);
  if (!drag)
  {
    butt->efgcolor = convertcolor(alias, acolor, fg);
    butt->dfgcolor = convertcolor(alias, acolor, color);
  }
  else
  {
    butt->efgcolor = NOCOLOR;
    butt->dfgcolor = NOCOLOR;
  }
  butt->bgcolor = convertcolor(alias, acolor, bg);
  if (tmpstr3[0] == '@')
  { /* linked to a popup menu */
    k = getname(Popups, sizeof (Popupmenu), InfState.numpopups, tmpstr3+1,
                "popup");
    butt->popupmenunum = k;
  } /* if (tempstr3... */
  else
    butt->popupmenunum = -1;
  butt->drag = drag;
  butt->depressed = FALSE;      /* button is not initially drawn depressed */
}
#endif

/**********************************************************************
Pass 1 for the interface file reader.  This pass counts the number of
objects and mallocs appropriate space.
GLOBALS: None.
**********************************************************************/
void readfilepass1(BigString InterfaceFile)
{
  FILE *fptr;
  int cnt;
#ifdef PARSER
  msgs = 0;
  aliases = 0;
  functions = 0;
  windows = -1;
  popups = -1;
  fields = 0;
  lineNumber = 0;
#else
  int msgs = 0;
  int aliases = 0;
  int functions = 0;
  int windows = -1;
  int windata[MAXWINDOW][2]; /* windata[i][0] is number of buttons,
                                windata[i][1] is number of graphics */
  int popups = -1;
  int popdata[MAXPOPUP]; /* Number of plabels in each popup */
  int fields = 0;
  HugeString linestr;   /* A line read from the file */
  HugeString token;     /* Next token from the current line */
#endif
#ifdef PARSER
  if ((fptr = GSopen(InterfaceFile, "r")) == NULL)
    GSelog (-1, "Cannot open interface file |%s|.\n", InterfaceFile);

  /*ifin = fptr;*/
  ifrestart(fptr);
  firstPass = TRUE;
  ifparse();
#else
#endif
  windows++; /* started at -1 to get index right */
  popups++;  /* started at -1 to get index right */
  fclose(fptr);

  /* Now, malloc the necessary space */
  InfState.numfuncs = functions;
  if (functions)
  {
    if ((InfState.funcs = (Function *)malloc(sizeof (Function) *
                      (long)functions)) == NULL)
    GSelog(-1, "Out of memory for functions\n");
  }
  InfState.nummsgs = msgs;
  if (msgs)
  {
    if ((InfState.msgs = (Message *)malloc(sizeof (Message) * (long)msgs))
        == NULL)
    GSelog(-1, "Out of memory for messages\n");
  }
  InfState.numpopups = popups;
  if (popups)
  {
    if ((Popups = (Popupmenu *)malloc(sizeof (Popupmenu) * (long)popups))
        == NULL)
    GSelog(-1, "Out of memory for popup menus\n");
    for (cnt=0; cnt<popups; cnt++)
    {
      Popups[cnt].numitems = popdata[cnt];
      /* Need an extra Menuitem for canceling menu */
      if ((Popups[cnt].item = (Menuitem *)malloc(sizeof (Menuitem) *
                                   (long)(popdata[cnt]))) == NULL)
        GSelog(-1, "Out of memory for popup menu labels\n");
    }
  }
  InfState.numwindows = windows;
  if (windows)
  {
    if ((Win = (GSWindow *)malloc(sizeof (GSWindow) * (long)windows)) == NULL)
      GSelog(-1, "Out of memory for windows\n");
    for (cnt=0; cnt<windows; cnt++)
    {
      Win[cnt].numbuttons = windata[cnt][0];
      if (windata[cnt][0]) /* if there are buttons */
        if ((Win[cnt].wb = (GSButton *)malloc(sizeof (GSButton) *
                                  (long)(windata[cnt][0]))) == NULL)
          GSelog(-1, "Out of memory for buttons\n");
      Win[cnt].numgraphics = windata[cnt][1];
      if (windata[cnt][1]) /* if there are graphics */
        if ((Win[cnt].graphics = (Graphic *)malloc(sizeof (Graphic) *
                                  (long)(windata[cnt][1]))) == NULL)
          GSelog(-1, "Out of memory for graphics\n");
    }
  }
  InfState.numfields = fields;
  if (fields)
  {
    if ((InfState.fields = (Field *)malloc(sizeof (Field) * (long)fields))
        == NULL)
    GSelog(-1, "Out of memory for fields\n");
  }
  /* initialize application key handler func to NULL */
  InfState.keyfunc = NULL;
  /* initialize application redraw func to GSredrawactivewindows */
  InfState.redrawfunc = GSredrawactivewindows;
}

/**********************************************************************
Initialize palette with standard and default colors.
GLOBALS: none
**********************************************************************/
void init_default_palette(char *Colorlist[], char Interface_Palette[])
{
	/* Copy color names and rgb values from default arrays;
		allocated dynamically so that they can be free()d.
	*/
	for (currcolor = 0; currcolor < NUM_STD_COLORS; currcolor++)
		Colorlist[currcolor] = str_dup(Std_Colorlist[currcolor]);

	for(;currcolor < NUM_STD_COLORS + NUM_DFLT_COLORS; currcolor++)
		Colorlist[currcolor] = str_dup(Dflt_Colorlist[currcolor-NUM_STD_COLORS]);
		
	Colorlist[currcolor] = NULL;
	memcpy(Interface_Palette, dflt_rgb, currcolor*3);
}

/**********************************************************************
Read the interface file.
GLOBALS: None.
**********************************************************************/
void readinterfacefile(BigString InterfaceFile)
{
  FILE       *fptr;                    /* File pointer to interface file */
#ifdef PARSER
  currmsg = -1;             /* Index for current message */
  currpop = -1;             /* Index for current popup menu */
  currplab;                 /* Index for current label in current
                                          popup menu */
  currwin = -1;             /* Index for current window */
/*  currbutt;                  Index for current button in current
                                          window */
/*  currgraphic;               Index for current graphic in current
                                          window */
  currfield = -1;           /* Index for current field */
  currfunc = -1;            /* Index for current function */
  /*int        p;*/
  clist = 0;  
  curr_alias = 0;           /* Why not -1? CAS *//* Index for current alias*/
  lineNumber = 0;
#else
  HugeString linestr;                  /* A line read from the file */
  BigString  token;                    /* Next token from the current line */
  BigString  tmpstr, tmpstr2, tmpstr3; /* Read simulation colors */
  BigString  color;                    /* The name of a color */
  BigString  fg;                       /* Foreground color name */
  BigString  fg2;                      /* Another foreground color name */
  BigString  bg;                       /* Background color */
  int        currmsg = -1;             /* Index for current message */
  int        currpop = -1;             /* Index for current popup menu */
  int        currplab;                 /* Index for current label in current
                                          popup menu */
  int        currwin = -1;             /* Index for current window */
  int        currbutt;                 /* Index for current button in current
                                          window */
  int        currgraphic;              /* Index for current graphic in current
                                          window */
  int        currfield = -1;           /* Index for current field */
  int        currfunc = -1;            /* Index for current function */
  int        len;                      /* Text string length */
  int        width;                    /* Line width */
  Field      *myfield;                 /* Current Field */
  Message    *msg;                     /* Current message */
  Graphic    *mygraphic;               /* Current Graphic */
  Function   *myfunc;                  /* Current Function */
  int        clist = 0;
  int        d1, d2, d3;
  int        p;
  int        returnvalue;              /* value returned by sscanf or other
                                          functs as needed */
  int        curr_alias = 0;           /* Why not -1? CAS *//* Index for current
                                                                alias */
  int		 tempactive;		/* This is needed to read in a menu's active status properly. JBR */
  FontSizeType oldfont;
#endif

  if ((fptr = GSopen(InterfaceFile, "r")) == NULL)
    GSelog (-1, "Cannot open interface file |%s|.\n", InterfaceFile);

#ifdef PARSER
  /*ifrestart(fptr);*/
  ifrestart(fptr);
  firstPass = FALSE;
  init_default_palette(Colorlist, Interface_Palette);
  ifparse();	
#else
 /* see scanfs.c */
#endif
  fclose(fptr);
}


/**********************************************************************
A dummy function used as target for no operation by interface element.
**********************************************************************/
static boolean nothing(int item)
{
  item = item; /*--- to appease the great compiler ---*/
  return(TRUE);
}


/**********************************************************************
Assign cancel menu items to popups.
**********************************************************************/
void assignpopups(void)
{
  int i, j;
  int len;

  return;

  /*for(i=0; i<InfState.numpopups; i++)
  { 
    j = Popups[i].numitems;
    Popups[i].item[j].active = 1;
    Popups[i].item[j].func = nothing;
    Popups[i].item[j].checked = FALSE;
    if ((Popups[i].item[j].text = malloc(strlen("Close Menu")+1)) == NULL)
      GSelog (-1, "Out of memory for menu close item label.\n");
    else
      strcpy(Popups[i].item[j].text, "Close Menu");
    len = GStextwidth("Close Menu")+2*POPUPOFFSET;
    if (Popups[i].checklist)
      len += CHECKWID;
    if (len > Popups[i].wd)
      Popups[i].wd = len;
    Popups[i].numitems++;
  }
  return; */
}


/**********************************************************************
Check if cursor is attempting to move beyond limits of entire screen.
GLOBALS: Cursx, Cursy, Maxx, Maxy.
**********************************************************************/
#define cursorlimits() ((Cursx <= (Maxx - CURSIZE/2 - 1)) &&\
                        (Cursx >= (CURSIZE/2)) && (Cursy >= (CURSIZE/2)) &&\
                        (Cursy <= (Maxy - CURSIZE/2 - 1)))



/*------------------------------------------------------------------------------
  Highlight a menu item if mouse is presently in it.  Also unhighlight
  an item when the mouse leaves.
  GLOBALS: Cursx, Cursy.
------------------------------------------------------------------------------*/
void highlight(int cx, int cy)
{
  static int oldpopup     = -1;  /*--- Popup window last time here. ---*/
  static int olditemslot  = -1;  /*---  Popup item last time here. ---*/
  static int olditem      = -1;  /*---  Popup item last time here. ---*/
         int curritem     = -1;
         int curritemslot = -1;
         int num;
         int i;
         int numactive    = 0;

  num = InfState.activepopupwindow;
  for (i=0; i<Popups[num].numitems; i++)
  {
    if (Popups[num].item[i].active != -1)
    {
      numactive++;
    }
  }
  if ((cx > Popups[num].x) && (cx < (Popups[num].x + Popups[num].wd)) &&
      (cy > Popups[num].y+1) &&
      (cy < (Popups[num].y + numactive*TSIZE + TSIZE/4 - 1)))
  {
    curritemslot = pickitemslot(cy, num);
    curritem = pickitem(cy, num);
  }
  else
  {
    curritemslot = -1;
    curritem = -1;
  }

  if ((num == oldpopup) && (curritemslot != olditemslot) && (olditemslot != -1))
  { /* clear the highlight from the old item */
    if (Popups[num].item[olditem].active != -1)
    {
      GSmouseoff();
      GSsetcolor(Popups[num].bg);
      GSPfill_rect(Popups[num].x+1, Popups[num].y + olditemslot*TSIZE+TSIZE/4,
                   Popups[num].wd-2, TSIZE);
      display_popup_label(num, olditem, olditemslot);
      GSmouseon();
    }
  }
  if ((curritemslot != -1) && (curritemslot != olditemslot))
  { /*--- highlight the new item ---*/
    if (Popups[num].item[curritem].active != -1)
    {
      GSmouseoff();
      GSsetcolor(InfState.highlight);
      GSPfill_rect(Popups[num].x+1, Popups[num].y + curritemslot*TSIZE+TSIZE/4,
                   Popups[num].wd-2, TSIZE);
      display_popup_label(num, curritem, curritemslot);
      GSmouseon();
    }
  }
  olditem = curritem;
  olditemslot = curritemslot;
  oldpopup = num;
}

/***********************************************************************
Process an arrow key.
GLOBALS: Cursx, Cursy.
***********************************************************************/
void process_arrow(int ch)
{
  int oldx, oldy;
  boolean move = FALSE;
  int step;

  oldx = Cursx;
  oldy = Cursy;

  if (Fast)
    step = CURSOR_RESO * 5;
  else
    step = CURSOR_RESO;

  while (step--)
  {
    switch (ch)
      { /* Process an arrow key */
      case KeyUp:
        Cursy -= 1;
        move = TRUE;
        break;
      case KeyDown:
        Cursy += 1;
        move = TRUE;
        break;
      case KeyLeft:
        Cursx -= 1;
        move = TRUE;
        break;
      case KeyRight:
        Cursx += 1;
        move = TRUE;
      break;
      } /* switch */

    if (move)
      if (cursorlimits())
        set_mouse_position(Cursx, Cursy);
      else
        { /* tried to move cursor off screen */
        Cursx = oldx;
        Cursy = oldy;
        break;
        }
    }
}


/***********************************************************************
Set the last control field of InfState
GLOBALS: InfState.
***********************************************************************/
void setlastcontrol(ControlGroupType type,
                    char *groupname,
                    char *controlname,
                    int controlnum)
{
  static BigString logstring;
  /*static double    k = 1000. / GS_CLOCKS_PER_SEC;*/

  static time_t t; /* ANSI C time struct */

  InfState.lastcontrol.type = type;
  strcpy(InfState.lastcontrol.gname, groupname);

  time(&t);
  strftime(InfState.lastcontrol.time, BIGSTRINGLEN, "%H:%M:%S", localtime(&t));

  switch (type)
  {
    case NONE :
      InfState.lastcontrol.control.number = -1;
      break;
    case WINDOW :
      strcpy(InfState.lastcontrol.control.name, controlname);
#ifdef SHOW_CONTROL_TIMES
      sprintf (logstring, "Window: %s -- Button: %s -- Time: %s -- ms: %ld\n",
               InfState.lastcontrol.gname, InfState.lastcontrol.control.name,
               InfState.lastcontrol.time, (long) (1000*clock())/CLK_TCK );
      GSelog(0, logstring);
#endif
      break;
    case MENU :
      InfState.lastcontrol.control.number = controlnum;
#ifdef SHOW_CONTROL_TIMES
      sprintf (logstring, "Menu: %s -- Item Num: %d -- Time: %s -- ms: %ld\n",
               InfState.lastcontrol.gname, InfState.lastcontrol.control.number,
               InfState.lastcontrol.time, (long)k*clock());
      GSelog(0, logstring);
#endif
      break;
  }
}



void writepalettefile (char *rgb)
 {
 FILE *fptr;
 int i;

 if ((fptr = GSopen ("palette.txt", "wb")) == NULL)
   GSelog (0, "Error opening palette file for writing.\n");
 else
   {
   for (i=0; i < 16; i++)
     fprintf (fptr, "%2d  %2d  %2d\n", rgb[i*3], rgb[i*3+1], rgb[i*3+2]);
   fclose (fptr);
   }
 }


/**********************************************************************
Toggle checkmark on checklist menu "i", item "itemnum".
**********************************************************************/
void toggle_check(int popnum, int itemnum, GSColor tcolor)
{
  int itemnumslot;  /* The physical slot for itemnum */
  int temp;

  Popups[popnum].item[itemnum].checked = !Popups[popnum].item[itemnum].checked;
  if ((popnum == InfState.activepopupwindow) &&
      (Popups[popnum].item[itemnum].active != -1))
  { /* This item is visible - toggle its checkmark on the screen */
    for (itemnumslot=0, temp=-1; temp<itemnum; itemnumslot++, temp++)
      if (Popups[popnum].item[itemnumslot].active == -1) /* invisible */
        temp--;
    itemnumslot--;
    if (pickitemslot(Cursy, popnum) == itemnumslot) /* This is current item */
      tcolor = InfState.highlight;                  /* so hightlight it */
    GSsetcolor(tcolor);
    GSPfill_rect(Popups[popnum].x+1, Popups[popnum].y+itemnumslot*TSIZE+TSIZE/4,
             Popups[popnum].wd-2, TSIZE);
    display_popup_label(popnum, itemnum, itemnumslot);
  }
}


/**********************************************************************
  Set the current window for drawing to winnum.
  set the clip rect to this window.
  GLOBALS: InfState
**********************************************************************/
void GSPsetcurrentwindow(int winnum)
{
  InfState.cw = winnum;
  GSPset_clip_rect(Win[winnum].x + Win[winnum].bordersize,
                   Win[winnum].y + Win[winnum].bordersize,
                   Win[winnum].wd - 2 * Win[winnum].bordersize,
                   Win[winnum].ht - 2 * Win[winnum].bordersize);
}

/**********************************************************************
  Temporarily disable buttons in a window behind a popup window.
  GLOBALS: InfState Win
**********************************************************************/
void disable_buttons_for_popup(int winnum)
{
  int butnum;
  int numbuttons;
  int currwin;

  currwin =  InfState.cw;
  numbuttons = Win[winnum].numbuttons;
  GSPsetcurrentwindow(winnum);
  for (butnum = 0; butnum < numbuttons; butnum++)
  {
    if (Win[winnum].wb[butnum].active == ENABLED &&
        !Win[winnum].wb[butnum].drag)
    {
      Win[winnum].wb[butnum].active = TMP_DISABLED;
      drawbutton(winnum, butnum, TRUE);
    }
  }
  GSPsetcurrentwindow(currwin);
}

/**********************************************************************
  Re-enable buttons in a window behind a popup window when the popup
  window is removed.
  GLOBALS: InfState Win
**********************************************************************/
void enable_buttons_after_popup(int winnum)
{
  int butnum;
  int numbuttons;
  int currwin;

  currwin =  InfState.cw;
  numbuttons = Win[winnum].numbuttons;
  GSPsetcurrentwindow(winnum);
  for (butnum = 0; butnum < numbuttons; butnum++)
  {
    if (Win[winnum].wb[butnum].active == TMP_DISABLED &&
        !Win[winnum].wb[butnum].drag)
    {
      Win[winnum].wb[butnum].active = (ButtonStatusType) 1;
      drawbutton(winnum, butnum, TRUE);
    }
  }
  GSPsetcurrentwindow(currwin);
}


/*------------------------------------------------------------------------------
  drawshadedframe: draws a highlighted frame around field.
------------------------------------------------------------------------------*/
void drawshadedframe (
  int		x,
  int		y,
  int		wd,
  int		ht,
  GSColor	tlcolor,
  GSColor	brcolor
)
{
  int		save_line_size;
  GSColor	save_color;

  save_color = GSgetcolor();
  GSsetcolor(tlcolor);
  save_line_size = GSPgetlinesize();
  GSsetlinesize(THIN);
  GSwdrawline(x-1,    y-1,    x+wd,   y-1);
  GSwdrawline(x-2,    y-2,    x+wd+1, y-2);
  GSwdrawline(x-1,    y-1,    x-1,    y+ht);
  GSwdrawline(x-2,    y-2,    x-2,    y+ht+1);
  GSsetcolor(brcolor);
  GSwdrawline(x,      y+ht,   x+wd,   y+ht);
  GSwdrawline(x-1,    y+ht+1, x+wd+1, y+ht+1);
  GSwdrawline(x+wd,   y,      x+wd,   y+ht);
  GSwdrawline(x+wd+1, y-1,    x+wd+1, y+ht+1);
  GSsetlinesize(save_line_size);
  GSsetcolor(save_color);
}
