/**********************************************************************
 GeoSim standard interface file interpreter and interface handler.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>

#include "geosim.h"
#include "gspriv.h"

#define INIT_SPEEDUP 1        /*  initial speed up for mouse button press */

/***********************************************************************
GLOBAL VARIABLES
***********************************************************************/

short    Etester = 0x0100; /* for "endian" testing */

/* Should these be static?? */
char *bm1 = NULL;  /*  backing store for popup palette editor */
char *bm2 = NULL;  /*  backing store for popup palette editor */


/*static clock_t    Start_time;  */ /* for timer functions */

extern char *Colorlist[];
extern char Interface_Palette[];
extern char *Alias[MAXALIASES+1];
extern char *Acolor[MAXALIASES+1];

extern GSWindow *Win;
extern boolean AlertUp;
extern Iconfig InfState;
extern Popupmenu *Popups;
extern int Cursx;
extern int Cursy;
extern int Maxx;
extern int Maxy;
extern int PopupSaveMsg;

/*--------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
  GSquit
    Quit the program and write a message to the stdout/error log 
------------------------------------------------------------------------------*/
void GSquit(char *format, ...)
{
  HugeString tmpstr;
  va_list argptr;

  GSPclose_events();

  if (format == NULL)
  {
	closegraph ();
	exit(0);
  }
  else
  {
	va_start(argptr, format);
	vsprintf (tmpstr, format, argptr);
	va_end(argptr);
	GSelog (-1, tmpstr);
  }
}

/***********************************************************************
 Return the cursor coordinates relative to the current window.
 GLOBALS: Cursx, Cursy, Win.
***********************************************************************/
void GSgetcursorxy (int *x, int *y)
{
  *x = Cursx - Win[InfState.cw].x;
  *y = Cursy - Win[InfState.cw].y;
}

/***********************************************************************
 Return the color at the requested index.  The caller only needs to know
 the index for the color on the interface file's color list.
 This is not a macro because the caller shouldn't know about InfState.
 GLOBALS: InfState.
***********************************************************************/
GSColor GSgetlistcolor(int lnum, int num)
{
  return (InfState.colorlist[lnum][num]);
}

/*------------------------------------------------------------------------------
 Returns origin (window relative) of a button
 GLOBALS: InfState, Win
------------------------------------------------------------------------------*/
void GSgetbuttonspecs (
  char			*wname,
  char			*bname,
  int			*x,
  int			*y,
  int			*wd,
  int			*ht,
  ButtonStatusType	*status
)
{
  int i,j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  if (x)
    *x = Win[i].wb[j].x;
  if (y)
    *y = Win[i].wb[j].y;
  if (wd)
    *wd = Win[i].wb[j].wd;
  if (ht)
    *ht = Win[i].wb[j].ht;
  if (status)
    *status = Win[i].wb[j].active;
}


/*------------------------------------------------------------------------------

------------------------------------------------------------------------------*/
void GSsetdragareaspecs (
  char			*wName,
  char			*daName,
  int			*x,
  int			*y,
  int			*wd,
  int			*ht,
  GSColor		*bgColor
)
{
  int	i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wName, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, daName, "button");
  if (x)
    Win[i].wb[j].x = *x;
  if (y)
    Win[i].wb[j].y = *y;
  if (wd)
    Win[i].wb[j].wd = *wd;
  if (ht)
    Win[i].wb[j].ht = *ht;
  if (bgColor)
    Win[i].wb[j].bgcolor = *bgColor;
}


/*------------------------------------------------------------------------------

------------------------------------------------------------------------------*/
void GSsetbuttonspecs (
  char		*wName,
  char		*bName,
  int		*x,
  int		*y,
  int		*wd,
  int		*ht,
  GSColor	*bgColor
)
{
  int	i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wName, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bName, "button");
  if (x)
    Win[i].wb[j].x = *x;
  if (y)
    Win[i].wb[j].y = *y;
  if (wd)
    Win[i].wb[j].wd = *wd;
  if (ht)
    Win[i].wb[j].ht = *ht;
  if (bgColor)
    Win[i].wb[j].bgcolor = *bgColor;
}

/*------------------------------------------------------------------------------
Added 8/27/95 - JBR
------------------------------------------------------------------------------*/
void GSsetbuttontext (
	char	*wName,
	char	*bName,
	char	*newText
)
{
  int	i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wName, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bName, "button");

  free (Win[i].wb[j].text);
  Win[i].wb[j].text = str_dup(newText);
 }	

/*------------------------------------------------------------------------------

------------------------------------------------------------------------------*/
void GSsetfieldrect (
  char			*fName,
  int			*x,
  int			*y,
  int			*wd,
  int			*ht,
  GSColor		*color
)
{
  int	i;
  Field *fld;

  i = getname(InfState.fields, sizeof (Field), InfState.numfields, fName,
              "field");
  fld = &(InfState.fields[i]);
  if (x)
    fld->x = *x;
  if (y)
    fld->y = *y;
  if (wd)
    fld->wd = *wd;
  if (ht)
    fld->ht = *ht;
  if (color)
    fld->color = *color;
}


/*------------------------------------------------------------------------------
 Returns origin (window relative) of a button
 GLOBALS: InfState, Win
------------------------------------------------------------------------------*/
void GSgetdragareaspecs (
  char			*wname,
  char			*bname,
  int			*x,
  int			*y,
  int			*wd,
  int			*ht,
  DragAreaStatusType	*status,
  GSColor		*bgcolor
)
{
  int i,j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  if (x)
    *x = Win[i].wb[j].x;
  if (y)
    *y = Win[i].wb[j].y;
  if (wd)
    *wd = Win[i].wb[j].wd;
  if (ht)
    *ht = Win[i].wb[j].ht;
  if (status)
    *status = Win[i].wb[j].active;
  if (bgcolor)
    *bgcolor = Win[i].wb[j].bgcolor;
}

/*------------------------------------------------------------------------------
 Returns origin of a window
 GLOBALS: InfState, Win
------------------------------------------------------------------------------*/
void GSgetwindowspecs (
  char *wname,
  int  *x,
  int  *y,
  int  *wd,
  int  *ht,
  int  *active
)
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");

  if (x)
    *x = Win[i].x;
  if (y)
    *y = Win[i].y;
  if (wd)
    *wd = Win[i].wd;
  if (ht)
    *ht = Win[i].ht;
  if (active)
    *active = Win[i].active;
}


/***********************************************************************
Activate a registered function to be called from inside event loop
***********************************************************************/
void GSactivatefunction(char *name)
{
  Function *func;
  int i;

  i = getname(InfState.funcs, sizeof (Function), InfState.numfuncs,
              name, "function");
  func = &(InfState.funcs[i]);
  func->active = TRUE;
  func->itime = -1;
}


/***********************************************************************
Deactivate a registered function (called from inside event loop)
***********************************************************************/
void GSdeactivatefunction(char *name)
{
  int i;

  i = getname(InfState.funcs, sizeof (Function), InfState.numfuncs,
			  name, "function");
  InfState.funcs[i].active = FALSE;
}


/***********************************************************************
 Return the enclosing box for a named field.
 GLOBALS: InfState.
***********************************************************************/
void GSgetfieldrect(char *name, int *x, int *y, int *wd, int *ht,
                  GSColor *color)
{
  int i;
  Field *fld;

  i = getname(InfState.fields, sizeof (Field), InfState.numfields,
              name, "field");
  fld = &(InfState.fields[i]);
  if (x)
    *x = fld->x;
  if (y)
    *y = fld->y;
  if (wd)
    *wd = fld->wd;
  if (ht)
    *ht = fld->ht;
  if (color)
    *color = fld->color;
}


/***********************************************************************
Public front end to fastrow.  Makes the coordinate window-relative.
GLOBALS: InfState, Win.
***********************************************************************/
void  GSwfastrow(int x, int y, GSColor row[], int n)
/*void  GSwfastrow(int x, int y, ColorIndexType row[], int n)*/
{
  x += Win[InfState.cw].x;
  y += Win[InfState.cw].y;
  GSPfastrow(x, y, row, n);
}



/***********************************************************************
 Set a pixel in a window
 GLOBALS: InfState, Win.
***********************************************************************/
void GSwdrawpoint(int x0, int y0)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
    GSelog (-1, "Bad wdrawpoint\n");
#endif
  x0 += Win[InfState.cw].x;
  y0 += Win[InfState.cw].y;
  GSPdraw_pixel(x0,y0);
}


/*------------------------------------------------------------------------------
  Get a pixel's color in a window
  GLOBALS: InfState, Win.

  -Bo&Colin 09/15/93: Check apps for use of this routine. If not used, kill it. 
  -Dave 08/10/94: It's used by Mental Maps.
------------------------------------------------------------------------------*/
GSColor GSwgetpoint(int x0, int y0)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wgetpoint\n");
#endif
  x0 += Win[InfState.cw].x;
  y0 += Win[InfState.cw].y;
  return (GSPget_pixel (x0,y0));
}


/***********************************************************************
 Draw a line in the current window from (x0, y0) to (x1, y1).
 x0, y0, x1, y1 are relative coordinates.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwdrawline(int x0, int y0, int x1, int y1)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wdrawline\n");
#endif
  x0 += Win[InfState.cw].x;
  y0 += Win[InfState.cw].y;
  x1 += Win[InfState.cw].x;
  y1 += Win[InfState.cw].y;
  GSPline(x0, y0, x1, y1);
}


/***********************************************************************
 Draw a polygon using current line style and color.  Then fills the
 poly using the current color.
***********************************************************************/
void GSwfillpoly(
  int numpoints,  /* number of vertices in polygon */
  int *polypoints /* array of 2*numpoints ints that are x-y coords of vertices*/
)
{
  int i;

  for (i=0; i<numpoints; i++)
  {
    polypoints[2*i]  += Win[InfState.cw].x;  /* add window x to poly x coords */
    polypoints[2*i+1] += Win[InfState.cw].y; /* add window y to poly y coords */
  }
  GSPfillpoly(numpoints, polypoints);
}



/***********************************************************************
 Draw a rectangle in the current window with dimensions (x0, y0, wd, ht).
 x0, y0 are relative coordinates.
 If the rectangle is 0 in one dimension, draw a line.
 If the rectangle is 0 in both dimensions, don't draw (but do set attributes).
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwdrawrect(int x0, int y0, int wd, int ht)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wdrawrect\n");
#endif
  x0 += Win[InfState.cw].x;
  y0 += Win[InfState.cw].y;
  /* Don't draw lines with no dimension */
  if (!wd && !ht)
	return;
  GSPframe_rect(x0, y0, wd, ht);
}

/***********************************************************************
 Draw an ellipse in the current window with dimensions (x0, y0, wd, ht).
 x0, y0 are relative coordinates.
 If the rectangle is 0 in one dimension, draw a line.
 If the rectangle is 0 in both dimensions, don't draw (but do set attributes).
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwdrawellipse(int x0, int y0, int wd, int ht)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wdrawellipse\n");
#endif
  x0 += Win[InfState.cw].x;
  y0 += Win[InfState.cw].y;
  /* Don't draw lines with no dimension */
  if (!wd && !ht)
	return;
  GSPframe_ellipse(x0, y0, wd, ht);
}

/***********************************************************************
 Clear a rectangle in the current window with dimensions (x, y, wd, ht)
 (i.e., set region to background color).
 x, y are relative coordinates.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwclearrect(int x, int y, int wd, int ht)
{
  int savecolor;
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wclearrect\n");
#endif
  savecolor = GSgetcolor();
  x += Win[InfState.cw].x;
  y += Win[InfState.cw].y;
  GSsetcolor(Win[InfState.cw].bgcolor);
  GSPfill_rect(x + 1, y + 1, wd - 2, ht - 2);
  GSsetcolor(savecolor);
}


/***********************************************************************
 Solid fill a rectangle in the current window with dimensions
 (x, y, wd, ht). x, y are relative coordinates.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwfillrect(int x, int y, int wd, int ht)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wfillrect\n");
#endif
  x += Win[InfState.cw].x;
  y += Win[InfState.cw].y;
  if (wd && ht)  /* Don't draw a bar with zero dimension */
	GSPfill_rect(x, y, wd, ht);
}


/***********************************************************************
 Solid fill an ellipse in the current window with dimensions
 (x, y, wd, ht). x, y are relative coordinates.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwfillellipse(int x, int y, int wd, int ht)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wfillrect\n");
#endif
  x += Win[InfState.cw].x;
  y += Win[InfState.cw].y;
  if (wd && ht)  /* Don't draw a bar with zero dimension */
	GSPfill_ellipse(x, y, wd, ht);
}

/***********************************************************************
 Set mouse postion within the current window.
 (x, y). x, y are relative coordinates.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwsetmouseposition(int x, int y)
{
#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad GSwsetmouseposition\n");
#endif
  x += Win[InfState.cw].x;
  y += Win[InfState.cw].y;
  set_mouse_position(x, y);
}


/***********************************************************************
 Put text in a window.  (x, y) is relative to the current window.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwwritemsg(int x, int y, char *string)
{
  GSPdraw_string(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y, string);
}

/***********************************************************************
 Put text in a window.  (x, y) is relative to the current window.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwwriteclippedmsg (
  int	clip_x,
  int	clip_y,
  int	clip_wd,
  int	clip_ht,
  int	msg_x,
  int	msg_y,
  char	*string
)
{
  int		tht = GStextheight(string);
  boolean	error = FALSE;

  if (clip_x < 0) {
    GSelog(0, "Negative value passed to clip_x in GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_y < 0) {
    GSelog(0, "Negative value passed to clip_y in GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_wd < 1) {
    GSelog(0, "Non-positive value passed to clip_wd in GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_ht < 1) {
    GSelog(0, "Non-positive value passed to clip_ht in GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_x >= Win[InfState.cw].wd)  {
    GSelog(0, "Value passed to clip_x is beyond right edge of window in");
    GSelog(0, " GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_y >= Win[InfState.cw].ht)  {
    GSelog(0, "Value passed to clip_y is beyond bottom edge of window in");
    GSelog(0, " GSwwriteclippedmsg\n");
    error = TRUE;
  }
  if (clip_x + clip_wd >= Win[InfState.cw].wd)  {
    clip_wd = Win[InfState.cw].wd - clip_x;
  }
  if (error) {
    return;
  }
  GSPset_clip_rect(Win[InfState.cw].x + clip_x + 1,
                   Win[InfState.cw].y + clip_y - tht,
                   clip_wd-2,  clip_ht);
/*  GSPset_clip_rect(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y + 1,
                   clip_wd-2,  clip_ht-2);*/
  GSPdraw_string(Win[InfState.cw].x + msg_x + 1, Win[InfState.cw].y + msg_y,
                 string);
  GSPset_clip_rect(Win[InfState.cw].x + Win[InfState.cw].bordersize,
                   Win[InfState.cw].y + Win[InfState.cw].bordersize,
                   Win[InfState.cw].wd - 2 * Win[InfState.cw].bordersize,
                   Win[InfState.cw].ht - 2 * Win[InfState.cw].bordersize);
}


/***********************************************************************
Put named text message in a window.  (x, y) is relative to the current
  window.
GLOBALS: Win, InfState.
***********************************************************************/
void GSwwritenamedmsg(int x, int y, char *msgname)
{
  int i;

  i = getname(InfState.msgs, sizeof (Message), InfState.nummsgs, msgname,
              "message");
  GSPdraw_string(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y,
            InfState.msgs[i].text);
}


/***********************************************************************
 Convert a double to a string containing an integer with commas.
 Then write it to the current window.  (x, y) is relative to the window.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwwritecommanum(int x, int y, double num)
{
  int i;
  int cnt;
  int len;
  BigString tmpstr;
  BigString lnum;

#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
    GSelog (-1, "Bad wwritecommanum\n");
#endif
  sprintf(tmpstr, "%.0f", num);
  len = strlen(tmpstr);
  cnt = 0;
  for (i=0; i<len; i++)
  {
    lnum[cnt] = tmpstr[i];
    cnt++;
    if ((len - 1 - i) % 3 == 0)
    {
      lnum[cnt] = ',';
      cnt++;
    }
  }
  lnum[cnt-1] = '\0';
  GSPdraw_string(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y, lnum);
}


/***********************************************************************
 Write a double to the current window by the caller's specs.
 (x, y) is relative to the window.
 GLOBALS: None.
***********************************************************************/
void GSwwritereal( int width,           /* min width of output format string */
                   int precision,
                int x, int y, double num)
{
  BigString tmpstr, fstr;

#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
	GSelog (-1, "Bad wwritereal\n");
#endif
  sprintf(fstr, "%%%d.%df", width, precision);
  sprintf(tmpstr, fstr, num);
  GSPdraw_string(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y, tmpstr);
}


/***********************************************************************
 Write a long to the current window by the caller's specs.
 (x, y) is relative to the window.
 GLOBALS: Win, InfState.
***********************************************************************/
void GSwwriteint(int width,             /* min width of output format string */
               int x, int y, long num)
{
  BigString tmpstr, fstr;

#ifdef DEBUG
  if ((InfState.cw < 0) || (InfState.cw >= InfState.numwindows))
    GSelog (-1, "Bad wwriteint\n");
#endif
  sprintf(fstr, "%%%dld", width);
  sprintf(tmpstr, fstr, num);
  GSPdraw_string(Win[InfState.cw].x + x + 1, Win[InfState.cw].y + y, tmpstr);
}

/**********************************************************************
 Return in "name" the name of the window containing the active button.
 GLOBALS: InfState, Win.
**********************************************************************/
char *GSgetcurrentwindowname(char *name)
{
  strcpy(name, Win[InfState.cw].name);
  return(name);
}


/**********************************************************************
 Activate a named window (without drawing).
 GLOBALS: Win, InfState.
**********************************************************************/
void GSsetactivenamedwindow(char *name)
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  Win[i].active = InfState.level;
  enable_buttons_after_popup(i);
}


/**********************************************************************
 Deactivate a named window (without removing).
 GLOBALS: Win, InfState.
**********************************************************************/
void GSsetdeactivenamedwindow(char *name)
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  Win[i].active = 0;
}


/**********************************************************************
 Make a named window the current window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSsetcurrentnamedwindow(char *name)
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  GSPsetcurrentwindow(i);
}

/**********************************************************************
 Remove the top window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSremovetopwindow(void)
{
  int i;

/*  i = getname(Win, sizeof (GSWindow), InfState.numwindows, "alert", "window");
 Passing "window" as the last param will cause an error exit if there is
no "alert" window.  In effect, this requires an alert window in order to
call this function.  - JBR */

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, "alert", NULL);
  if ((i != -1) && (InfState.cw == i))
    AlertUp = FALSE;    
  removewindow(InfState.cw);
}

/**********************************************************************
 Remove named window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSremovenamedwindow( char *name) /* name of the window */
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  if (GSstricmp(name, "alert") == 0)
    AlertUp = FALSE;
  removewindow(i);
}


/**********************************************************************
 Draw named window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSdrawnamedwindow(
  char *name,  /* name of the window */
  int bstore, /* Will be BSTORE if backing store to be used */
  int popup)  /* Will be POPUP if this is a popup window */
{
  int i;
  int j, win_bstore;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  if ((win_bstore = Win[i].bstore_to_use) != 0)
  {
    Win[i].bstore = bstore;
    if (bstore == BSTORE) {
		if (InfState.winbstore != BSTORE)
			GSelog(-1, "Error - interface was not initialized with"
						"backing store.\n"
						"In GSinterfaceinit, pass BSTORE as the bstore"
						" argument.\n");
		if (InfState.backingstore[win_bstore].inuse)
			GSelog(-1, "Error - backing store %d is already occupied.\n",
															win_bstore);
		InfState.backingstore[win_bstore].inuse = TRUE;
		if (!GSPto_bstore(win_bstore, Win[i].x, Win[i].y, Win[i].wd, Win[i].ht))
		GSelog(-1, "Error backing Window |%s| to b-store (%ld): %d %d %d %d\n",
               name, InfState.backingstore[win_bstore].bsize, Win[i].x,
               Win[i].y, Win[i].wd, Win[i].ht);
		}
  } 

  Win[i].popup = popup;
  if (popup)
  {
    for (j = 0; j < InfState.numwindows; j++)
    {
      if (i != j && (Win[j].active == InfState.level))
      {
        disable_buttons_for_popup(j);
      }
    }
    InfState.level++;
  }
  Win[i].active = InfState.level;
  drawwindow(i);
}


/**********************************************************************
 Redraw named window.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSredrawnamedwindow(char *name)
{
  int i;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, name, "window");
  if (Win[i].active > 0)
    drawwindow(i);
}


/**********************************************************************
 Redraw all active windows.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSredrawactivewindows(void)
{
  int i;
  int j;

/*  for (i=0; i<InfState.numwindows; i++)
    if (Win[i].active > 0)
      drawwindow(i);*/
  for (j = 1; j <= InfState.level; j++)
    for (i=0; i<InfState.numwindows; i++)
      if (Win[i].active == j)
        drawwindow(i);
}

/**********************************************************************
 Bring up a named popup menu (caller's access to an on-screen popup
 menu).
 GLOBALS: Popups, InfState.
**********************************************************************/
void GScreatenamedpopup (char *pname)
{
  int i;

  i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, pname, "popup");
  PopupSaveMsg = GSgetcurrmsg();
  createpopupmenu(i);
}


/**********************************************************************
 Enable a named button.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSenablebutton (
  char *wname,  /* Window name */
  char *bname,  /* Button name */
  int draw)     /* Will be set to DRAW if should draw. */
{
  int i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  Win[i].wb[j].active = ENABLED;
  GSPsetcurrentwindow(i);
  if (draw == DRAW)
    drawbutton(i, j, FALSE);
}

/**********************************************************************
 Make a named button invisible.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSinvisiblebutton (
  char *wname,  /* Window name */
  char *bname,  /* Button name */
  int draw)     /* Will be set to DRAW if should draw (erase). */
{
  int i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  Win[i].wb[j].active = INVISIBLE;
  GSPsetcurrentwindow(i);
  if (draw == DRAW)
    erasebutton(i, j);
}

/**********************************************************************
 Make a named button visible.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSvisiblebutton (
  char *wname,  /* Window name */
  char *bname,  /* Button name */
  int draw)     /* Will be set to DRAW if should draw. */
{
  int i, j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  Win[i].wb[j].active = ENABLED;
  GSPsetcurrentwindow(i);
  if (draw == DRAW)
    drawbutton(i, j, FALSE);
}

/**********************************************************************
 Disable a named button.
 GLOBALS: Win, InfState.
**********************************************************************/
void GSdisablebutton (
  char *wname,  /* Window name */
  char *bname,  /* Button name */
  int draw)     /* Will be set to DRAW if should draw. */
{
  int i,j;

  i = getname(Win, sizeof (GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof (GSButton), Win[i].numbuttons, bname, "button");
  Win[i].wb[j].active = DISABLED;
  GSPsetcurrentwindow(i);
  if (draw == DRAW)
    drawbutton(i, j, FALSE);
}


/*------------------------------------------------------------------------------
 Remove a popup menu item.
 GLOBALS: Popups, InfState.
------------------------------------------------------------------------------*/
void GSinvisiblemenuitem (
  char	*pname,	/*-- Popup window name --*/
  char	*iname	/*-- Item name --*/
)
{
  int i, j;

  i = getname(Popups, sizeof(Popupmenu), InfState.numpopups, pname, "popup");
  j = getname(Popups[i].item,sizeof(Menuitem),Popups[i].numitems,iname,"item");
  Popups[i].item[j].active = INVISIBLE;
}

/*------------------------------------------------------------------------------
 Restore a popup menu item.
 GLOBALS: Popups, InfState.
------------------------------------------------------------------------------*/
void GSvisiblemenuitem (
  char	*pname,	/*-- Popup window name --*/
  char	*iname	/*-- Item name --*/
)
{
  int i, j;

  i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, pname, "popup");
  j = getname(Popups[i].item,sizeof(Menuitem),Popups[i].numitems,iname,"item");
  Popups[i].item[j].active = ENABLED;
}

/*------------------------------------------------------------------------------
 Enable a popup menu item.
 GLOBALS: Popups, InfState.
------------------------------------------------------------------------------*/
void GSenablemenuitem (
  char	*pname,	/*-- Popup window name --*/
  char	*iname	/*-- Item name --*/
)
{
	int i, j;
	
	if (iname == NULL)
		GSelog(-1, "Error: no name given to 'GSenablemenuitem'.\n");
	
	i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, pname, "popup");
	j = getname(Popups[i].item,sizeof(Menuitem),Popups[i].numitems,iname,"item");
	Popups[i].item[j].active = ENABLED;
}


/*------------------------------------------------------------------------------
 Disable a popup menu item.
 GLOBALS: Popups, InfState.
------------------------------------------------------------------------------*/
void GSdisablemenuitem (
  char	*pname,	/*-- Popup window name --*/
  char	*iname	/*-- Item name --*/
)
{
	int i, j;
	
	if (iname == NULL)
		GSelog(-1, "Error: no name given to 'GSdisablemenuitem'.\n");
	
	i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, pname, "popup");
	j = getname(Popups[i].item,sizeof(Menuitem),Popups[i].numitems,iname,"item");
	Popups[i].item[j].active = DISABLED;
}


/***********************************************************************
 Initialize the mouse, making sure mouse is present.
 GLOBALS: Cursx, Cursy.
***********************************************************************/
void GSmouseinit (
  void
)
{
  Cursx = Maxx/2;
  Cursy = Maxy/2;

  /*  initialize mouse  */
  if (!mouse_present())
  {
    printf("You must have a mouse installed to run this program.\n");
    exit (-1);
  }
  GSmouseon();
}


/***********************************************************************
 This handles the interface.
 GLOBALS: InfState, Win, Popups, Cursx, Cursy.
***********************************************************************/
void GSinterface(void)
{
  GSmouseinit();
  GSPdo_events_forever();
}

/***********************************************************************
 Initialize the user interface.
 GLOBALS: InfState, Maxx, Maxy, Cursx, Cursy.
***********************************************************************/
void GSinterfaceinit (
  BigString		appl_name,
  BigString		ifile,	/* Interface file */
  int			bstore,	/* BSTORE iff application wants to use bstore*/
  char			*elog_file_mode,	/* "w"rite or "a"ppend */
  GraphicsModeType	graphmode,		/* G640x480x16 or G640x480x256 */
/*  int			argc,
  char			**argv*/
  GS_MAIN_ARGS_F
)
{
  int i, win_bstore;
  int maxwinwd[MAXBSTORES+1], maxwinht[MAXBSTORES+1];

  InfState.speedup = INIT_SPEEDUP;
  InfState.numwindows = -1;
  InfState.numpopups = 0;
  InfState.cw = 0;
  InfState.cb = -1;
  InfState.level = 1;
  InfState.activepopupwindow = -1;
  InfState.highlight = 0;

  init_error_log (elog_file_mode);


/*  GraphicsMode = graphmode;*/
  GraphicsMode =  (graphmode==G640x480x256)?256:16;
  
/*  GSPinit_system(appl_name, argc, argv);*/
	GSPinit_system(appl_name, GS_MAIN_ARGS_A);
  GSPinit_events();

  readfilepass1(ifile);

  readinterfacefile(ifile);

  GSPinit_palette(Interface_Palette, GraphicsMode);

  assignpopups();

  for (i=1; i<=MAXBSTORES; maxwinwd[i]=0, maxwinht[i]=0, i++);

  /* find max bstore req'd by popup menus */
  for (i=0; i<InfState.numpopups; i++)
  {
    if ((long)Popups[i].wd > (long)(maxwinwd[POPUPBSTORE]))
    {
      maxwinwd[POPUPBSTORE] = Popups[i].wd;
    }
    if ((long)(TSIZE/2+Popups[i].numitems*TSIZE) > (long)(maxwinht[POPUPBSTORE]))
    {
      maxwinht[POPUPBSTORE] = (TSIZE/2+Popups[i].numitems*TSIZE);
    }
  }
  
  InfState.winbstore = bstore;

  if (bstore == BSTORE) /* find max bstore req'd by windows */
    for (i=0; i<InfState.numwindows; i++)
      if ((win_bstore = Win[i].bstore_to_use) != 0)
      {  /* if this win can use	bstore, then see how much it needs */
	  if (win_bstore > MAXBSTORES)
	  GSelog(-1, "Error:  can't have bstore # %d", win_bstore);
        if (Win[i].wd > maxwinwd[win_bstore])
          maxwinwd[win_bstore] = Win[i].wd;
        if (Win[i].ht > maxwinht[win_bstore])
          maxwinht[win_bstore] = Win[i].ht;
      }

  for (i=1; i <= MAXBSTORES; i++) {
	InfState.backingstore[i].inuse = FALSE;
    if (!GSPinit_bstore(i, maxwinwd[i], maxwinht[i]))
      GSelog(-1,
	         "Error:  cannot initialize backing store %d.  Out of memory.\n",
             i);
	}
}


/***********************************************************************
Draw an unfilled rectangle whose pixels are the inverse of their
background.
***********************************************************************/
void GSdrawxorbox (int x, int y, int wd, int ht)
{
  int savecolor;

  savecolor = GSgetcolor();
  GSPxormode(ON);
  GSsetcolor(GSconvertcolor("white"));
  GSwdrawrect (x, y, wd, ht);
  GSsetcolor(savecolor);
  GSPxormode(OFF);
}


/*------------------------------------------------------------------------------
  Draw a line whose pixels are the inverse of their background.
------------------------------------------------------------------------------*/
void GSdrawxorline (
  int x0,
  int y0,
  int x1,
  int y1
)
{
  int savecolor;

  savecolor = GSgetcolor();
  GSPxormode(ON);
  GSsetcolor(GSconvertcolor("white"));
  GSwdrawline (x0, y0, x1, y1);
  GSsetcolor(savecolor);
  GSPxormode(OFF);
}


/*------------------------------------------------------------------------------
  Draw a line whose pixels are the inverse of their background.
------------------------------------------------------------------------------*/
void GSdrawinvertline (
  int x0,
  int y0,
  int x1,
  int y1
)
{
  int savecolor;

  savecolor = GSgetcolor();
  GSPinvertmode(ON);
  GSsetcolor(GSconvertcolor("white"));
  GSwdrawline (x0, y0, x1, y1);
  GSsetcolor(savecolor);
  GSPinvertmode(OFF);
}


/*------------------------------------------------------------------------------
  Draw an unfilled rectangle whose pixels are the inverse of their background
------------------------------------------------------------------------------*/
void GSdrawinvertbox (int x, int y, int wd, int ht)
{
  int savecolor;

  savecolor = GSgetcolor();
  GSPinvertmode(ON);
  GSsetcolor(GSconvertcolor("white"));
  GSwdrawrect (x, y, wd, ht);
  GSsetcolor(savecolor);
  GSPinvertmode(OFF);
}


/***********************************************************************/
/*  Set the origin (absolute) of a window                              */
/***********************************************************************/
void GSsetwindowspecs (
  char	*wname,
  int	*x,
  int	*y,
  int	*wd,
  int	*ht
)
{
  int i;

  for (i=0; i < InfState.numwindows; i++) {
    if (GSstricmp (wname, Win[i].name) == 0) {
      if (wd != NULL) {
        if (*wd > 0) {
          Win[i].wd = *wd;
	}
      }

      if (ht != NULL) {
        if (*ht > 0) {
          Win[i].ht = *ht;
	}
      }
      /*-- note: must do dimensions first, since they are referenced below --*/
      if (x != NULL) {
        if (*x + Win[i].wd < Maxx) { /* As long as window fits horizontally, */
          Win[i].x = *x;             /* set the x coord to specified x; */
	}
        else if (Win[i].wd < Maxx) { /* else, if width is less than Maxx, */
          Win[i].x = Maxx - Win[i].wd; /* set x coord as far right as possible */
	}
      }
      /* Otherwise, don't change window's x coord. */

      if (y != NULL) {
        if (*y + Win[i].ht < Maxy) {    /* Do the same for vertical dimension */
          Win[i].y = *y;
	}
        else if (Win[i].ht < Maxy) {
          Win[i].y = Maxy - Win[i].ht;
	}
      }
      return;
    }
  }

  GSelog(-1, "Invalid window |%s| in set window origin.\n", wname);
  return;
}

/***********************************************************************/
/*  Set the origin (absolute) of a window                              */
/***********************************************************************/
void GSsetmenuorigin (char *mname, int *x, int *y)
{
  int i;

  for (i=0; i < InfState.numpopups; i++) {
    if (GSstricmp (mname, Popups[i].name) == 0) {
      if (x != NULL) {
        if (*x + Popups[i].wd < Maxx) {
          /* As long as the menu fits horizontally, set x coord to specified x */
          Popups[i].x = *x;
	}
        else if (Popups[i].wd < Maxx) {
          /* otherwise, if width < Maxx, set x coord as far right as possible. */
          Popups[i].x = Maxx - Popups[i].wd;
	}
      } /* otherwise, don't change window's x coord. */
      if (y != NULL) {
        if (*y + Popups[i].ht < Maxy) {  /* Do the same for vertical dimension */
          Popups[i].y = *y;
	}
        else if (Popups[i].ht < Maxy) {
          Popups[i].y = Maxy - Popups[i].ht;
        }
      } /* otherwise, don't change window's y coord. */
      return;
    }
  }

  GSelog(-1, "Invalid popup |%s| in set menu origin.\n", mname);
  return;
}

/***********************************************************************/
/*  Save the status of the buttons                                     */
/***********************************************************************/
void GSsavewindows (FILE *fptr)
{
  int i, j;

  for (i=0; i < InfState.numwindows; i++)
  {
    if (fwrite (&(Win[i].active), sizeof (boolean), 1, fptr) != 1)
      GSelog (0, "Error while saving state of program.  Win[%d].active\n", i);
    for (j=0; j < Win[i].numbuttons; j++)
    {
      if (fwrite (&(Win[i].wb[j].active), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog (0, "Error while saving state of program.");
        GSelog (0, "Win[%d].wb[%d].active\n", i, j);
      }
    }
  }
  for (i=0; i < InfState.numpopups; i++)
  {
    if (fwrite (&(Popups[i].active), sizeof (boolean), 1, fptr) != 1)
    {
      GSelog (0, "Error while saving state of program.");
      GSelog (0, "Popups[%d].active \n", i);
    }
    for (j=0; j < Popups[i].numitems; j++)
    {
      if (fwrite (&(Popups[i].item[j].active), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog (0, "Error while saving state of program.");
        GSelog (0,  "Popups[%d].item[%d].active\n", i, j);
      }
      if (fwrite (&(Popups[i].item[j].checked), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog (0, "Error while saving state of program.");
        GSelog (0, "Popups[%d].item[%d].checked\n", i, j);
      }
    }
  }
}

/***********************************************************************/
/*  Load the status of the buttons                                     */
/***********************************************************************/
void GSloadwindows (FILE *fptr)
{
  int i, j;

  for (i=0; i < InfState.numwindows; i++)
  {
    if (fread (&(Win[i].active), sizeof (boolean), 1, fptr) != 1)
    {
      GSelog(0, "Error while loading saved state of program.");
      GSelog(0, "Win[%d].active\n", i);
    }
    for (j=0; j < Win[i].numbuttons; j++)
    {
      if (fread (&(Win[i].wb[j].active), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog(0, "Error while loading saved state of program.");
        GSelog(0, "Win[%d].wb[%d].active\n", i, j);
      }
    }
  }
  for (i=0; i < InfState.numpopups; i++)
  {
    if (fread (&(Popups[i].active), sizeof (boolean), 1, fptr) != 1)
    {
      GSelog(0, "Error while loading saved state of program.");
      GSelog(0, "Popups[%d].active \n", i);
    }
    for (j=0; j < Popups[i].numitems; j++)
    {
      if (fread (&(Popups[i].item[j].active), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog(0, "Error while loading saved state of program.");
        GSelog(0, "Popups[%d].item[%d].active\n", i, j);
      }
      if (fread (&(Popups[i].item[j].checked), sizeof (boolean), 1, fptr) != 1)
      {
        GSelog (0, "Error while loading saved state of program.");
        GSelog (0, "Popups[%d].item[%d].checked\n", i, j);
      }
    }
  }
}

/**********************************************************************
Convert a color name to a color index. (EGA/VGA only)
GLOBALS: None.
**********************************************************************/
GSColor GSconvertcolor (char *color)
{
  int currcolor;

  /*currcolor = convertcolor(Alias, Acolor, color);*/
  currcolor = GSiselword(color, Colorlist);
  if (currcolor == -1)
  {
    currcolor = GSiselword(color, Alias);
    if (currcolor != -1)
    {
      currcolor = GSiselword(Acolor[currcolor], Colorlist);
      if (currcolor == -1)
      {
        return (15); 
      }
    }
  }
  return((GSColor)currcolor);
}

/**********************************************************************
Toggle checkmark on checklist menu item "itemnum".
**********************************************************************/
void GStogglecheck(char *menuname, char *itemname)
{
  int i, j;

  i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, menuname,
              "popup");
              
  j = getname(Popups[i].item, sizeof (Menuitem), Popups[i].numitems, itemname,
  				"menu item");
  
  toggle_check(i, j, Popups[i].bg);
/*  if ((itemnum >= 0)
   && (itemnum < Popups[i].numitems)
   && (Popups[i].checklist))
  {
    toggle_check(i, itemnum, Popups[i].bg);
  }
  else
  {
    GSelog(0, "Error in call to GStoggle_check.");
    GSelog(-1, "Menu: |%s|, checklist = %d, itemnum = %d, numitems = %d",
           Popups[i].name, Popups[i].checklist, itemnum, Popups[i].numitems);
  }*/
}

int GSgetmenuitemnum(char *menuname, char *itemname)
{
	int i, j;
	
	i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, menuname,
	          "popup");
	          
	j = getname(Popups[i].item, sizeof (Menuitem), Popups[i].numitems, itemname,
				"menu item");
	
	return j;
}

char *GSgetmenuitemname(char *menuname, int itemnum)
{
	int i;
	i = getname(Popups, sizeof (Popupmenu), InfState.numpopups, menuname,
	          "popup");

	if ((itemnum >= 0)  && (itemnum < Popups[i].numitems))
		return Popups[i].item[itemnum].name;
	else
		return NULL;
}

/**********************************************************************
Return the most recently pressed control and it's information.
**********************************************************************/
void GSgetlastcontrol(ControlType *control)
{
  control->type = InfState.lastcontrol.type;
  strcpy(control->gname, InfState.lastcontrol.gname);
  strcpy(control->time, InfState.lastcontrol.time);
  if (InfState.lastcontrol.type == WINDOW)
	strcpy(control->control.name, InfState.lastcontrol.control.name);
  else
	control->control.number = InfState.lastcontrol.control.number;
}

/**********************************************************************
Clear the last control struct.
Use this if tutorial wants to check for the same control consecutively.
**********************************************************************/
void GSclearlastcontrol(void)
{
  InfState.lastcontrol.type = NONE;
  strcpy(InfState.lastcontrol.gname, "");
  InfState.lastcontrol.control.number = -1;
}

/**********************************************************************
GSgetmsg.  Returns the text of the named message.
**********************************************************************/
char *GSgetmsg (char *name)
  {
  int i;

  i = getname(InfState.msgs, sizeof (Message), InfState.nummsgs, name, NULL);
  if (i >= 0)
	return(InfState.msgs[i].text);
  else
	return ("");
  }

/*****************************************************************************
Timing functions, meant to be used like a stopwatch.  Call GSPstart_timer to
set the timer to 0.  Call GSPread_timer thereafter to get the number of
milliseconds that have passed since the call to GSPstart_timer.

Timing funcs moved to platform source.  JBR
*****************************************************************************/

#if 0
void GSPstart_timer(void) { Start_time = clock(); }

long GSPread_timer(void)
  {
  /*
	CLK_TCK is the number of processor "clicks" per second.

	CLK_TCK should be defined in the platform's
	time.h header.  If not, define it in the XXXsys.h
	header.  JBR
  */
  long curTime = (long) ((clock() - Start_time)*1000 / CLK_TCK);
  
  return curTime;
  }
#endif

/*****************************************************************************
 Returns size in bytes of file f
*****************************************************************************/
long GSfilesize (FILE *f )
{
  long  pos, sz = 0;

  pos = ftell(f);
  if (fseek(f, 0, SEEK_END))
    GSelog(-1, "Error: GSfilesize unable to determine file size.");
  sz = ftell(f);
  fseek(f, pos, SEEK_SET);
  return sz;
}


/*****************************************************************************
  Converts path separator character in file name to platform specific
  character.  Then fopens the file.
*****************************************************************************/
FILE *GSopen(char *filename, char *mode)
{
  char tempname[HUGESTRINGLEN], *tptr, *fptr, c;
  char path_separator = PATH_SEPARATOR;

  tptr = tempname;
  fptr = filename;
  if (path_separator == ':')  *tptr++ = ':';

  while (1)
  {
    c = *fptr++;
    *tptr++ = (c == '/')? PATH_SEPARATOR : c;
    if (c == '\0')  break;
  }

  return fopen(tempname, mode);
}

/*****************************************************************************
Returns whether this machine is big-endian.
*****************************************************************************/
/*boolean GSbigendian(void)
{
	return (BIG_ENDIAN != 0);
}
Now done with a macro.  JBR
*/


/*------------------------------------------------------------------------------
  GSframefield: draws a highlighted frame around field.
------------------------------------------------------------------------------*/
void GSframefield (
  char *window,	/*--- current window when frame is drawn;
                              optionally NULL(semantics: don't change win) ---*/
  char *field,	/*--- field to be framed ---*/
  GSColor   tlcolor,	/*--- color for top and left edges of frame ---*/
  GSColor   brcolor	/*--- color for bottom and right edges of frame ---*/
)
{
  int		fx, fy, fwd, fht;
  int		save_window;

  save_window = InfState.cw;  
  if (window)
  {
    GSsetcurrentnamedwindow(window);
  }
  if (field == NULL)
  {
    GSelog(0, "NULL passed to parameter field in GSframefield; \n");
    GSelog(0, "frame not drawn.\n");
    return;
  }
  GSgetfieldrect(field, &fx, &fy, &fwd, &fht, NULL);
  drawshadedframe(fx, fy, fwd, fht, tlcolor, brcolor);
  GSPsetcurrentwindow(save_window);
}


/*------------------------------------------------------------------------------
  GSframedragarea: draws a highlighted frame around dragarea.
------------------------------------------------------------------------------*/
void GSframedragarea (
  char *window,	/*--- window which contains dragarea ---*/
  char *dragarea,	/*--- dragarea to be framed ---*/
  GSColor   tlcolor,	/*--- color for top and left edges of frame ---*/
  GSColor   brcolor	/*--- color for bottom and right edges of frame ---*/
)
{
  int		fx, fy, fwd, fht;
  int		save_window;

  if (window == NULL)
  {
    GSelog(0, "NULL passed to parameter window in GSframedragarea\n");
    GSelog(0, "frame not drawn.\n");
    return;
  }
  if (dragarea == NULL)
  {
    GSelog(0, "NULL passed to parameter dragarea in GSframedragarea\n");
    GSelog(0, "frame not drawn.\n");
    return;
  }
  save_window = InfState.cw;  
  GSsetcurrentnamedwindow(window);
  GSgetdragareaspecs(window, dragarea, &fx, &fy, &fwd, &fht, NULL, NULL);
  drawshadedframe(fx, fy, fwd, fht, tlcolor, brcolor);
  GSPsetcurrentwindow(save_window);
}




