/**********************************************************************
Log file routines.  All GeoSim routines have access (and are encouraged
to use) the two logfiles "error.log" (from GSelog) and
"output.log" (from GSolog).
**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "geosim.h"
#include "version.h"
#include "gspriv.h"

boolean GSDebug = FALSE;

boolean GSpermit_log = TRUE; // Switch logging on/off

static FILE *Out_log = NULL;
static FILE *Err_log = NULL;

/**********************************************************************
Initialize the output log "output.log".
**********************************************************************/
boolean GSinitlog (
  char *mode_string  /* "w" = write, "a" = append (see fopen() for more) */
) {
  if (!GSpermit_log) return(TRUE);

  if ((Out_log = fopen("output.log", mode_string)) == NULL)
    return (FALSE);

  return(TRUE);
}


/**********************************************************************
Write a message to the output log "output.log" and flush it.
**********************************************************************/
void GSolog(char *format, ...)
{
  va_list argptr;

  if (!GSpermit_log) return;

  if (Out_log == NULL)
    GSelog(-1, "Tried to write to uninitialized output.log\n");
  va_start(argptr, format);
  vfprintf(Out_log, format, argptr);
  va_end(argptr);
  fflush(Out_log);
}


/**********************************************************************
Initialize error log "error.log".
This is done either from GSinterface_init or upon initial call to
GSelog.
**********************************************************************/
boolean init_error_log(char *mode_string)  /* "w" = write, "a" = append */
{
  if (!GSpermit_log) return(TRUE);

  if (Err_log == NULL)      /* make sure Err_log is not already open */
  {
    if ((Err_log = fopen("error.log", mode_string)) == NULL)
      return (FALSE);
    fprintf(Err_log, "GeoSim Interface Handler Library Version %s\n", GSVERSION);
    fflush(Err_log);
  }
  return (TRUE);
}


/**********************************************************************
Write and flush a message to the error log "error.log".
**********************************************************************/
void GSelog(int done, char *format, ...)
{
  va_list argptr;

  if (!GSpermit_log) return;

  if (Err_log == NULL)
    if(!init_error_log("w"))
    {
      printf("Unable to write to error.log\n");
      exit(-1);
    }
  va_start(argptr, format);
  vfprintf(Err_log, format, argptr);
  va_end(argptr);
  fflush(Err_log);
  if (done != 0)
  {
	closegraph();
	fclose(Err_log);
	if (Out_log)
	  fclose(Out_log);
	va_start(argptr, format);
	vprintf(format, argptr);
	va_end(argptr);
	/* Wait some time before exiting.  This is so that the user can see the
	   message before returning to a graphical environment */
	GSdelay(ERRORDELAY);
	exit (done);
  }
}
