#include <string.h>
#include <ctype.h>
#include "geosim.h"
#include "gspriv.h"

#define  RETURN                13

extern Iconfig InfState;
extern GSWindow *Win;

static boolean LoginWindowUp = FALSE;
static boolean (*ReturnFunc)(void);
GSColor        LoginTextColor;
BigString      LoginBuffer;
char           *LoginName;

/*------------------------------------------------------------------------------
  login
    Called when a character is typed at the keyboard.
    Processes a character typed at the keyboard.
------------------------------------------------------------------------------*/
boolean login (
  int   mouse_x,  /*--- not used here, but required by event handler ---*/
  int   mouse_y,  /*--- not used here, but required by event handler ---*/
  char  ch        /*--- character typed at the keyboard ---*/
)
{
  int       text_x, text_y;
  GSColor   text_color;
  int       background_x, background_y, background_wd, background_ht;
  GSColor   background_color;
  int       length;
  int       cursor_x;
  BigString tmp_string;
  BigString save_window;
  GSColor   save_color;
  FontSizeType save_font;


  /*--- The following two assignments are to appease the great compiler ---*/
  mouse_x = mouse_x;
  mouse_y = mouse_y;

  /*--- Don't process character unless "login" is the current window. ---*/
  if (GSstricmp(GSgetcurrentwindowname(save_window), "login") != 0)
    return (FALSE);

  save_color = GSgetcolor();
  save_font = GSPgetfont();
  GSsetfont(LARGE);
  GSsetcurrentnamedwindow("login");
  GSgetfieldrect("login", &background_x, &background_y, &background_wd,
                 &background_ht, &background_color);
  text_x = background_x + 3 ;
  text_y = background_y + 14 ;
  text_color = LoginTextColor;

  if (ch == RETURN)
  {
    InfState.keyfunc = NULL;
    GSremovenamedwindow("login");
    LoginWindowUp = FALSE;
    GSsetcolor(save_color);
    GSsetfont(save_font);
    GSsetcurrentnamedwindow(save_window);
    if (ReturnFunc)
      return ((*ReturnFunc)());
    else
      return (TRUE);
  }
  else
  {
    if (isprint(ch))
    {
      length = strlen(LoginName);
      if (length < BIGSTRINGLEN)
      {
        strcpy(tmp_string, LoginName);
        tmp_string[length] = ch;
        tmp_string[length+1] = '\0';
        if (GStextwidth(tmp_string) <= background_wd-6)
        {
          strcpy(LoginName, tmp_string);
        }
      }
    }
    else if ((ch == '\b') && (strlen(LoginName) > 0))
    {
      (LoginName)[strlen(LoginName)-1] = '\0';
    }
    strcpy(tmp_string, LoginName);
    if (strlen(tmp_string) < BIGSTRINGLEN)
    {
      strcat(tmp_string, "_");
      if (GStextwidth(tmp_string) > background_wd-6)
      {
        tmp_string[strlen(tmp_string)-1] = '\0';
      }
    }
    GSsetcolor(background_color);
    GSwfillrect(background_x, background_y, background_wd, background_ht);
    GSsetcolor(text_color);
    GSwwritemsg(text_x, text_y, tmp_string);
    GSsetcurrentnamedwindow(save_window);
    GSsetcolor(save_color);
    GSsetfont(save_font);
    return TRUE;
  }
}

/*------------------------------------------------------------------------------
  GSdologinscreen
   Draws the ``login'' window and activates a built-in keyboard handler
   for a user login.  This is a convenient way of getting the user's name
   for program output, security checks, etc.  The application programmer
   must declare a window named ``login'' and a field named ``login'' in
   the interface file.  The login  window is treated as a popup window. 
   The login field is for user text entry; it is positioned relative to
   the login window.  Only keyboard input received while ``login'' is the
   current window is processed.
------------------------------------------------------------------------------*/
void GSdologinscreen (
  char      *prompt,
  int       bstore,
  boolean   (*return_func)(void),
  GSColor   login_text_color,
  BigString user_name
)
{
  int        x, y, wd, ht;
  GSColor    color;
  GSColor    save_color;
  FontSizeType save_font;
  int        text_x, text_y;
  BigString  tmp_string;

  if (user_name == NULL)
    LoginName = LoginBuffer;
  else
    LoginName = user_name;
  save_color = GSgetcolor();
  save_font = GSPgetfont();
  ReturnFunc = return_func;
  LoginTextColor = login_text_color;
  InfState.keyfunc = login;
  if (LoginWindowUp == TRUE)
  {
    GSredrawnamedwindow("login");
  }
  else
  {
    GSdrawnamedwindow("login", bstore, POPUP);
    LoginWindowUp = TRUE;
  }
  GSgetfieldrect("login", &x, &y, &wd, &ht, &color);
  GSsetcolor(color);
  GSwfillrect(x, y, wd, ht);
  GSsetfont(LARGE);
  GSsetcolor(Win[InfState.cw].fgcolor);
  GSwwritemsg(x+wd/2-GStextwidth(prompt)/2, y-10, prompt);
  GSframefield("login", "login", InfState.buttonBRcolor, InfState.buttonTLcolor);
/*  GSsetcolor (InfState.buttonBRcolor);
  GSwdrawline(x-1, y-1, x+wd, y-1);
  GSwdrawline(x-2, y-2, x+wd+1, y-2);
  GSwdrawline(x-1, y-1, x-1, y+ht);
  GSwdrawline(x-2, y-2, x-2, y+ht+1);
  GSsetcolor (InfState.buttonTLcolor);
  GSwdrawline(x, y+ht, x+wd, y+ht);
  GSwdrawline(x-1, y+ht+1, x+wd+1, y+ht+1);
  GSwdrawline(x+wd, y, x+wd, y+ht);
  GSwdrawline(x+wd+1, y-1, x+wd+1, y+ht+1);*/
  GSsetcolor(login_text_color);
  GSsetfont(LARGE);
  text_x = x + 3 ;
  text_y = y + 14 ;
  strcpy(tmp_string, LoginName);
  if (strlen(tmp_string) < BIGSTRINGLEN)
  {
    strcat(tmp_string, "_");
  }
  GSwwritemsg(text_x, text_y, tmp_string);
  GSsetcolor(save_color);
  GSsetfont(save_font);
}
