/***********************************************************************
Runlength encode an image file.
***********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "gsincl.h"

#define TAG4 0x844e5552L  /* ID tag: 'RUN' + 0x04000000 + 0x80000000L */
#define TAG8 0x884e5552L  /* ID tag: 'RUN' + 0x08000000 + 0x80000000L */
#define TAG12 0x8c4e5552L /* ID tag: 'RUN' + 0x0c000000 + 0x80000000L */
#define TAG16 0x904e5552L /* ID tag: 'RUN' + 0x10000000 + 0x80000000L */

int main(int argc, char **argv)
{
  unsigned long tag;
  unsigned short *headers;
  FILE *infp, *outfp;
  char *inbuff;
  short *shortbuff; /* short type alias for inbuff */
  int rows, cols;
  int outcount;
  unsigned short lastval;
  int runcount;
  unsigned char charval;
  unsigned short shortval;
  int i, j;
  int bpp, inbytes, outbytes;

  if ((argc != 4) && (argc != 6))
  {
    printf("Usage:\n");
    printf("For .npm file:\n");
    printf("\trencode <bits-per-pixel> <infile> <outfile>\n");
    printf("For header-less image:\n");
    printf("\trencode <bits-per-pixel> <infile> <outfile> <#rows> <#cols>\n");
    exit(0);
  }
  bpp = atoi(argv[1]);
  if ((bpp != 4) && (bpp != 8) && (bpp != 12) & (bpp != 16))
  {
    printf("Don't know how to encode an image with %d bits per pixel\n", bpp);
    exit(0);
  }
  if((infp = fopen(argv[2], "rb")) == NULL)
  {
    printf("Unable to open input file |%s|\n", argv[2]);
    exit(0);
  }
  if (argc == 6)
  {
    rows = atoi(argv[4]);
    cols = atoi(argv[5]);
  }
  else
  { /* .npm file */
    fread(&cols, 2, 1, infp);
    fread(&rows, 2, 1, infp);
  }
  if((outfp = fopen(argv[3], "wb")) == NULL)
  {
    printf("Unable to open output file |%s|\n", argv[3]);
    exit(0);
  }
  switch (bpp) {
    case 4: tag = TAG4; inbytes = 1; outbytes = 1; break;
    case 8: tag = TAG8; inbytes = 1; outbytes = 2; break;
    case 12: tag = TAG12; inbytes = 2; outbytes = 2; break;
    case 16: tag = TAG16; inbytes = 2; outbytes = 3; break;
    default:
      printf("Unknown tag type\n");
      exit(0);
  }
  fwrite(&tag, 4, 1, outfp);
  fwrite(&rows, 2, 1, outfp);
  fwrite(&cols, 2, 1, outfp);
  fseek(outfp, 2*(rows+1)+8, SEEK_SET); /* Skip space for row vector */
  if((inbuff = malloc(cols * inbytes)) == NULL)
  {
    printf("Unable to malloc input buffer\n");
    exit(0);
  }
  shortbuff = (short *)inbuff;
  headers = malloc((rows+1)*2);
  headers[0] = 8 + (rows+1)*2;  /* Byte position in file for the first row */
  for(i=0; i<rows; i++)
  { /* Process a row */
    fread(inbuff, inbytes, cols, infp);
    outcount = 0;
    lastval = -1;
    runcount = 0;
    for(j=0; j< cols; j++)
    { /* Process a pixel */
      switch (bpp) {
      case 4:
        if((inbuff[j] == lastval) && (outcount != 16))
          outcount++;
        else
        {
          if(outcount != 0)
          {
	    charval = (unsigned char)((outcount << 4) + lastval);
            if (fwrite(&charval, 1, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
            runcount++;
          }
          lastval = inbuff[j];
          outcount = 1;
        }
        break;
      case 8:
        if((inbuff[j] == lastval) && (outcount != 256))
          outcount++;
        else
        {
          if(outcount != 0)
          {
	    charval = (unsigned char)outcount % 256;
	    if (fwrite(&charval, 1, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
            charval = (unsigned char)lastval;
	    if (fwrite(&charval, 1, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
            runcount++;
          }
          lastval = inbuff[j];
          outcount = 1;
        }
        break;
      case 12:
        if((shortbuff[j] == lastval) && (outcount != 16))
          outcount++;
        else
        {
          if(outcount != 0)
          {
	    shortval = (unsigned short)((outcount << 12) |
					((unsigned short)lastval & 0xfff));
            if (fwrite(&shortval, 2, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
            runcount++;
          }
          lastval = shortbuff[j];
          outcount = 1;
        }
        break;
      case 16:
        if((shortbuff[j] == lastval) && (outcount != 256))
          outcount++;
        else
        {
          if(outcount != 0)
          {
            charval = outcount % 256;
            /* 1 byte for runlength ... */
            if (fwrite(&charval, 1, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
	    /* and 2 bytes for pixel value */
            if (fwrite(&lastval, 2, 1, outfp) != 1)
            {
              printf("Error writing\n");
              exit(0);
            }
            runcount++;
          }
          lastval = shortbuff[j];
          outcount = 1;
        }
        break;
      default:
        printf("Unknown pixel type\n");
        exit(0);
      }
    } /* Process a pixel */
    /* Output the last run */
    switch (bpp) {
    case 4:
      charval = (unsigned char)((outcount << 4) + lastval);
      if (fwrite(&charval, 1, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      break;
    case 8:
      charval = (unsigned char)outcount % 256;
      if (fwrite(&charval, 1, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      charval = (unsigned char)lastval;
      if (fwrite(&charval, 1, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      break;
    case 12:
      shortval = (unsigned short)((outcount << 12) |
                  ((unsigned short)lastval & 0xffffff));
      if (fwrite(&shortval, 2, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      break;
    case 16:
      outcount = outcount % 256;
      /* 1 byte for runlength ... */
      if (fwrite(&outcount, 1, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      /* and 2 bytes for pixel value */
      if (fwrite(&lastval, 2, 1, outfp) != 1)
      {
        printf("Error writing\n");
        exit(0);
      }
      break;
    default:
      printf("Unknown pixel type\n");
      exit(0);
    }
    runcount++;
    headers[i+1] = runcount * outbytes;
  } /* Process a row */
  fseek(outfp, 8, SEEK_SET);
  fwrite(headers, 2, (rows+1), outfp);
  return(TRUE);
}
