#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "geosim.h"
#include "gspriv.h"
#include "gsscreen.h"


/* These definitions should become variables if we
   support different resolutions, numcolors, etc.*/

#define BITSPERPIXEL       4
/*#define NUMCOLORS          16*/
static int kNUMCOLORS = 16;
/*#define COLORRES           6 */   /* number of bits for a primary color */
static kCOLORRES = 6;
                                /* in vga, primary colors range from 0 -63, hence 6 bits */

#define BM                 19778    /* Is this defined elsewhere in a Windows header? -Bo*/

extern char Interface_Palette[];
extern Iconfig InfState;

/********************************************************************
 Dumps screen to filename
 fname contains the prefix of the file to save to (eg., ./migmod).
 The name of the file saved is returned in fname (eg., ./migmod01.bmp).
 ********************************************************************/
boolean GSdumpscreen(
  char * fname,        /* prefix of file name to which to save. and ...*/
		               /* actual filename opened is returned in fname */
  GraphicsFileType filetype)    /* type of graphics file to create */
{
  FILE *fp;                     /* stream to save screen to. */
  boolean dumpsuccess = TRUE;   /* return value of GSscreen_to_bmp() */
  BigString tempfname;
  BigString extension;
  int i;

  /* fill the extension string with the appropriate characters */
  switch (filetype)
  {
    case BMP_16   : strcpy(extension, ".bmp"); break;
    case PS_COLOR :
    case PS_GRAY  : strcpy(extension, ".ps"); break;
    case GIF_16   : strcpy(extension, ".gif"); break;
    case NO_FILE  :
    default      : GSelog(0, "Invalid screen save file type.\n");
                   return FALSE;
  }

  /*--- open output file.  Start at fname00.bmp and keep trying until
        fname99.bmp.  Then fail if file not open. ---*/
  for(i = 0; i < 100; i++)
  {
    sprintf(tempfname, "%s%02d%s", fname, i, extension);
    if ((fp = GSopen(tempfname, "r")) == NULL)
    {
      if ((filetype != GIF_16))
      {
        if ((fp = GSopen(tempfname, "wb")) == NULL)
        {
          GSelog(0, "Cannot open graphics output file |%s|.\n", tempfname);
          dumpsuccess = FALSE;
        }
      }
      else
      {
        if ((GIF_Create(tempfname, SCREENWIDTH, SCREENHEIGHT, kNUMCOLORS,
            kCOLORRES)) != GIF_OK)
        {
          GSelog(0, "Cannot open graphics output file |%s|.\n", tempfname);
          dumpsuccess = FALSE;
        }
      }
      strcpy(fname, tempfname);
      break;
    }
    else fclose(fp);
  }

  if (dumpsuccess)                  /* if file openned successfully */
  {
    switch (filetype)
    {
      case BMP_16 :
        dumpsuccess = GSPscreen_to_bmp(fp);
        break;
      case PS_COLOR :
        dumpsuccess = GSPscreen_to_ps(fp, TRUE);
        break;
      case PS_GRAY :
        dumpsuccess = GSPscreen_to_ps(fp, FALSE);
        break;
      case GIF_16 :
        dumpsuccess = GSPscreen_to_gif();
        break;
      case NO_FILE : default :
        dumpsuccess = FALSE;
        GSelog(0, "Invalid screen save file type.\n");
    };
  }
  if (!dumpsuccess)
  {
    GSelog(0, "Error:  Screen not copied to output file.\n");
  }

  if ((filetype != GIF_16))
    fclose(fp);
  return dumpsuccess;
}

static GSDWORD reverse_bytes_dword(GSDWORD thedword)
{
  char 	*tempchar;
  GSBYTE 	tempbyte;

  tempchar = (char *)&thedword;
  tempbyte = tempchar[3];	/* swap first and last bytes */
  tempchar[3] = tempchar[0];
  tempchar[0] = tempbyte;
  tempbyte = tempchar[1];	/* swap middle two bytes */
  tempchar[1] = tempchar[2];
  tempchar[2] = tempbyte;

  return thedword;
}
 
static GSWORD   reverse_bytes_word(GSWORD theword)
{
  char 	*tempchar;
  GSBYTE 	tempbyte;

  tempchar = (char *)&theword;
  tempbyte = tempchar[1];	/* swap the two bytes */
  tempchar[1] = tempchar[0];
  tempchar[0] = tempbyte;

  return theword;
} 

/********************************************************************
 Dumps screen to Windows 3.1 DIB file
 GeoSim Private
 ********************************************************************/
boolean GSPscreen_to_bmp( FILE *outfile )
{
  GSBITMAPFILEHEADER  bmfh;
  GSBITMAPINFOHEADER  bmih;
  GSRGBQUAD          *aColors;
  GSBYTE              aBitmapBits[SCREENWIDTH/2 + SCREENWIDTH%2];
  GSBYTE              rowBuff[SCREENWIDTH];
  GSBYTE              pix1, pix2;
  GSBYTE              tempbyte;
  int               i, j;
  GSDWORD             tempsize;
  GSColor           savecolor;
  BigString         CurrentWindow;

  savecolor = GSgetcolor();
  GSgetcurrentwindowname(CurrentWindow);

  /* Fill the GSBITMAPINFOHEADER structure */
  bmih.biSize = (GSbigendian()) ? reverse_bytes_dword(sizeof(GSBITMAPINFOHEADER))
                             : sizeof(GSBITMAPINFOHEADER);
  bmih.biWidth =  (GSbigendian()) ? reverse_bytes_dword(SCREENWIDTH):SCREENWIDTH;
  bmih.biHeight = (GSbigendian()) ? reverse_bytes_dword(SCREENHEIGHT):SCREENHEIGHT;
  bmih.biPlanes = (GSbigendian()) ? reverse_bytes_word(1):1;
  bmih.biBitCount = (GSbigendian()) ? reverse_bytes_word(BITSPERPIXEL):BITSPERPIXEL;
  bmih.biCompression = (GSbigendian()) ? reverse_bytes_dword(BI_RGB):BI_RGB;
  bmih.biSizeImage = (GSbigendian()) ?
       reverse_bytes_dword((long)((long)SCREENWIDTH*(long)SCREENHEIGHT) /2)
     : ((long)((long)SCREENWIDTH*(long)SCREENHEIGHT) /2);
  bmih.biXPelsPerMeter = (GSbigendian()) ? reverse_bytes_dword(0):0;
  bmih.biYPelsPerMeter =  (GSbigendian()) ? reverse_bytes_dword(0):0;
  bmih.biClrUsed = (GSbigendian()) ? reverse_bytes_dword(16):16;
  bmih.biClrImportant = (GSbigendian()) ? reverse_bytes_dword(16):16;

  /* Fill the GSBITMAPFILEHEADER structure */
  bmfh.bfType = (GSbigendian()) ? reverse_bytes_word(BM):BM;
  tempsize = sizeof(GSBITMAPFILEHEADER) + sizeof(GSBITMAPINFOHEADER)
		+ sizeof(GSRGBQUAD)*(16) + bmih.biSizeImage;
  bmfh.bfSize = (GSbigendian()) ? reverse_bytes_dword(tempsize):tempsize;
  bmfh.bfReserved1 = (GSbigendian()) ? reverse_bytes_word(0):0;
  bmfh.bfReserved2 = (GSbigendian()) ? reverse_bytes_word(0):0;
  tempsize = sizeof(GSBITMAPFILEHEADER) + sizeof(GSBITMAPINFOHEADER)
		   + sizeof(GSRGBQUAD)*(16);
  bmfh.bfOffBits =  (GSbigendian()) ? reverse_bytes_dword(tempsize):tempsize;

  /* Store the palette information in aColors */
  aColors = (GSRGBQUAD *) malloc(sizeof(GSRGBQUAD)*16);
  for (i=0; i<16; i++)
  {
    aColors[i].rgbRed = (GSBYTE)Interface_Palette[i*3]<<2;
    aColors[i].rgbGreen = (GSBYTE)Interface_Palette[i*3+1]<<2;
    aColors[i].rgbBlue = (GSBYTE)Interface_Palette[i*3+2]<<2;
    aColors[i].rgbReserved = 0;
  }

  /* Write the three headers to the file */
  if ( (fwrite(&bmfh.bfType, sizeof(GSWORD), 1, outfile) != 1) )
  {
    GSelog(0, "Error writing Bit Map File Header to file.\n");
    return FALSE;
  }
   if ( (fwrite(&bmfh.bfSize, sizeof(GSDWORD), 1, outfile) != 1) )
  {
    GSelog(0, "Error writing Bit Map File Header to file.\n");
    return FALSE;
  }
  if ( (fwrite(&bmfh.bfReserved1, sizeof(GSWORD), 1, outfile) != 1) )
  {
    GSelog(0, "Error writing Bit Map File Header to file.\n");
    return FALSE;
  }
  if ( (fwrite(&bmfh.bfReserved2, sizeof(GSWORD), 1, outfile) != 1) )
  {
    GSelog(0, "Error writing Bit Map File Header to file.\n");
    return FALSE;
  }
  if ( (fwrite(&bmfh.bfOffBits, sizeof(GSDWORD), 1, outfile) != 1) )
  {
    GSelog(0, "Error writing Bit Map File Header to file.\n");
    return FALSE;
  }
 if ((fwrite(&bmih, sizeof(GSBITMAPINFOHEADER), 1, outfile) != 1))
  {
    GSelog(0, "Error writing Bit Map Info Header to file.\n");
    return FALSE;
  }
  if ((fwrite(aColors, (sizeof(GSRGBQUAD)*16), 1, outfile) != 1))
  {
    GSelog(0, "Error writing Bit Map palette structure to file.\n");
    return FALSE;
  }

  free(aColors);

  GSelog(0, "Struct Sizes:\n bmfh = %d, bmih = %d, ",
         sizeof(GSBITMAPFILEHEADER), sizeof(GSBITMAPINFOHEADER));
  GSelog(0, "acolors = %d\nDWORD = %d, GSWORD = %d, GSBYTE = %d\n",
         (sizeof(GSRGBQUAD)*16), sizeof(GSDWORD), sizeof(GSWORD), sizeof(GSBYTE));
  GSsetcolor(GSconvertcolor("white"));
  GSsetcurrentnamedwindow("root");

  /* Dump screen to file.  Start in lower left corner. */
  for(i = SCREENHEIGHT - 1; i >= 0; i--)
  {
    GSPget_row(i, rowBuff);
    for (j = 0; j < SCREENWIDTH/2; j++)
    {
      aBitmapBits[j] = ((rowBuff[2*j] << 4) | (rowBuff[2*j+1]));
    }

#if 0  /* include the following code if SCREENWIDTH becomes a variable */
    if (SCREENWIDTH%2)                      /* if screen width is odd */
    {
      aBitmapBits[j] = (GSBYTE)GSPget_pixel(j*2, i);
    }
#endif

    for (j = 0; j < 5; j++)
    { /*--- draw "stripes" up the side of the window to show progess ---*/
      GSwdrawpoint(j, i);
      GSwdrawpoint(SCREENWIDTH-j, i);
    }

    /* Write this row to the file */
    if ((fwrite(aBitmapBits, sizeof(aBitmapBits), 1, outfile) != 1))
    {
      GSelog(0, "Error writing Bit Map pixels to file.\n");
      return FALSE;
    }
  }

  GSsetcolor(savecolor);
  GSsetcurrentnamedwindow(CurrentWindow);
  InfState.redrawfunc();
  return TRUE;
}

#define PSCOLORPREAMBLE \
"%!PS-Adobe-2.0 EPSF-2.0\n" \
"%%Creator: GeoSim Interface Library -- Project GeoSim\n" \
"%%BoundingBox: 66 76 546 716\n" \
"%%Pages: 1\n" \
"%%DocumentFonts:\n" \
"%%EndComments\n" \
"%%EndProlog\n\n" \
"%%Page: 1 1\n\n" \
"% remember original state\n" \
"/origstate save def\n\n" \
"% build a temporary dictionary\n" \
"20 dict begin\n\n" \
"% define string to hold a scanline's worth of data\n" \
"/pix 1920 string def\n\n" \
"% print in landscape mode\n" \
"90 rotate 0 -612 translate\n\n" \
"% lower left corner\n" \
"76 66 translate\n\n" \
"% size of image (on paper, in 1/72inch coords)\n" \
"640.00800 480.02400 scale\n\n" \
"% define 'colorimage' if it isn't defined\n" \
"%   ('colortogray' and 'mergeprocs' come from xwd2ps\n" \
"%     via xgrab)\n" \
"/colorimage where   % do we know about 'colorimage'?\n" \
"  { pop }           % yes: pop off the 'dict' returned\n" \
"  {                 % no:  define one\n" \
"    /colortogray {  % define an RGB->I function\n" \
"      /rgbdata exch store    % call input 'rgbdata'\n" \
"      rgbdata length 3 idiv\n" \
"      /npixls exch store\n" \
"      /rgbindx 0 store\n" \
"      /grays npixls string store  % str to hold the result\n" \
"      0 1 npixls 1 sub {\n" \
"        grays exch\n" \
"        rgbdata rgbindx       get 20 mul    % Red\n" \
"        rgbdata rgbindx 1 add get 32 mul    % Green\n" \
"        rgbdata rgbindx 2 add get 12 mul    % Blue\n" \
"        add add 64 idiv        % I = .5G + .31R + .18B\n" \
"        put\n" \
"        /rgbindx rgbindx 3 add store\n" \
"      } for\n" \
"      grays\n" \
"    } bind def\n\n" \
"    % Utility procedure for colorimage operator.\n" \
"    % This procedure takes two procedures off the\n" \
"    % stack and merges them into a single procedure.\n" \
"    /mergeprocs { % def\n" \
"      dup length\n" \
"      3 -1 roll\n" \
"      dup\n" \
"      length\n" \
"      dup\n" \
"      5 1 roll\n" \
"      3 -1 roll\n" \
"      add\n" \
"      array cvx\n" \
"      dup\n" \
"      3 -1 roll\n" \
"      0 exch\n" \
"      putinterval\n" \
"      dup\n" \
"      4 2 roll\n" \
"      putinterval\n" \
"    } bind def\n\n" \
"    /colorimage { % def\n" \
"      pop pop     % remove 'false 3' operands\n" \
"      {colortogray} mergeprocs\n" \
"      image\n" \
"    } bind def\n" \
"  } ifelse          % end of 'false' case\n\n\n\n" \
"640 480 8                       % dimensions of data\n" \
"[640 0 0 -480 0 480]            % mapping matrix\n" \
"{currentfile pix readhexstring pop}\n" \
"false 3 colorimage\n\n"

#define PSGREYPREAMBLE \
"%!PS-Adobe-2.0 EPSF-2.0\n" \
"%%Creator: GeoSim Interface Library -- Project GeoSim\n" \
"%%BoundingBox: 66 76 546 716\n" \
"%%Pages: 1\n" \
"%%DocumentFonts:\n" \
"%%EndComments\n" \
"%%EndProlog\n\n" \
"%%Page: 1 1\n\n" \
"% remember original state\n" \
"/origstate save def\n\n" \
"% build a temporary dictionary\n" \
"20 dict begin\n\n" \
"% define string to hold a scanline's worth of data\n" \
"/pix 640 string def\n\n" \
"% print in landscape mode\n" \
"90 rotate 0 -612 translate\n\n" \
"% lower left corner\n" \
"76 66 translate\n\n" \
"% size of image (on paper, in 1/72inch coords)\n" \
"640.00800 480.02400 scale\n\n" \
"640 480 8                       % dimensions of data\n" \
"[640 0 0 -480 0 480]            % mapping matrix\n" \
"{currentfile pix readhexstring pop}\n" \
"image\n\n" \


#define PSPOSTAMBLE \
"showpage\n\n" \
"% stop using temporary dictionary\n" \
"end\n\n" \
"% restore original state\n" \
"origstate restore\n\n" \
"%%Trailer\n"


#define PS_HEX_PERLINE 33

/********************************************************************
 Dumps screen to color PostScript level 2 file
 GeoSim Private
 ********************************************************************/
boolean GSPscreen_to_ps( FILE *outfile,     /* File pointer to output file */
                         boolean use_color) /* TRUE iff output should be color PostScript, otherwise */
                                            /* output will be greyscale postscript */
{
  GSColor   savecolor;         /* save and restore the old current color */
  BigString CurrentWindow;     /* save and restore the old current window */
  int       i, j;              /* counters */
  int       outputcolumn = 0;  /* to keep the output lines to PS_HEX_RGB_PERLINE
                                  chars */
  int       pixcolorindex;     /* index of pixel */
  BigString outpixelhex;       /* output to ps file.  Will contain RGB in color
                                  mode, or grey value in greyscale mode */
  GSColor   rowbuf[SCREENWIDTH];

  savecolor = GSgetcolor();
  GSgetcurrentwindowname(CurrentWindow);
  GSsetcolor(GSconvertcolor("white"));
  GSsetcurrentnamedwindow("root");

  if (use_color)
  {
    if ((fputs(PSCOLORPREAMBLE, outfile) == EOF))
    {
      GSelog(0, "Error writing preamble to color PostScript file.\n");
      return FALSE;
    }
  }
  else
  {
    if ((fputs(PSGREYPREAMBLE, outfile) == EOF))
    {
      GSelog(0, "Error writing preamble to greyscale PostScript file.\n");
      return FALSE;
    }
  }

  /* Dump screen to file.  Start in upper right corner. */
  for(i = 0; i < SCREENHEIGHT; i++)
  {
    GSPget_row(i, rowbuf);
    for(j=0; j<SCREENWIDTH; j++)
    {
/*      pixcolorindex = GSwgetpoint(j, i);********/
      pixcolorindex = rowbuf[j];
      if (use_color)
      {
        sprintf(outpixelhex, "%02x%02x%02x",
                Interface_Palette[pixcolorindex*3]<<2,
                Interface_Palette[pixcolorindex*3+1]<<2,
                Interface_Palette[pixcolorindex*3+2]<<2);
      }
      else
      {
        /*  I = .31R + .5G + .18B */
        sprintf(outpixelhex, "%02x",
                (int)(0.31*((int)Interface_Palette[pixcolorindex*3]<<2)
              + 0.5*((int)Interface_Palette[pixcolorindex*3+1]<<2)
              + 0.18*((int)Interface_Palette[pixcolorindex*3+2]<<2)));
      }

      if ((fputs(outpixelhex, outfile) == EOF ))
      {
        GSelog(0, "Error writing pixel RGB values to PostScript file.\n");
        return FALSE;
      }

      outputcolumn += use_color?3:1;
      if ((outputcolumn %= PS_HEX_PERLINE) == 0)
        /* put a newline at the end of this text row in the output file */
        if ((fputs("\n", outfile) == EOF))
        {
          GSelog(0, "Error writing newline character to PostScript file.\n");
          return FALSE;
        }
    }

    outputcolumn = 0;
    if ((fputs("\n", outfile) == EOF))        /* put a newline at the end of this screen row's rgb values */
    {
      GSelog(0, "Error writing newline character to PostScript file.\n");
      return FALSE;
    }

    for (j=0; j<5; j++)
    { /*--- draw "stripes" up the side of the window to show progess ---*/
      GSwdrawpoint(j, i);
      GSwdrawpoint(SCREENWIDTH-j, i);
    }
  }

  if ((fputs(PSPOSTAMBLE, outfile) == EOF))
  {
    GSelog(0, "Error writing postamble to PostScript file.\n");
    return FALSE;
  }

  GSsetcolor(savecolor);
  GSsetcurrentnamedwindow(CurrentWindow);
  InfState.redrawfunc();
  return TRUE;
}


/********************************************************************
 Utility function that returns pixel color value of screen
 coord (x, y)
 ********************************************************************/
static int gpixel(int x, int y)
{
  static GSColor rowbuf[SCREENWIDTH]; /* for retrieving a row of pixels */
  int            temppixelvalue;      /* temporary holder of (x, y) value */
  int            j;                   /* counter */

  if (x == 0)
  { /*--- collect a new row of pixels if this is the start of a row ---*/
    GSPget_row(y, rowbuf);
    if (y > 0)
    { /*--- if y = 0, then y-1 = -1, which is not valid row ---*/
      for (j=0; j<5; j++)
      { /*--- draw "stripes" up the side of the window to show progess ---*/
        GSwdrawpoint(j, y-1);
        GSwdrawpoint(SCREENWIDTH-j, y-1);
      }
    }
  }
  temppixelvalue = rowbuf[x];
  return (int)temppixelvalue;
}


/********************************************************************
 Dumps screen to GIF 16 color GIF87a format
 GeoSim Private
 ********************************************************************/
boolean GSPscreen_to_gif(void)
{
  int i;                    /* counter */
  GSColor   savecolor;      /* save and restore the old current color */
  BigString CurrentWindow;  /* save and restore the old current window */
  boolean   success = FALSE;/* return value of this function */

  savecolor = GSgetcolor();
  GSgetcurrentwindowname(CurrentWindow);
  GSsetcolor(GSconvertcolor("white"));
  GSsetcurrentnamedwindow("root");

  for (i = 0; i < kNUMCOLORS; i++)
      GIF_SetColor(i, (int)Interface_Palette[i*3],         /* red[i]   */
                      (int)Interface_Palette[i*3+1],       /* green[i] */
                      (int)Interface_Palette[i*3+2]);      /* blue[i]  */

  if ((GIF_CompressImage(0, 0, -1, -1, gpixel)) == GIF_OK)
  {
    if ((GIF_Close()) == GIF_OK)
      success = TRUE;
    else
      GSelog(0, "Error closing GIF file.\n");
  }
  else
    GSelog(0, "Error compressing GIF file.\n");

  GSsetcolor(savecolor);
  GSsetcurrentnamedwindow(CurrentWindow);
  InfState.redrawfunc();
  return success;
}
