#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "geosim.h"
#include "gspriv.h"

typedef struct {
	char *wname;
	char *daname;
	int fx, fy, fwd, fht; /* dims of sclist field */
	GSColor buttonCol, topLeftCol, botRightCol, fieldCol;
	int numListItems, numVisibleItems;
	int numCols;
	int curPos;
	void (*sclistDraggedFunc)(int);
	char *upbtn, *dwnbtn;
	float /*scaleVal, */pixPerItem;
	GSlistHandle sclistList;
	ButtonStatusType upstate, dwnstate;
	} GSsclistType;
	
#define MAX_THUMBS 5
#define FIRST_HANDLE 1000

/* Internal protos */
void AdjustButtons(GSsclistHandle sclist, boolean redraw);
GSsclistHandle getsclisthandle (void);
int thPSizeFunc(GSlistHandle list, int);
/**/

static GSsclistType *Thumb[MAX_THUMBS];
static boolean FirstTime = TRUE; /* kludge way to init Thumb[] */

extern Iconfig InfState;
extern GSWindow *Win;

void AdjustButtons(GSsclistHandle sclist, boolean redraw)
{
	ButtonStatusType up, down, newup, newdown;
	GSsclistType *cur = Thumb[sclist-FIRST_HANDLE];
	
	newup = (cur->curPos > 0)? ENABLED : DISABLED;
	newdown = (cur->curPos < (cur->numListItems) - (cur->numVisibleItems))? 
			ENABLED : DISABLED;
	
	if (cur->upbtn)
		GSgetbuttonspecs(cur->wname, cur->upbtn, NULL, NULL, NULL, NULL, &up);
	if (cur->dwnbtn)
		GSgetbuttonspecs(cur->wname, cur->dwnbtn, NULL, NULL, NULL, NULL, &down);
	
	if ((up != newup)&&(cur->upbtn)) {
		if (newup == ENABLED)
			GSenablebutton(cur->wname, cur->upbtn, redraw);
		else
			GSdisablebutton(cur->wname, cur->upbtn, redraw);
		}
	if ((down != newdown)&&(cur->dwnbtn)) {
		if (newdown == ENABLED)
			GSenablebutton(cur->wname, cur->dwnbtn, redraw);
		else
			GSdisablebutton(cur->wname, cur->dwnbtn, redraw);
		}
}

static boolean validhandle(GSsclistHandle sclist)
{
	return ((sclist>=FIRST_HANDLE) && (sclist<FIRST_HANDLE+MAX_THUMBS));
}

GSsclistHandle getsclisthandle (void)
{
	int i;
	
	if (FirstTime) {
		for (i=0; i<MAX_THUMBS; i++)
			Thumb[i] = NULL;
		FirstTime = FALSE;
		}
	
	for (i=0; i<MAX_THUMBS; i++) {
	
		if (Thumb[i]==NULL) {
			Thumb[i] = (GSsclistType*)malloc(sizeof(GSsclistType));
			if (Thumb[i] == NULL)
				GSelog(-1, "Out of memory for sclist.\n");
			else
				return (GSsclistHandle) (i+FIRST_HANDLE);
			}
		}
	GSelog(-1, "Reached maximum number of sclists.\n");
	return 0;
}

void GSdismantlesclist(GSsclistHandle sclist)
{
	if (validhandle(sclist)) {
		if (Thumb[sclist-FIRST_HANDLE]->sclistList)
			GSdismantlelist(Thumb[sclist-FIRST_HANDLE]->sclistList);
		free(Thumb[sclist-FIRST_HANDLE]);
		Thumb[sclist-FIRST_HANDLE] = NULL;
		}
	else
		GSelog(-1, "GSdismantlesclist called with invalid handle.\n");
}

GSsclistHandle GSscrollbarsetup(
	char *wname,
	char *daname,
	GSColor buttonCol, 
	GSColor upLeftCol, 
	GSColor botRightCol,
	int numListItems, int numVisibleItems,
	int cols,
	int curPos,
	void (*sclistDraggedFunc)(int),
	char *upbtn, char *dwnbtn
	)
{
	GSsclistHandle retHandle = getsclisthandle(), sclist;

	sclist = retHandle - FIRST_HANDLE;	
	
	Thumb[sclist]->wname = wname;
	Thumb[sclist]->daname = daname;
	Thumb[sclist]->buttonCol = buttonCol;
	Thumb[sclist]->topLeftCol = upLeftCol;
	Thumb[sclist]->botRightCol = botRightCol;
	Thumb[sclist]->numListItems = numListItems;
	Thumb[sclist]->numVisibleItems = numVisibleItems;
	Thumb[sclist]->numCols = cols;
	if (curPos > numListItems - numVisibleItems)
		curPos = MAX(0, numListItems - numVisibleItems);
		
	Thumb[sclist]->curPos = curPos;
	Thumb[sclist]->sclistDraggedFunc = sclistDraggedFunc;
	Thumb[sclist]->upbtn = upbtn;
	Thumb[sclist]->dwnbtn = dwnbtn;

	GSgetdragareaspecs(wname, daname, 
				&Thumb[sclist]->fx,
				&Thumb[sclist]->fy,
				&Thumb[sclist]->fwd,
				&Thumb[sclist]->fht,
				NULL,
				&Thumb[sclist]->fieldCol);

	Thumb[sclist]->pixPerItem = ((float)Thumb[sclist]->fht)/
		((float)((MAX(Thumb[sclist]->numListItems,Thumb[sclist]->numVisibleItems))));
	Thumb[sclist]->sclistList = -1;
	
	AdjustButtons(retHandle, TRUE);
	
	return retHandle;
}

void GSdrawsclist(GSsclistHandle sclist)
{
	int tx, ty, twd, tht;
	GSColor black = GSconvertcolor("black");
	GSsclistType *cur;
	int fx, fy, fwd, fht;
	
	if (!validhandle(sclist))
		GSelog(-1, "Invalid GSsclistHandle passed to GSdrawsclist.\n");
	
	cur = Thumb[sclist-FIRST_HANDLE];
	fx = cur->fx;
	fy = cur->fy;
	fwd = cur->fwd;
	fht = cur->fht;
	
	GSsetcurrentnamedwindow(cur->wname);
	GSsetcolor(cur->fieldCol);
	GSwfillrect(fx, fy, fwd, fht);
	
	GSsetlinesize(1);
	GSsetcolor(cur->botRightCol);
	GSwdrawline(fx, fy, fx+fwd, fy);
	GSwdrawline(fx, fy, fx, fy+fht);
	GSsetcolor(cur->topLeftCol);
	GSwdrawline(fx+fwd, fy, fx+fwd, fy+fht);
	GSwdrawline(fx, fy+fht, fx+fwd, fy+fht);
	
/*	Calc sclist dims */
	tx = fx;
	twd = fwd;
	ty = fy + (int) ((float)cur->curPos * cur->pixPerItem);
	tht = (int) ((float)cur->numVisibleItems * cur->pixPerItem);
	
	GSsetcolor(cur->buttonCol);
	GSwfillrect(tx, ty, twd, tht);
	
	GSsetcolor(cur->topLeftCol);
	GSwdrawline(tx, ty, tx+twd, ty);
	GSwdrawline(tx, ty, tx, ty+tht);
	GSsetcolor(cur->botRightCol);
	GSwdrawline(tx+twd, ty, tx+twd, ty+tht);
	GSwdrawline(tx, ty+tht, tx+twd, ty+tht);

	if (Thumb[sclist-FIRST_HANDLE]->sclistList >= 0)
		GSlistfromitem(Thumb[sclist-FIRST_HANDLE]->sclistList, 
								Thumb[sclist-FIRST_HANDLE]->curPos);
}

void GSsetsclistnumitems(GSsclistHandle sclist, int numItems, boolean redraw)
{
	if (!validhandle(sclist))
		GSelog(-1, "Invalid GSsclistHandle passed to GSsetsclistnumitems.\n");
	
	if (numItems == Thumb[sclist-FIRST_HANDLE]->numListItems)
		return;
		
	Thumb[sclist-FIRST_HANDLE]->numListItems = numItems;
	if (Thumb[sclist-FIRST_HANDLE]->curPos > 
				numItems - Thumb[sclist-FIRST_HANDLE]->numVisibleItems) {
		Thumb[sclist-FIRST_HANDLE]->curPos
			 = MAX(0,numItems - Thumb[sclist-FIRST_HANDLE]->numVisibleItems);
		}
		
	Thumb[sclist-FIRST_HANDLE]->pixPerItem = ((float)Thumb[sclist-FIRST_HANDLE]->fht)/
		((float)((MAX(Thumb[sclist-FIRST_HANDLE]->numListItems,
				Thumb[sclist-FIRST_HANDLE]->numVisibleItems))));
	AdjustButtons(sclist, TRUE);
	
	if (redraw)
		GSdrawsclist(sclist);
}

void GSsetsclistpos(GSsclistHandle sclist,int newPos, boolean redraw)
{
	if (!validhandle(sclist))
		GSelog(-1, "Invalid GSsclistHandle passed to GSsetsclistpos.\n");
	
	if (newPos > Thumb[sclist-FIRST_HANDLE]->numListItems - Thumb[sclist-FIRST_HANDLE]->numVisibleItems)
		newPos = MAX(0, Thumb[sclist-FIRST_HANDLE]->numListItems - Thumb[sclist-FIRST_HANDLE]->numVisibleItems);
	else if (newPos < 0)
		newPos = 0;
		
	if (newPos == Thumb[sclist-FIRST_HANDLE]->curPos)
		return;
	
	Thumb[sclist-FIRST_HANDLE]->curPos = newPos;
	
	if (Thumb[sclist-FIRST_HANDLE]->sclistDraggedFunc)
		(Thumb[sclist-FIRST_HANDLE]->sclistDraggedFunc)(newPos);
	
	AdjustButtons(sclist, TRUE);
	
	if (redraw) {
		GSdrawsclist(sclist);
		}
}

int GSgetsclistpos(GSsclistHandle sclist)
{
	return Thumb[sclist-FIRST_HANDLE]->curPos;
}

int GSgetmaxpos(GSsclistHandle sclist)
{
	return MAX(0,Thumb[sclist-FIRST_HANDLE]->numListItems-Thumb[sclist-FIRST_HANDLE]->numVisibleItems);
}

void GShandlesclist(GSsclistHandle sclist, 
				int relx, int rely, DragStatusType status)
{
	int fx, fy, fwd, fht;
	int ty, tht;
	GSsclistType *cur; 
	int nc;
	
	if (!validhandle(sclist))
		GSelog(-1, "Invalid GSsclistHandle passed to GShandlesclist.\n");
	
	cur = Thumb[sclist-FIRST_HANDLE];
	nc = Thumb[sclist-FIRST_HANDLE]->numCols;
	
	fx = cur->fx;
	fy = cur->fy;
	fwd = cur->fwd;
	fht = cur->fht;
	ty = fy + (int) ((float)cur->curPos * cur->pixPerItem);
	tht = (int) ((float)cur->numVisibleItems * cur->pixPerItem);
	
	if (tht >= fht)
		return;  /* nothing to do */

	switch(status) {
	
		case DRAG_INIT:
		case DRAG_FINAL:
			break;
			
		case DRAG_PROCESS:
			if ((rely >= 0) && (rely <= fht)) { 
				GSsetsclistpos(sclist,
							  ((int)( (float)(rely-(tht/2)) / cur->pixPerItem)/nc)*nc,
							  	/* round down to multiple of nc */
							  TRUE);
				}
			break;
			
		case MOUSE_CLICK:
			if (rely+fy < ty)
				GSsetsclistpos(sclist, 
							  cur->curPos - (cur->numVisibleItems) + nc,
							  TRUE);
			else if (rely+fy > ty + tht)
				GSsetsclistpos(sclist, 
							  cur->curPos + (cur->numVisibleItems) - nc,
							  TRUE);
			break;
		}
}

void GShandleupbutton(GSsclistHandle sclist)
{
	GSsetsclistpos(sclist, Thumb[sclist-FIRST_HANDLE]->curPos
					-Thumb[sclist-FIRST_HANDLE]->numCols, TRUE);
}

void GShandledownbutton(GSsclistHandle sclist)
{
	GSsetsclistpos(sclist, Thumb[sclist-FIRST_HANDLE]->curPos
					+Thumb[sclist-FIRST_HANDLE]->numCols, TRUE);
}

void GSdisablesclistbuttons(GSsclistHandle sclist)
{
	GSsclistType *cur = Thumb[sclist-FIRST_HANDLE];
	
	if (cur->upbtn) {
		GSdisablebutton(cur->wname, cur->upbtn, TRUE);
		}
	if (cur->dwnbtn) {
		GSdisablebutton(cur->wname, cur->dwnbtn, TRUE);
		}
	
}

void GSenablesclistbuttons(GSsclistHandle sclist)
{
	AdjustButtons(sclist, TRUE);
}	

int thPSizeFunc(GSlistHandle list, int unused)
{
	/* find sclist with this list - could be better */

	int i;
	int nc;
	for (i=0; i < MAX_THUMBS; i++) {
		if (Thumb[i]->sclistList == list)
			break;
		}
	
	if (i >= MAX_THUMBS)
		GSelog(-1, "Error:  thPSizeFunc called by list with no sclist.\n");
	
	nc = Thumb[i]->numCols;
	
	return (MIN(Thumb[i]->numVisibleItems,Thumb[i]->numListItems));
}

/*
	This function holds the record for the 
	# of params of a GIL call - 23.
*/
void GSsclistsetup(
	char	*wname,			/*--- name of window holding list and sclist ---*/
	char	*daname,		/*--- name of drag-area holding list ---*/
	int	cols,			/*--- Number of columns on page ---*/
	char	*labelfield,		/*--- Field descriptor for a label ---*/
	int	colgap,			/*--- Gap (in pixels) between columns ---*/
	int	rowgap,			/*--- Gap (in pixels) between rows ---*/
	GSjustType just,		/*-- justification of labels --*/	
	char	*(*label_func)(int),	/*--- returns requested label ---*/
	void	(*settext_func)(GSlistHandle, int), /* sets text attributes for items */
	void	(*in_func)(int),	/*--- response when cursor in list ---*/
	void	(*out_func)(void),	/*--- response when cursor leaves list ---*/
	void	(*click_func)(int),	/*--- response when user clicks ---*/
	char *thDaname,
	GSColor buttonCol, GSColor topLeftCol, GSColor botRightCol,
	int numListItems,
	int curPos,
	void (*sclistDraggedFunc)(int),
	char *upbtn, char *dwnbtn,
	GSlistHandle *newList,
	GSsclistHandle *newThumb
	)
{
	GSlistHandle thList;
	GSsclistHandle sclist;
	
	int numVisItems, draght, itemht;
	thList = GSlistsetup(wname, daname, cols, labelfield, colgap, rowgap, just, thPSizeFunc, 
							label_func, settext_func, in_func, out_func, click_func);
	GSgetdragareaspecs(wname, daname, NULL, NULL, NULL, &draght, NULL, NULL);
	GSgetfieldrect(labelfield, NULL, NULL, NULL, &itemht, NULL);
	itemht += rowgap;
	numVisItems = cols * draght / itemht;
	sclist = GSscrollbarsetup(wname, thDaname, buttonCol, topLeftCol, botRightCol,
		numListItems, numVisItems, cols, curPos, sclistDraggedFunc, upbtn, dwnbtn);
	
	Thumb[sclist-FIRST_HANDLE]->sclistList = thList;
	
	if (newList != NULL)
		*newList = thList;
	if (newThumb != NULL)
		*newThumb = sclist;
}
