/*#include <windows.h>*/
#include <time.h>
#include "geosim.h"
#include "gspriv.h"

//#define EVENT_LOG // Turn on mouse event logging

/* Globals */
static HWND gHwnd;
static GSEvent gCurEvent;
static HDC gHdc;  /* needs to stay valid at all times */
static bool updating = FALSE;
static PAINTSTRUCT gPs;

static HBITMAP gBitmap[MAXBSTORES+1], gRowBitmap;
static HDC gHdcMem[MAXBSTORES+1], gRowMem;

static int gBswd[MAXBSTORES+1], gBsht[MAXBSTORES+1];

static GSColor gCurColor;

static LOGBRUSH gLogBrush;
static LOGPEN gLogPen;
static LOGFONT gSmallFont, gLargeFont;
static HFONT gHSmall, gHLarge; 
static int gLargeH, gSmallH;  /* large, small font heights */
static int gLargeA, gSmallA;  /* large, small font ascents */

static char kFontName[] = "Helvetica"; //"MS Sans Serif"; //"Times New Roman";
#define kFontHeight 12

static FontSizeType gCurFont;

static bool madePalette = FALSE;
static HPALETTE origPalette, curPalette;

#ifdef EVENT_LOG
static long evCount=0;
#endif

#define GS_TIMER 31
#define kInterval 10 // 10 ms between WM_TIMER messages - 100 per sec.
static bool gTimerSet = FALSE;

static DWORD gStartTime;
/**/

extern char Interface_Palette[3*PAL_SIZE];
static unsigned long GetPaletteColor(GSColor col)
{
	return GetNearestColor(gHdc, 
							PALETTERGB(Interface_Palette[((int) col)*3] << 2,
								Interface_Palette[(((int) col)*3)+1] << 2,
								Interface_Palette[(((int) col)*3)+2] << 2)
							);
}
/* */

/*  
	HACK ALERT:  isatty() is used in the flex scanner.  We should
	never be scanning a tty, so we're just returning false here. 
*/

int isatty(int x) { return 0; }

void GSPflush() { }

bool GShasFPU() { return TRUE; }
long GSmemavail()
{
	return GetFreeSpace(1);
}

void GSPinit_events(void) {}
void GSPclose_events(void) {}
void GSPsuspendgraphics(void) {}
void GSPresumegraphics(void) {}
void setpalettebar(void) {}
void set_mouse_position(int x, int y) {}

void GSbell(void) { MessageBeep(0); } /* can this be called outside of the WndProc? */
int mouse_present(void) { return (0 != GetSystemMetrics(SM_MOUSEPRESENT)); }

void GSPline(int x1, int y1, int x2, int y2) { POINT t; MoveToEx(gHdc, x1, y1, &t); LineTo(gHdc, x2, y2); }

void GSsetcolor(GSColor col)
{
	unsigned long palColor = GetPaletteColor(col);
	gCurColor = col;

	SetTextColor(gHdc, palColor);

	DeleteObject(SelectObject(gHdc, GetStockObject(BLACK_PEN)));
	gLogPen.lopnColor = palColor;
	SelectObject(gHdc, CreatePenIndirect(&gLogPen));

	DeleteObject(SelectObject(gHdc, GetStockObject(BLACK_BRUSH)));
	gLogBrush.lbColor = palColor;
	SelectObject(gHdc, CreateBrushIndirect(&gLogBrush));
}

GSColor GSgetcolor(void) { return gCurColor; }

void GSsetfont(FontSizeType i) 
{ 
	gCurFont = i;  
	if (i==SMALL)
		SelectObject(gHdc, gHSmall);
	else if (i==LARGE)
		SelectObject(gHdc, gHLarge);
}

FontSizeType GSPgetfont() { return gCurFont; }

int GStextheight() { return (gCurFont == SMALL)? gSmallA : gLargeA; }

int GStextwidth(char *s) 
{
	SIZE sz;
	
	if (!GetTextExtentPoint(gHdc, s, strlen(s), &sz)) 
		GSelog(-1, "GetTextExtentPoint failed!\n");
	
	return sz.cx;
}

void GSsetlinesize(int sz)
{
	if ((sz != THIN) && (sz != THICK))
		return;

	DeleteObject(SelectObject(gHdc, GetStockObject(BLACK_PEN)));
	gLogPen.lopnWidth.x = gLogPen.lopnWidth.y = sz;
	SelectObject(gHdc, CreatePenIndirect(&gLogPen));
}

int GSPgetlinesize(void) { return gLogPen.lopnWidth.x; }

void GSPinvertmode(int on) { GSPxormode(on); }

void GSPxormode(int on) { SetROP2(gHdc, on? R2_XORPEN : R2_COPYPEN); }

void GSPdraw_pixel(int x, int y)
{
	SetPixel(gHdc, x, y, GetPaletteColor(gCurColor));
}

static char translate_key(char k)
{
	if (k == '\b')
		return KeyRubout;
		
	return k; // for now - will this do arrows, backspace, etc?
}

long CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	// get an event;  stash in the global event structure

	/*
		This is where device contexts are replaced.  Question:  do old pens
		and brushes stay selected in the new context?
	*/
#ifdef EVENT_LOG
	static bool bd = FALSE;
	evCount++;

	if ((bd) && (message != WM_LBUTTONUP))
		GSelog(0, "EvCount=%ld; message=%d.\n", evCount, message);
		
#endif
	
	if (updating) {
		updating = FALSE;
		EndPaint(gHwnd, &gPs);
		gHdc = GetDC(gHwnd);
		}

	switch(message) {
		
		case WM_DESTROY:
			PostQuitMessage(0);   // will this exit through GSquit()?
			return 0;

		case WM_LBUTTONDOWN:
			 gCurEvent.type = EvMouseDown;
			 gCurEvent.mousex = LOWORD(lParam);
			 gCurEvent.mousey = HIWORD(lParam);
#ifdef EVENT_LOG
				bd = TRUE;
			 GSelog(0, "ButtonDown:  EvCount=%ld\n", evCount);
#endif
			 return 0;

		case WM_LBUTTONUP:
			 gCurEvent.type = EvMouseUp;
			 gCurEvent.mousex = LOWORD(lParam);
			 gCurEvent.mousey = HIWORD(lParam);
#ifdef EVENT_LOG
				bd = FALSE;
			 GSelog(0, "ButtonUp:  EvCount=%ld\n", evCount);
#endif
			 return 0;

		case WM_MOUSEMOVE:
			gCurEvent.type = EvNone;
			gCurEvent.mousex = LOWORD(lParam);
			gCurEvent.mousey = HIWORD(lParam);
			return 0;

		case WM_PAINT:
			gCurEvent.type = EvExpose;
			ReleaseDC(gHwnd, gHdc);
			gHdc = BeginPaint(gHwnd, &gPs);
			updating = TRUE;
			return 0;

		case WM_CHAR:
			gCurEvent.type = EvKeyDown;
			gCurEvent.key = translate_key((char) wParam);
			return 0;

		case WM_KEYDOWN:
			gCurEvent.type = EvKeyDown;
			switch((char) wParam) {
				case VK_UP:
					gCurEvent.key = KeyUp;
					return 0;
				case VK_DOWN:
					gCurEvent.key = KeyDown;
					return 0;
				case VK_RIGHT:
					gCurEvent.key = KeyRight;
					return 0;
				case VK_LEFT:
					gCurEvent.key = KeyLeft;
					return 0;
				case VK_DELETE:
					gCurEvent.key = KeyRubout;
					return 0;
				}
			gCurEvent.type = EvNone;
			break;
			
		case WM_CLOSE:
			GSquit(NULL);
			return 0; // never will get here
			
		case WM_NULL:
		case WM_TIMER:
			gCurEvent.type = EvNone;
			return 0;

		default:
			gCurEvent.type = EvNone;
			// mouse coords should be valid
			break;
		}

	return DefWindowProc(hwnd, message, wParam, lParam);
}

int GSPnext_event(GSEvent *ev)
{
	MSG msg;

	if (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);

		memcpy((char *) ev, (char *) &gCurEvent, sizeof(gCurEvent));
		}
	else
		GSquit(NULL); // right ???
	return TRUE;
}

void GSPinit_system(char *appl_name, GS_MAIN_ARGS_F)
{
	WNDCLASS wndClass;
	int i;
	HPEN hPen;
	HBRUSH hBrush;
	TEXTMETRIC tm;
	DWORD windStyle;
	RECT deskRect, cliRect, windRect;
	int deskWidth, deskHeight, cliWidth, cliHeight, windWidth, windHeight;
		
	
	if (!hPrevInstance) {
		wndClass.style = CS_OWNDC; /*keep our device context settings around */
		wndClass.lpfnWndProc = WndProc; //DefWindowProc;
		wndClass.cbClsExtra = 0;
		wndClass.cbWndExtra = 0;
		wndClass.hInstance = hInstance;
		wndClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
		wndClass.hCursor = LoadCursor(NULL, IDC_ARROW);
		wndClass.hbrBackground = GetStockObject(WHITE_BRUSH);
		wndClass.lpszMenuName = NULL;
		wndClass.lpszClassName = "gil"; //appl_name;

		RegisterClass(&wndClass);
		}

	GetWindowRect(GetDesktopWindow(), &deskRect);
	deskWidth = deskRect.right - deskRect.left;
	deskHeight = deskRect.bottom - deskRect.top;
	
	windStyle = WS_CAPTION | WS_OVERLAPPED | WS_MINIMIZEBOX | WS_SYSMENU;
	
	gHwnd = CreateWindow("gil", //appl_name,
					appl_name,
					windStyle,
					CW_USEDEFAULT,
					CW_USEDEFAULT,
					SCREENWIDTH,
					SCREENHEIGHT,
					NULL,
					NULL,
					hInstance,
					NULL);
	
	GetClientRect(gHwnd, &cliRect);
	GetWindowRect(gHwnd, &windRect);
	
	cliWidth = cliRect.right - cliRect.left;
	cliHeight = cliRect.bottom - cliRect.top;
	windWidth = windRect.right - windRect.left;
	windHeight = windRect.bottom - windRect.top;
	
	MoveWindow(gHwnd,
			(deskWidth - (SCREENWIDTH + (windWidth-cliWidth)))/2,
			(deskHeight - (SCREENHEIGHT + (windHeight - cliHeight)))/2,
			SCREENWIDTH + (windWidth-cliWidth),
			SCREENHEIGHT + (windHeight - cliHeight),
			FALSE
			);

	ShowWindow(gHwnd, nCmdShow);
	//UpdateWindow(gHwnd);
	PostMessage(gHwnd, WM_PAINT, 0, 0);
	gHdc = GetDC(gHwnd);

	for (i=1; i <= MAXBSTORES; i++) {
		gBitmap[i] = NULL;
		gHdcMem[i] = NULL;
		}

	SetBkMode(gHdc, TRANSPARENT); // text background
	SetPolyFillMode(gHdc, WINDING);  // fill in all closed regions in polygon

	hPen = SelectObject(gHdc, GetStockObject(BLACK_PEN));
	SelectObject(gHdc, hPen); // ugly way to get default object
	hBrush = SelectObject(gHdc, GetStockObject(BLACK_BRUSH));
	SelectObject(gHdc, hBrush);

	GetObject(hPen, sizeof(LOGPEN), (LPSTR) &gLogPen);
	GetObject(hBrush, sizeof(LOGBRUSH), (LPSTR) &gLogBrush);

	SelectObject(gHdc, CreatePenIndirect(&gLogPen));
	SelectObject(gHdc, CreateBrushIndirect(&gLogBrush));
		// We have to create the pen and brush so we can delete them later.
		
	memset((char *) &gSmallFont, 0, sizeof(gSmallFont));
	gSmallFont.lfHeight = kFontHeight;
	strcpy(gSmallFont.lfFaceName, kFontName);
	
	memcpy((char *) &gLargeFont, (char *) &gSmallFont, sizeof(gSmallFont));
	gLargeFont.lfWeight = 700;
	
	gHSmall = CreateFontIndirect(&gSmallFont);	
	gHLarge = CreateFontIndirect(&gLargeFont);
	
	SelectObject(gHdc, gHLarge);
	GetTextMetrics(gHdc, &tm);
	gLargeH = tm.tmHeight;
	gLargeA = tm.tmAscent;
	
	SelectObject(gHdc, gHSmall);
	GetTextMetrics(gHdc, &tm);
	gSmallH = tm.tmHeight;
	gSmallA = tm.tmAscent;
	
	gCurFont = SMALL;
	gRowBitmap = NULL;
	gRowMem = NULL;
	if ((gRowBitmap = CreateCompatibleBitmap(gHdc, SCREENWIDTH, 1)) == NULL)
		GSelog(-1, "Can't allocate row bitmap.\n");
	
	gRowMem = CreateCompatibleDC(gHdc);
	SelectObject(gRowMem, gRowBitmap);
	
	if (SetTimer(gHwnd, GS_TIMER, kInterval, NULL))
		gTimerSet = TRUE;
	else
		GSelog(-1, "Can't set timer!\n");
}

void GSdelay(unsigned long ms)
{
	unsigned long start, end;
	start = (unsigned long) clock();
	end = (unsigned long) start + ((ms * CLOCKS_PER_SEC) / 1000);
	
	while (((unsigned long) clock()) < end);
}

void GSPfill_ellipse(int left, int top, int wd, int ht)
{
	Ellipse(gHdc, left, top, left+wd, top+ht);
}

void GSPframe_ellipse(int left, int top, int wd, int ht)
{
	HBRUSH hBrush = SelectObject(gHdc, GetStockObject(HOLLOW_BRUSH));
	Ellipse(gHdc, left, top, left+wd, top+ht);
	SelectObject(gHdc, hBrush);
}

void GSPfill_rect(int left, int top, int wd, int ht)
{
	Rectangle(gHdc, left, top, left+wd, top+ht);
}

void GSPframe_rect(int left, int top, int wd, int ht)
{
	HBRUSH hBrush = SelectObject(gHdc, GetStockObject(HOLLOW_BRUSH));
	Rectangle(gHdc, left, top, left+wd, top+ht);
	SelectObject(gHdc, hBrush);
}

void GSPset_clip_rect(int left, int top, int wd, int ht)
{
	RECT rect;
	HRGN hRgn;
	
	GetClientRect(gHwnd, &rect);
	
	rect.left += left;
	rect.top += top;
	rect.right = rect.left + wd;
	rect.bottom = rect.top + ht;
	
	hRgn = CreateRectRgnIndirect(&rect);
	SelectClipRgn(gHdc, hRgn);
	DeleteObject(hRgn);
}

GSColor GSPget_pixel(int x, int y)
{
	if (madePalette)
		return (GSColor) GetNearestPaletteIndex(curPalette, GetPixel(gHdc, x, y));
	else
		return 0;
}

GSColor *GSPget_row(int r, GSColor *rowbuff)
{
	int c;
	for (c=0; c < SCREENWIDTH; c++)
		rowbuff[c] = GSPget_pixel(c, r);
	
	return rowbuff;
}


void GSPfastrow(int x, int y, GSColor *row, int n)
{
	int i;
	for (i=0; i<n; i++)
		SetPixel(gRowMem, i, 0, GetPaletteColor(row[i]));
	
	BitBlt(gHdc, x, y, n, 1, gRowMem, 0, 0, SRCCOPY);
}

void GSPdraw_string(int x, int y, char *str)
{
	char	*c;

	for (c = str;  *c != 0;  c++)
		if (!isprint(*c))  *c = ' ';

	TextOut(gHdc, x, y - ((gCurFont==SMALL)? gSmallA : gLargeA) , str, strlen(str));
	SetTextJustification(gHdc, 0, 0);
}

void GSPfillpoly(int numPoints, int p[])
{
	int i;
	POINT *pts = (POINT *) malloc(sizeof(POINT) * numPoints);
	if (pts == NULL)
		GSelog(-1, "Couldn't allocate point array in GSPfillpoly.\n");

	for (i=0; i < numPoints; i++)  {
		pts[i].x = p[2*i];
		pts[i].y = p[2*i + 1];
		}

	Polygon(gHdc, pts, numPoints);
	
	free(pts);
}

int GSPinit_bstore(int bstorenum, int wd, int ht)
{
	gBswd[bstorenum] = gBsht[bstorenum] = 0;

	if ((wd <= 0) || (ht <= 0)) {
		gBitmap[bstorenum] = NULL;
		return TRUE;
		}

	if ((gBitmap[bstorenum] = CreateCompatibleBitmap(gHdc, wd, ht)) == NULL)
		return FALSE;

	gHdcMem[bstorenum] = CreateCompatibleDC(gHdc);
	SelectObject(gHdcMem[bstorenum], gBitmap[bstorenum]);

	return TRUE;
}

/*
	Copy from screen to internal bitmap
*/

int GSPto_bstore(int bstorenum, int left, int top, int wd, int ht)
{
	if (gBswd[bstorenum] > 0)
		return FALSE;

	if (gBitmap[bstorenum] == NULL)
		GSelog(-1, "Error:  attempt to copy from null bstore %d.\n", bstorenum);

	gBswd[bstorenum] = wd;
	gBsht[bstorenum] = ht;

	BitBlt(gHdcMem[bstorenum], 0, 0, wd, ht, gHdc, left, top, SRCCOPY);

	return TRUE;
}

/*
	Copy from internal bitmap to screen
*/

int GSPfrom_bstore(int bstorenum, int left, int top)
{
	BitBlt(gHdc, left, top, gBswd[bstorenum], gBsht[bstorenum],
													gHdcMem[bstorenum], 0, 0, SRCCOPY);
	gBswd[bstorenum] = gBsht[bstorenum] = 0;
   return TRUE;
}

void GSclick(void)
{
	// How do you "CLICK" in Windoze???
}

int GSstricmp(char *s1, char *s2) /* changed 9/11/95 - was stricmp */
{
	unsigned char	c1, c2;
	
	while (1)
	{
		c1 = tolower(*s1++);
		c2 = tolower(*s2++);
		
		if (c1 < c2)		return -1;
		else if (c1 > c2)	return  1;
		else if (c1 == 0)	return  0;
	}
}

void GSPinit_palette(char pal[], int numColors)
{
	int i;
	HPALETTE hPal;
	
	LOGPALETTE *palPtr = (LOGPALETTE *) 
		LocalAlloc(LPTR, sizeof(LOGPALETTE) + numColors*sizeof(PALETTEENTRY));
	
	if (palPtr == NULL)
		GSelog(-1, "Couldn't allocate logical palette.\n");
	
	palPtr->palNumEntries = numColors;
	palPtr->palVersion = 0x0300;
	
	for (i=0; i < numColors; i++) {
		palPtr->palPalEntry[i].peRed = LOWORD(pal[3*i]<<2);
		palPtr->palPalEntry[i].peGreen = LOWORD(pal[(3*i)+1]<<2);
		palPtr->palPalEntry[i].peBlue = LOWORD(pal[(3*i)+2]<<2);
		palPtr->palPalEntry[i].peFlags = PC_RESERVED;
		}
	
	hPal = CreatePalette(palPtr);
	LocalFree((HLOCAL) palPtr);
	
	if (hPal == NULL)
		GSelog(-1, "Couldn't allocate palette.\n");
	
	if (!madePalette) {
		origPalette = SelectPalette(gHdc, hPal, FALSE);
		madePalette = TRUE;
		}
	else
		DeleteObject(SelectPalette(gHdc, hPal, FALSE));
	
	curPalette = hPal;
/*	if (RealizePalette(gHdc) == NULL)
		GSelog(-1, "Couldn't realize palette.\n");*/
	RealizePalette(gHdc);
}

void closegraph(void)
{
	/*
		This should be the ONLY place where the window can be destroyed.
	*/

	int i;

	for (i=1; i <= MAXBSTORES; i++) {
		if (gHdcMem[i] != NULL)
			DeleteObject(gHdcMem[i]);
		if (gBitmap[i] != NULL)
			DeleteObject(gBitmap[i]);
		}
	if (gRowBitmap)
		DeleteObject(gRowBitmap);
	if (gRowMem)
		DeleteObject(gRowMem);

	DeleteObject(SelectObject(gHdc, GetStockObject(BLACK_PEN)));
	DeleteObject(SelectObject(gHdc, GetStockObject(BLACK_BRUSH)));

	DeleteObject(gHSmall);
	DeleteObject(gHLarge);
	
	if (madePalette)
		DeleteObject(SelectPalette(gHdc, origPalette, FALSE));
	
	ReleaseDC(gHwnd, gHdc);

	if (gTimerSet)
		KillTimer(gHwnd, GS_TIMER);

	DestroyWindow(gHwnd);
}

void GSPstart_timer()
{
	gStartTime = GetTickCount();
}

long GSPread_timer()
{
	return GetTickCount() - gStartTime;
}
