/*------------------------------------------------------------------------------
  Core routines of the GeoSim interface library
  X-Windows version
------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Shell.h>

#include "geosim.h"
#include "gspriv.h"

extern int Cursx;
extern int Cursy;

typedef unsigned char    Uchar;
typedef unsigned short   Ushort;
typedef unsigned long    Ulong;

static Display           *Disp;          /* X display (server) we're on */
static XWindowAttributes X;              /* various screen attributes */
static Window            Win;            /* main window */
static Pixmap            Bstore[MAXBSTORES+1];      /* used for backing store */
static GC                Gc, Gc0;        /* handy graphics contexts */
static int               Bswd[MAXBSTORES+1],           /* backing-store width */
                         Bsht[MAXBSTORES+1];          /* backing-store height */
static GSColor           Curcol;         /* current color index */
static XFontStruct       *Lilfont,       /* the little (plain) font */
                         *Bigfont,       /* the big (bold) font */
                         *Curfont;       /* one of the two above */
static XColor            Pal[PAL_SIZE];       /* our color palette */
static char              Rowbuf[SCREENWIDTH*4];  /* used for fast row copying */
static XImage            *Row;           /* XImage linked to Rowbuf */

static Window            ParentWin;      /* -dhines Parent of Win */

static clock_t	Start_time; /* for timer funcs */

#if 1
#define FLUSH   XFlush(Disp)
#else
#define FLUSH
#endif

/*--------------------------------------------------------------------------*/

/* Various one-line functions, some of which are empty. */

void	set_mouse_borders(int x1, int y1, int x2, int y2) {}
void	setpalettebar      (void) {}
void	GSPinit_events     (void) {}
void	GSPclose_events    (void) {}
void	GSPclear_events    (void) {}
void	GSPsuspendgraphics (void) {}
void	GSPrestoregraphics (void) {}
void	GSclick            (void) {}

void	GSPflush           (void) { XFlush(Disp); }
FontSizeType	GSPgetfont         (void) { return Curfont==Lilfont? SMALL : LARGE; }
void	GSbell             (void) { XBell(Disp, 50); }
int	GShasFPU           (void) { return TRUE; }
int	mouse_present      (void) { return TRUE; }
long	GSmemavail         (void) { return 2000000000; }
int	GStextheight(char *str) {return Curfont->ascent + Curfont->descent - 2;}
void	GSsetcolor(GSColor c) {XSetForeground(Disp, Gc, Pal[Curcol = c].pixel);}
GSColor	GSgetcolor         (void) { return Curcol; }
void	GSPdraw_pixel      (int x, int y) { XDrawPoint(Disp, Win, Gc, x, y); }

/*--------------------------------------------------------------------------*/

void get_mouse_position(short *x, short *y)
{
  Window    w;
  int       i, tx, ty;

  XQueryPointer(Disp, Win, &w, &w, &i, &i, &tx, &ty, &i);
  *x = tx;
  *y = ty;
}


void set_mouse_position (int x, int y)    /* eliminate?? */
{
  XWarpPointer(Disp, None, Win, 0, 0, 0, 0, x, y);
}
  

static char translate_key(XEvent *xev)
{
/* Translate an X key event into an ASCII character code, or
** one of our special control key codes.
*/

  KeySym    xkey;

  XLookupString((XKeyEvent *) xev, NULL, 0, &xkey, NULL);

  switch (xkey)
  {
    case XK_Return: case XK_KP_Enter:  return KeyEnter;
    case XK_Tab:  return KeyTab;
    case XK_Up:  return KeyUp;
    case XK_Left:  return KeyLeft;
    case XK_Down:  return KeyDown;
    case XK_Right:  return KeyRight;
    case XK_BackSpace: case XK_Delete:  return KeyRubout;
    case XK_Escape:  return KeyEscape;
    case XK_F1: return KeyF1;
    default:
        return (xkey > 0  &&  xkey < 128)? (char) xkey : 0;
  }
}

void GSPinvertmode(int on)
{
  /* set plane mask to make portable from monitor to monitor -dhines */
  XSetPlaneMask(Disp, Gc, (on? 0x08: 0xff));
  XSetFunction(Disp, Gc, (on? GXinvert : GXcopy));
}

void GSPxormode(int on)
{
  /* set plane mask to make portable from monitor to monitor -dhines */
/*XSetPlaneMask(Disp, Gc, (on? 0x08: 0xff));*/
  XSetFunction(Disp, Gc, (on? GXxor : GXcopy));
}

int GSPnext_event(GSEvent *ev)
{
  /*--------------------------------------------------------------
    If there is an X event on the queue,  convert it to a GeoSim
    event in "ev" and return TRUE.  Otherwise, put a GeoSim null
    event in "ev" and return FALSE.
  --------------------------------------------------------------*/

  XEvent    xev;

  while (XCheckMaskEvent(Disp, -1L, &xev)) /* while there are X events...*/
  {
    switch (xev.type)
    {
      case ButtonPress:
      case ButtonRelease:
        if (xev.xbutton.button != Button1)
          break;
        ev->type = (xev.type==ButtonPress? EvMouseDown : EvMouseUp);
        ev->mousex = xev.xbutton.x;
        ev->mousey = xev.xbutton.y;
        return TRUE;

      case KeyPress:
        ev->type = EvKeyDown;
        ev->key = translate_key(&xev);
        ev->mousex = xev.xkey.x;
        ev->mousey = xev.xkey.y;
        return TRUE;

      case Expose:
        if (xev.xexpose.count != 0)
          break;
        ev->type = EvExpose;
        ev->mousex = xev.xexpose.x;
        ev->mousey = xev.xexpose.y;
        return TRUE;

    } /* end switch */

  } /* end while */

  /* If we arrive here, either there are no events, or there are no
     events we wish to recognize.  Thus, we return the null event.  */

  ev->type = EvNone;
  get_mouse_position(&ev->mousex, &ev->mousey);

  /************************TEST****************************/
  if ((ev->mousex != Cursx) || (ev->mousey != Cursy))
  { /*--- if mouse has moved since last time through loop ---*/
    return TRUE;
  }
  /************************TEST****************************/

  return FALSE;
}

/*--------------------------------------------------------------------------*/

int GStextwidth(char *str)
{
/* Return the pixel-width of a string.  The addition of the string's
** character length is to accomodate boldface strings -- i.e. one pixel
** per character is added.
*/

  int    len;

  len = strlen(str);
  return XTextWidth(Curfont, str, len);
}

#if 1
int GSstricmp(char *s1, char *s2)      /* changed from stricmp */
{
/* Compare two strings, like the ANSI function strcmp, but with
** insensitivity to case.
*/

  Uchar    c1, c2;

  while (1)
  {
    c1 = tolower(*s1++);
    c2 = tolower(*s2++);

    if (c1 < c2)
      return -1;
    else if (c1 > c2)
      return  1;
    else if (c1 == 0)
      return  0;
  }
}
#endif

void GSdelay (unsigned long ms)   /* formerly delay()  */
{
  clock_t    end;
  
  end = clock() + (ms*1000) / CLK_TCK;
  while ( (long) (end-clock()) > 0 )    {}
}

  
void GSPline(int x1, int y1, int x2, int y2)
{
  XDrawLine(Disp, Win, Gc, x1, y1, x2, y2);
  FLUSH;
}


GSColor GSPget_pixel(int x, int y)
{
  /*------------------------------------------------------------
    Read a pixel from the main window, and return the GeoSim
    palette index corresponding to the X pixel value we read.
    Because of the horrendously slow nature of this process,
    this function should probably be abolished.
  ------------------------------------------------------------*/

  Ulong    value;
  int      i;

  XGetSubImage(Disp, Win, x, y, 1, 1, 0xFFFFFFFF, ZPixmap, Row, 0, 0);

  switch (X.depth)
  {
    default: value = *(Uchar *) Rowbuf;  break;
    case 16: value = *(Ushort *) Rowbuf;  break;
    case 32: value = *(Ulong *) Rowbuf;  break;
  }

  for (i = 0; i < PAL_SIZE; i++)
  {
    if (Pal[i].pixel == value)
    {
      return i;
    }
  }
  return 0;
}

GSColor *GSPget_row(int row_n, GSColor *row_b)
{
  /*------------------------------------------------------------
    Read a pixel from the main window, and return the GeoSim
    palette index corresponding to the X pixel value we read.
    Because of the horrendously slow nature of this process,
    this function should probably be abolished.
  ------------------------------------------------------------*/

  Ulong    value;
  int      i;
  int      j;
  Uchar    *cbuf;
  Ushort   *sbuf;
  Ulong    *lbuf;

  if (row_b == NULL)
  {
    GSelog(-1, "Attempt to retrieve pixel row into null buffer\n");
  }

  if (row_n < 0 || row_n >= SCREENWIDTH)
  {
    GSelog(-1, "Attempt to retrieve non-existent pixel row %d\n", row_n);
  }

  XGetSubImage(Disp, Win, 0, row_n, SCREENWIDTH, 1, 0xFFFFFFFF,
               ZPixmap, Row, 0, 0);
 
  switch (X.depth)
  {
/*    default: value = *(Uchar *) Rowbuf;  break;
    case 16: value = *(Ushort *) Rowbuf;  break;
    case 32: value = *(Ulong *) Rowbuf;  break;*/
    default: cbuf = (Uchar *) Rowbuf;  break;
    case 16: sbuf = (Ushort *) Rowbuf;  break;
    case 32: lbuf = (Ulong *) Rowbuf;  break;
  }

  for (i = 0; i < SCREENWIDTH; i++)
  {
    switch (X.depth)
    {
      default: value = *cbuf++;  break;
      case 16: value = *sbuf++;  break;
      case 32: value = *lbuf++;  break;
    }
    for (j = 0; j < PAL_SIZE; j++)
    {
      if (Pal[j].pixel == value)
      {
        row_b[i] = j;
        break;
      }
    }
    if (j == PAL_SIZE)
    {
      row_b[i] = 0;
/*      return NULL;*/
    }
  }
  return row_b;
}

void GSPfastrow(int x, int y, GSColor *row, int n)
/*void GSPfastrow(int x, int y, ColorIndexType *row, int n)*/
{
  /*---------------------------------------------------------------
    Quickly copy a row of colors to the main window.  The special
    color value of 255 is currently mapped to color 0.
  ---------------------------------------------------------------*/

  int        wd = n;
  Uchar    *cbuf;
  Ushort   *sbuf;
  Ulong    *lbuf;

/*  XGetSubImage(Disp, Win, x, y, wd, 1, 0xFFFFFFFF, ZPixmap, Row, 0, 0);*/

  switch (X.depth)
  {
    default:
    case 8:
      for (cbuf = (Uchar *) Rowbuf; n-- > 0; cbuf++, row++)
        *cbuf = Pal[*row==255? 0 : *row].pixel;
      break;

    case 16:
      for (sbuf = (Ushort *) Rowbuf; n-- > 0; sbuf++, row++)
        *sbuf = Pal[*row==255? 0 : *row].pixel;
      break;

    case 32:
      for (lbuf = (Ulong *) Rowbuf; n-- > 0; lbuf++, row++)
        *lbuf = Pal[*row==255? 0 : *row].pixel;
      break;
  }

  XPutImage(Disp, Win, Gc, Row, 0, 0, x, y, wd, 1);
  FLUSH;
}

  
void GSPfill_ellipse(int left, int top, int wd, int ht)
{
  XFillArc(Disp, Win, Gc, left, top, wd, ht, 0, 360*64);
  FLUSH;
}


void GSPframe_ellipse(int left, int top, int wd, int ht)
{
  XDrawArc(Disp, Win, Gc, left, top, wd, ht, 0, 360*64);
  FLUSH;
}


void GSPfill_rect(int left, int top, int wd, int ht)
{
  XFillRectangle(Disp, Win, Gc, left, top, wd, ht);
  FLUSH;
}


void GSPframe_rect(int left, int top, int wd, int ht)
{
  if (wd > 0 && ht > 0)
    XDrawRectangle(Disp, Win, Gc, left, top, wd-1, ht-1);
  FLUSH;
}


void GSPfillpoly(int n, int p[])
{
  XPoint    xp[20];
  int       i;

  for (i = 0;  i < n;  i++)
  {
      xp[i].x = p[2*i];
      xp[i].y = p[2*i+1];
  }

  XFillPolygon(Disp, Win, Gc, xp, n, Nonconvex, CoordModeOrigin);
  FLUSH;
}

      
void GSPdraw_string(int x, int y, char *str)
{
  char    *c;

  for (c = str;  *c != 0;  c++)
    if (!isprint(*c))  *c = ' ';

  XDrawString(Disp, Win, Gc, x, y, str, c-str);
  FLUSH;
}


void GSsetlinesize (int size)
{
  if (size == THIN || size == THICK)
    XSetLineAttributes(Disp, Gc, size, LineSolid, CapRound, JoinRound);
}

int GSPgetlinesize (void)
{
  XGCValues values;

  XGetGCValues(Disp, Gc, GCLineWidth, &values);
  return (values.line_width);
}

void GSPset_focus(void)
{
  XSetInputFocus(Disp, Win, RevertToParent, CurrentTime);
}

void GSPinit_palette( char pal[],         /* palette of r g b colors */
                      int n)              /* number of colors */
{
  int      i;
  char    *c;

  for (i = 0, c = pal;  i < n;  i++, c += 3)
  {
    Pal[i].red   = ((unsigned short) c[0]) << 10;
    Pal[i].green = ((unsigned short) c[1]) << 10;
    Pal[i].blue  = ((unsigned short) c[2]) << 10;
    XAllocColor(Disp, X.colormap, Pal+i);
  }
}


void GSsetfont(FontSizeType type)
{
    Curfont = type==SMALL? Lilfont : Bigfont;
    XSetFont(Disp, Gc, Curfont->fid);
}


void GSPset_clip_rect(int left, int top, int wd, int ht)
{
  XRectangle    r;

  r.x = left;
  r.y = top;
  r.width = wd;
  r.height = ht;

  XSetClipRectangles(Disp, Gc, 0, 0, &r, 1, Unsorted);
}

/*--------------------------------------------------------------------------*/

int GSPinit_bstore(int bstorenum, int wd, int ht)
{
  Bswd[bstorenum] = Bsht[bstorenum] = 0;
  if ((wd <= 0) || (ht <= 0))
  {
    Bstore[bstorenum] = NULL;
    return True;
  }

  Bstore[bstorenum] = XCreatePixmap(Disp, Win, wd, ht, X.depth);
  return Bstore? True : False;
}


int GSPto_bstore(int bstorenum, int left, int top, int wd, int ht)
{
  if (Bswd[bstorenum] > 0)  return False;
  if (Bstore[bstorenum] == NULL)
    GSelog(-1, "Error: attempt to copy to null bstore %d.\n",
                         bstorenum);

  XFlush(Disp);
  XCopyArea(Disp, Win, Bstore[bstorenum], Gc0, left, top, 
            Bswd[bstorenum]=wd, Bsht[bstorenum]=ht, 0, 0);
  return True;
}


int GSPfrom_bstore(int bstorenum, int left, int top)
{
  if (Bswd[bstorenum] < 1)  return False;
  if (Bstore[bstorenum] == NULL)
    GSelog(-1, "Error: attempt to copy from null bstore %d.\n", bstorenum);
  XCopyArea(Disp, Bstore[bstorenum], Win, Gc0, 0, 0, 
			Bswd[bstorenum], Bsht[bstorenum], left, top);
  FLUSH;
  Bswd[bstorenum] = Bsht[bstorenum] = 0;
  return True;
}

/*--------------------------------------------------------------------------*/

void GSPinit_system (
  char	*appl_name,
  int	argc,
  char	**argv
)
{
/*  int         argc = 0;*/
  Widget      top, draw;
  XSizeHints  xsh;
  Arg         args[7];

  Window      *cw; /* child windows; dummy var for XQueryTree */
  Window      rw;  /* root window; dummy var for XQueryTree */
  int         nc;  /* number of children; dummy var for XQueryTree */

  top = XtInitialize("", "", NULL, 0, &argc, argv);
  draw = XtVaCreateManagedWidget("", coreWidgetClass, top, XtNwidth,
                                 SCREENWIDTH, XtNheight, SCREENHEIGHT, NULL);

  /*--- Set resource values ---*/
  if (appl_name != NULL)
  {
    XtSetArg(args[0], XtNtitle, appl_name);
    XtSetArg(args[1], XtNiconName, appl_name);
  }
  else
  {
    XtSetArg(args[0], XtNtitle, "GIL Application");
    XtSetArg(args[1], XtNiconName, "GIL Application");
  }
  /*--- args 2-5 prevent resizing of GIL application window ---*/
  XtSetArg(args[2], XtNminHeight, SCREENHEIGHT);
  XtSetArg(args[3], XtNmaxHeight, SCREENHEIGHT);
  XtSetArg(args[4], XtNminWidth, SCREENWIDTH);
  XtSetArg(args[5], XtNmaxWidth, SCREENWIDTH);
  XtSetArg(args[6], XtNinput, TRUE);
  XtSetValues(top, (ArgList)&args, 7);

  XtSetMappedWhenManaged(top, False);
  XtRealizeWidget(top);
  Disp = XtDisplay(top);
  Win = XtWindow(draw);
  XGetWindowAttributes(Disp, Win, &X);

  if (X.depth < 8)
    GSelog(-1, "Screen cannot support 8-bit color!\n");

  Gc0 = XCreateGC(Disp, Win, 0, NULL);
  Gc  = XCreateGC(Disp, Win, 0, NULL);
  Row = XCreateImage(Disp, X.visual, X.depth,
    ZPixmap, 0, Rowbuf, SCREENWIDTH, 1, X.depth, 0);
  
  Lilfont = Curfont =
    XLoadQueryFont(Disp, "-*-helvetica-medium-r-*-*-11-*-*-*-*-*-*-*");
  if (Lilfont == NULL)
    Lilfont = Curfont = 
      XLoadQueryFont(Disp, "-*-helvetica-medium-r-*-*-10-*-*-*-*-*-*-*");
  if (Lilfont == NULL)
    GSelog(-1, "Unable to load small font");  

  Bigfont = XLoadQueryFont(Disp, "-*-helvetica-bold-r-*-*-11-*-*-*-*-*-*-*");
  if (Bigfont == NULL)
    Bigfont = XLoadQueryFont(Disp, "-*-helvetica-bold-r-*-*-10-*-*-*-*-*-*-*");
  if (Bigfont == NULL)
    GSelog(-1, "Unable to load large font");  

  XSetFont(Disp, Gc, Curfont->fid);
  
  XSelectInput(Disp, Win, ButtonPressMask | ButtonReleaseMask | KeyPressMask |
               ExposureMask);

  XMapRaised (Disp, Win);

  xsh.min_width = xsh.max_width = SCREENWIDTH;
  xsh.min_height = xsh.max_height = SCREENHEIGHT;
  xsh.flags = PMinSize | PMaxSize;
  XSetWMNormalHints(Disp, Win, &xsh);

  XtMapWidget(top);

  /*--- discover parent window of Win ---*/
  XQueryTree(Disp, Win, &rw, &ParentWin, &cw, &nc);
  /*------------------------------------------------------------
    select FocusIn, FocusOut events for parent window; FocusIn
    is sent to parent window (not Win) when the user clicks on
    the window title bar, for example.  When the parent window
    gets focus,  focus is given explicitly to the child window
    to allow keyboard input, regardless of where pointer is. 
  ------------------------------------------------------------*/

  XSelectInput(Disp, ParentWin, KeyPressMask);
}
  

void closegraph(void)
{
  if (Disp)               /*--- if !Disp, then init_system not called yet ---*/
    XtCloseDisplay(Disp);
}

void GSPstart_timer(void) { Start_time = clock(); }

long GSPread_timer(void)
  {
  /*
    CLK_TCK is the number of processor "clicks" per second.

    CLK_TCK should be defined in the platform's
    time.h header.  If not, define it in the XXXsys.h
    header.  JBR
  */
  long curTime = (long) ((clock() - Start_time)*1000 / CLK_TCK);

  return curTime;
}

/*==========================================================================*/
